/*======================================================================================================================
  (c) Copyright OMRON Europe BV 2012
  All Right Reserved
  OMRON Development & Engineering
  -----------------------------------------------------------------------------------------------------------------*//**
  @file

  Implements helper functions for the generation of the ARM A-tags list required to start the Linux kernel

*//*--------------------------------------------------------------------------------------------------------------------
      Unique identifier generated by svn:
      $Header$

======================================================================================================================*/

/*=== Include files ==================================================================================================*/
#include "atags.h"
#include "debug.h"
#include "string.h"

/*=== Constant definitions (File scope) ==============================================================================*/

/*=== Macro definitions (File scope) =================================================================================*/

/*=== Type definitions (File scope) ==================================================================================*/

/*=== Type definitions for events (File scope) =======================================================================*/

/*=== Function declarations (File scope) =============================================================================*/

/*=== Macro-function definitions (File scope) ========================================================================*/

/*=== Data definitions (File scope) ==================================================================================*/

/*=== Local Function definitions (File scope) ========================================================================*/

/*=== Global function definitions (Global system or Local sub-system scope) ==========================================*/
/**
 * Formats an ARM start A-tag
 * Used to indicate the start of the specified A-tags
 * @param[out] ptag             Destination memory for the start A-tag
 * @return     The memory address following the generated start A-tag
 */
struct tag* atags_addStartTag(struct tag* ptag)
{
  ptag->hdr.tag = ATAG_CORE;
  ptag->hdr.size = tag_size (tag_core);

  ptag->u.core.flags = 0;
  ptag->u.core.pagesize = 0;
  ptag->u.core.rootdev = 0;

  return tag_next(ptag);
}

/**
 * Formats an ARM memory A-tag
 * @param[out] ptag             Destination memory for the ARM memory A-tag
 * @param[in]  start            The start address of the RAM that Linux should use
 * @param[in]  size             The size of the RAM that Linux should use
 * @return     The memory address following the generated memory A-tag
 */
struct tag* atags_addMemoryTag(struct tag*  ptag,
                               UNS32        start,
                               UNS32        size )
{
  ptag->hdr.tag = ATAG_MEM;
  ptag->hdr.size = tag_size (tag_mem32);

  ptag->u.mem.start = start;
  ptag->u.mem.size = size;

  return tag_next(ptag);
}

/**
 * Formats a Linux cmdline A-tag
 * @param[out] ptag             Destination memory for the cmdline A-tag
 * @param[in]  linuxCmdLine     The Linux cmdline string to add to the A-Tags. If @p NULL, the cmdline A-tag is skipped.
 * @param[in]  linuxCmdLineAdd  Additional string data for the Linux cmdline. If @p NULL, will be skipped.
 * @return     The memory address following the generated cmdline A-tag
 */
struct tag* atags_addCommandlineTag(struct tag*                 ptag,
                                    char const* const           commandline,
                                    char const* const* const    commandlineAdd )
{
  char const* p;
  unsigned commandLineLen;

  if (!commandline)
    return ptag;

  /* eat leading white space */
  for (p = commandline; *p == ' '; p++);

  /* skip non-existent command lines so the kernel will still
   * use its default command line.
   */
  if (*p == '\0')
    return ptag;

  strcpy (ptag->u.cmdline.cmdline, p);
  commandLineLen = strlen(p);

  if(commandlineAdd != NULL)
  {
    unsigned index = 0;

    while (commandlineAdd[index] != NULL)
    {
      char const* a;

      /* Continue with a separator */
      ptag->u.cmdline.cmdline[commandLineLen++] = ' ';

      /* eat leading white space */
      for (a = commandlineAdd[index]; *a == ' '; a++);

      strcpy (ptag->u.cmdline.cmdline + commandLineLen, a);
      commandLineLen += strlen(a);

      index++;
    }
  }

  ptag->hdr.tag = ATAG_CMDLINE;
  ptag->hdr.size =
    (sizeof (struct tag_header) + commandLineLen + 1 + 4) >> 2;

  return tag_next(ptag);
}

/**
 * Formats an ARM initrd A-tag
 * @param[out] ptag             Destination memory for the initrd A-tag
 * @param[in]  initrd_start     The start address of the initrd image
 * @param[in]  initrd_end       The end address of the initrd image
 * @return     The memory address following the generated initrd A-tag
 */
struct tag* atags_addInitrdTag(struct tag* ptag, UNS32 initrd_start, UNS32 initrd_end)
{
  /* an ATAG_INITRD node tells the kernel where the compressed
   * ramdisk can be found. ATAG_RDIMG is a better name, actually.
   */
  ptag->hdr.tag = ATAG_INITRD2;

  ptag->hdr.size = tag_size (tag_initrd);
  ptag->u.initrd.start = initrd_start;
  ptag->u.initrd.size = initrd_end - initrd_start;

  return tag_next(ptag);
}

/**
 * Formats an ARM end A-tag
 * Used to indicate the end of the specified A-tags
 * @param[out] ptag             Destination memory for the end A-tag
 * @return     The memory address following the generated end A-tag
 */
struct tag* atags_addEndTag(struct tag* ptag)
{
  ptag->hdr.tag = ATAG_NONE;
  ptag->hdr.size = 0;

  return tag_next(ptag);
}


/* END OF FILE */
