/*======================================================================================================================
  (c) Copyright OMRON Europe BV 2011
  All Right Reserved
  OMRON Development & Engineering
  -----------------------------------------------------------------------------------------------------------------*//**
  @file

  This module generates and parses O-tags.
  O-tags are used to configure parameters for the Linux kernel. For each image (a board can have multiple images like
  the application and maintenance image) an O-tag can be defined. It is also possible to create an O-tag which applies to
  all images. The number of tags is only limited by the reserved memory space to store them. Unknown tags are skipped and
  not parsed. However, it is possible to include them in a list.
*//*--------------------------------------------------------------------------------------------------------------------
      Unique identifier generated by svn:
      $Header$

======================================================================================================================*/

/*=== Include files ==================================================================================================*/
#include "otags.h"
#include "cksum.h"
#include "string.h"

/*=== Constant definitions (File scope) ==============================================================================*/
/* O-TAGS types */

/**
 * The O-tag identifier for an ID O-tag.
 * This O-tag type must be the first in a valid O-tag list.
 */
#define OTAG_ID             (0x544F0000)

/**
 * The O-tag identifier for an MAC O-tag.
 */
#define OTAG_MAC            (0x544F0001)

/**
 * The O-tag identifier for an UPC string O-tag.
 */
#define OTAG_UPC_STR        (0x544F0002)

/**
 * The O-tag identifier for the additional Linux command-line arguments O-tag.
 */
#define OTAG_LNX_CMD_L_ADD  (0x544F0003)

/**
 * The O-tag identifier for the Linux command-line override O-tag.
 */
#define OTAG_LNX_CMD_L_OVR  (0x544F0004)

/**
 * The O-tag identifier for an END O-tag.
 * This O-tag type must be the last O-tag in a list to terminate it.
 */
#define OTAG_END            (0x544F0FFF)

/* Current O-tag subsystem config */
/**
 * The magic number used in the ID O-tag to indicate a valid list.
 */
#define OTAG_MAGIC_NR       (0x4F4D524E)

/**
 * The major version of the O-tag protocol.
 */
#define OTAG_MAJOR_VER (1)

/**
 * The minor version of the O-tag protocol.
 */
#define OTAG_MINOR_VER (1)

/* Tag options */
/**
 * Option to indicate for which image the tag is applicable.
 * 0x80 means for image 0. 0x81 for image 1, ... 0xFF means applicable for all known images.
 */
#define USE_FOR_IMAGE_X         (0x80U)
#define USE_FOR_IMAGE_BIT_MASK  (0x7FU)

/*=== Macro definitions (File scope) =================================================================================*/

/*=== Type definitions (File scope) ==================================================================================*/
/**
 *  The header which is used in every O-tag.
 */
typedef struct {
  UNS32 tagId;    /**< One of the unique O-tag identifiers */
  UNS32 options;  /**< On of the options for an O-tag */
  UNS32 size;     /**< The size (in UNS32) for a complete O-tag, including header, padding, etc */
  UNS32 crc32;    /**< The CRC-32 for a complete O-tag */
} T_OTAG_HEADER;

/**
 * Structure containing the members for an ID O-tag
 */
typedef struct {
  UNS32 magicNumber;  /**< The special magic number to indicate a valid O-tag list */
  UNS16 otagVerMajor; /**< The major version of the list */
  UNS16 otagVerMinor; /**< The minor version of the list */
} T_OTAG_ID;

/**
 * Structure containing the members for an UPC string O-tag
 */
typedef struct {
  UNS8  str[1];         /**< At least one byte for storage. However, string can be extended passed this one byte */
} T_OTAG_UPC_STR;

/**
 * Structure containing the members for an Linux commandline string O-tag
 */
typedef struct {
  char  str[1];         /**< At least one byte for storage. However, string can be extended passed this one byte */
} T_OTAG_LINUX_CMD_LINE;

/**
 * Structure containing the members for an end O-tag
 */
typedef struct {
  UNS32 dummy;          /**< ot used, just reserving some mem space */
} T_OTAG_END;

/**
 * A complete O-tag definition.
 */
typedef struct {
  T_OTAG_HEADER hdr;
  union {
    T_OTAG_ID             id;
    T_OTAG_MAC            mac;
    T_OTAG_UPC_STR        upc;
    T_OTAG_LINUX_CMD_LINE linuxCmd;
    T_OTAG_END            end;
  } u;
} T_OTAG;

/*=== Type definitions for events (File scope) =======================================================================*/

/*=== Function declarations (File scope) =============================================================================*/

/*=== Macro-function definitions (File scope) ========================================================================*/
/**
 * Spools to the next O-Tag
 * @param[in] t     The previous tag
 */
#define NEXT_OTAG(t)      ((T_OTAG*)((UNS32*)(t) + (t)->hdr.size))

/**
 * Determines if a found O-tag needs to be processed for the current image.
 * @remarks Needs to be used in a condition check (like 'if').
 * @param [in] *tag       The tag for which to determine processing;
 * @param [in] bootImage  The current boot image to test the O-tag options against;
 */
#define NEED_TAG_PROCESSING(tag, bootImage) (((tag->hdr.options & USE_FOR_IMAGE_BIT_MASK) == USE_FOR_IMAGE_BIT_MASK) ||\
                                             ((tag->hdr.options & USE_FOR_IMAGE_BIT_MASK) == (unsigned)bootImage))

/*=== Data definitions (File scope) ==================================================================================*/

/*=== Local Function definitions (File scope) ========================================================================*/

/*=== Global function definitions (Global system or Local sub-system scope) ==========================================*/

/*=== Global function definitions (Global system or Local sub-system scope) ==========================================*/

/**
 * Returns the O-tag identifier to indicate which O-tag is currently active.
 * @param [in] *tag A pointer to the O-tag currently active;
 * @return The unique 32-bits O-tag identifier.
 */
UNS32 otags_getType(T_OTAG const* const tag)
{
  return tag->hdr.tagId;
}

/**
 * Retrieves the ID part for an @p OTAG_ID O-tag.
 * @param [in] *tag A pointer to the O-tag currently active;
 * @return A @p T_OTAG_ID pointer to the struct in the current O-tag.
 */
T_OTAG_ID const* getIdTag(T_OTAG const* const tag)
{
    T_OTAG_ID const* id;
    UNS32 const otagType = otags_getType(tag);

    if (otagType == OTAG_ID) {
        id = &tag->u.id;
    } else {
        id = NULL;
    }
    return id;
}

/**
 * Checks the CKSUM for the current O-tag.
 * @param [in] *tag A pointer to the O-tag currently active;
 * @return @p TRUE if the O-tag is valid and has a correct CRC. Otherwise @p FALSE.
 */
BOOL verifyCrc(T_OTAG const* const tag)
{
    UNS32   crc32;
    crc32 = cksum_calculateSubstituted(&tag->hdr, tag->hdr.size * sizeof(UNS32),
                                       &tag->hdr.crc32, 0U, sizeof(tag->hdr.crc32));

    return ((crc32 == tag->hdr.crc32) ? TRUE : FALSE);
}

/**
 * Determines if an O-tag list is actually an O-tag list.
 * @param [in] *list Start of the O-tag list.
 * @return @p TRUE if the list is valid. Otherwise @p FALSE.
 */
BOOL isValidList(T_OTAG const* const list)
{
  T_OTAG const*             currentOtag = list;
  BOOL                      isValidList = FALSE;
  BOOL                      validCrc;
  T_OTAG_ID const* const    idTag = getIdTag(currentOtag);

  //Get the first otag which should include a magic number and version.
  if (idTag != NULL)
  {
    //Check the CRC
    validCrc = verifyCrc(currentOtag);

    if ((validCrc             == TRUE)                    &&
        (idTag->magicNumber   == OTAG_MAGIC_NR)           &&
        (idTag->otagVerMajor  == OTAG_MAJOR_VER)          &&
        (idTag->otagVerMinor  == OTAG_MINOR_VER))
    {
      //We have found a valid list.
      isValidList = TRUE;
    }
    else
    {
      isValidList = FALSE;
    }
  }

  return isValidList;
}

/**
 * Parse O-tags from memory
 * @param[in]   buffer
 * @param[in]   imageNr
 * @param[out]  linuxCmdLine
 * @return
 */
T_OTAG_VALUES* otags_parseList(void const* const buffer, unsigned const imageNr)
{
    T_OTAG const* const         baseList            = (T_OTAG const* const ) buffer;
    BOOL                        listHasCrcErrors    = FALSE;
    BOOL const                  isValid             = isValidList(baseList);
    static T_OTAG_VALUES const  otagInit            = OTAG_VALUES_INIT;
    static T_OTAG_VALUES        otagValues;
    T_OTAG const*               currentOtag;
    UNS32                       type;
    INT32                       strSize;
    BOOL                        validCrc;

    // Re-initialise the data
    otagValues = otagInit;

    if (isValid != FALSE) {
        //Valid O-tag list. Parse it.
        //We now skip the first O-tag because it is an ID-tag.
        currentOtag = NEXT_OTAG(baseList);

        while (currentOtag != NULL) {
            type        = otags_getType(currentOtag);
            validCrc    = verifyCrc(currentOtag);

            if( (NEED_TAG_PROCESSING(currentOtag, imageNr)) &&
                (validCrc == TRUE) ) {
                switch (type) {
                    case OTAG_MAC: {
                        if (otagValues.mac.count < ARRAY_SIZE(otagValues.mac.address)) {
                            otagValues.mac.address[otagValues.mac.count] = &currentOtag->u.mac;
                            otagValues.mac.count++;
                        } else {
                            otagValues.errors |= OTAG_TOO_MANY_MACS;
                        }
                        break;
                    }
                    case OTAG_LNX_CMD_L_ADD: {
                        strSize = (currentOtag->hdr.size*sizeof(UNS32)) - sizeof(currentOtag->hdr);
                        if(strSize > 0 && otagValues.linuxCmdAdd.string == NULL) {
                            otagValues.linuxCmdAdd.string  = currentOtag->u.linuxCmd.str;
                            otagValues.linuxCmdAdd.length  = strSize;
                        } else {
                            otagValues.errors |= OTAG_TOO_MANY_LINUX_ADD_CMDS;
                        }
                        break;
                    }
                    case OTAG_LNX_CMD_L_OVR: {
                        strSize = (currentOtag->hdr.size*sizeof(UNS32)) - sizeof(currentOtag->hdr);
                        if(strSize > 0 && otagValues.linuxCmdOverride.string == NULL) {
                            otagValues.linuxCmdOverride.string  = currentOtag->u.linuxCmd.str;
                            otagValues.linuxCmdOverride.length  = strSize;
                        } else {
                            otagValues.errors |= OTAG_TOO_MANY_LINUX_OVR_CMDS;
                        }
                        break;
                    }
                    case OTAG_UPC_STR: {
                        strSize = (currentOtag->hdr.size*sizeof(UNS32)) - sizeof(currentOtag->hdr);
                        if(strSize > 0  && otagValues.upc.string == NULL) {
                            otagValues.upc.string       = currentOtag->u.upc.str;
                            otagValues.upc.length       = strSize;
                        } else {
                            otagValues.errors |= OTAG_TOO_MANY_UPC_STRINGS;
                        }
                        break;
                    }
                    default: {
                        //Unsupported tag.
                    }
                }
            } else if (validCrc == FALSE) {
                //Invalid CRC. Report it.
                otagValues.errors |= OTAG_CRC_ERRORS;
                listHasCrcErrors = TRUE;
            }
            if (type != OTAG_END) {
                currentOtag = NEXT_OTAG(currentOtag);
            } else {
                currentOtag = NULL;
            }
        }
    } else {
        otagValues.errors |= OTAG_NO_LIST;
    }
    return &otagValues;
}

/* END OF FILE */
