/*
 * TI Booting and Flashing Utilities
 *
 * Generic SPI driver file for common SPI peripheral
 *
 * Copyright (C) 2009 Texas Instruments Incorporated - http://www.ti.com/
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as 
 * published by the Free Software Foundation version 2.
 *
 * This program is distributed "as is" WITHOUT ANY WARRANTY of any
 * kind, whether express or implied; without even the implied warranty
 * of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 */

/* --------------------------------------------------------------------------
  AUTHOR      : Daniel Allred
 --------------------------------------------------------------------------- */

// General type include
#include "tistdtypes.h"

// Device specific CSL
#include "device.h"

// Util functions
#include "util.h"

// This module's header file 
#include "spi.h"


/************************************************************
* Explicit External Declarations                            *
************************************************************/


/************************************************************
* Local Macro Declarations                                  *
************************************************************/

/************************************************************
* Local Function Declarations                               *
************************************************************/

static Uint32 LOCAL_setupMode(SPI_InfoHandle hSPIInfo);

/************************************************************
* Local Variable Definitions                                *
************************************************************/


/************************************************************
* Global Variable Definitions                               *
************************************************************/

#ifdef USE_IN_ROM
SPI_InfoObj gSPIInfo;
#endif


/************************************************************
* Global Function Definitions                               *
************************************************************/

// Initialze NAND interface and find the details of the NAND used
SPI_InfoHandle SPI_open(Uint32 const        spiPeripheralNum,
                        Uint32 const        csNum,
                        SPI_Role const      role,
                        SPI_Mode const      mode,
                        SPI_Config* const   config)
{
  DEVICE_SPIRegs *SPI;
  SPI_InfoHandle hSPIInfo;
  
  if (spiPeripheralNum >= SPI_PERIPHERAL_CNT)
    return NULL;
    
#ifdef USE_IN_ROM
  hSPIInfo = (SPI_InfoHandle) &gSPIInfo;
#else
  hSPIInfo = (SPI_InfoHandle) UTIL_allocMem(sizeof(SPI_InfoObj));
#endif

  // Assign the correct register base
  hSPIInfo->peripheralNum = spiPeripheralNum;  
  hSPIInfo->csNum = csNum;
  hSPIInfo->regs = (void *) (((Uint32)SPI0) + (SPI_MEMORYMAP_SPACING *spiPeripheralNum) );
  SPI = (DEVICE_SPIRegs *) hSPIInfo->regs;
  
  // Assign mode and role
  hSPIInfo->role = role;
  hSPIInfo->mode = mode;
  hSPIInfo->config = config;
  
  // Reset the SPI
  SPI_reset(hSPIInfo);

  LOCAL_setupMode(hSPIInfo);

  // Disable CS if in 3-pin mode
  SPI_disableCS(hSPIInfo);
  
  // Enable SPI
  SPI->SPIGCR1 |= ( 1 << 24 );
  
  return hSPIInfo;
}

// Routine to reset the SPI device
Uint32 SPI_reset(SPI_InfoHandle const hSPIInfo)
{
  DEVICE_SPIRegs *SPI = (DEVICE_SPIRegs *) hSPIInfo->regs;

  // Put the SPI in reset
  SPI->SPIGCR0 = 0x00;
  UTIL_waitLoop( 1000 );

  // Release SPI from reset
  SPI->SPIGCR0 = 0x01;

  return E_PASS;
}

Uint32 SPI_readBytes(SPI_InfoHandle const hSPIInfo, Uint32 const byteCnt, void* const dest)
{
    Uint32 i;

    if (hSPIInfo->config->charLen == 16) {
        Uint16* const tempDest = (Uint16* const) dest;

        for (i = 0; i < (byteCnt >> 1); i++) {
            tempDest[i] = SPI_xferOneChar(hSPIInfo, 0x0000);
        }
    } else if (hSPIInfo->config->charLen == 8) {
        Uint8* const tempDest = (Uint8* const) dest;

        for (i = 0; i < byteCnt; i++) {
            tempDest[i] = SPI_xferOneChar(hSPIInfo, 0x00);
        }
    }

    return E_PASS;
}

Uint32 SPI_writeBytes(SPI_InfoHandle const hSPIInfo, Uint32 const byteCnt, void const* const src)
{
  Uint32 i;

  if (hSPIInfo->config->charLen == 16)
  {
    Uint16 const* const tempSrc = (Uint16 const* const) src;

    for (i=0; i< (byteCnt>>1); i++)
    {
      SPI_xferOneChar(hSPIInfo,tempSrc[i]);
    }
  }
  else if (hSPIInfo->config->charLen == 8)
  {
    Uint8 const* const tempSrc = (Uint8 const* const) src;

    for (i=0; i< byteCnt; i++)
    {
      SPI_xferOneChar(hSPIInfo,tempSrc[i]);
    }
  }

  return E_PASS;
}
void SPI_sync(SPI_InfoHandle const hSPIInfo,
              Uint32 const         srcByteCnt,
              void const* const    src,
              Uint32 const         destByteCnt,
              void* const          dest)
{
    SPI_enableCS(hSPIInfo);
    SPI_writeBytes(hSPIInfo, srcByteCnt, src);

    if (srcByteCnt < destByteCnt) {
        SPI_readBytes(hSPIInfo, destByteCnt - srcByteCnt, dest + srcByteCnt);
    }
    SPI_disableCS(hSPIInfo);
}

Uint32 SPI_xferOneChar(SPI_InfoHandle const hSPIInfo, Uint32 const dataOut)
{
  Uint32 spiflg, spibuf;
  DEVICE_SPIRegs *SPI = (DEVICE_SPIRegs *) hSPIInfo->regs;

  Uint32 mask = (0x1 << hSPIInfo->config->charLen) - 1;

  // Write output data
  SPI->SPIDAT0 = dataOut & mask;

  do 
  { 
    spiflg = SPI->SPIFLG;
    if (spiflg & SPI_RXINTFLAG)
    {
      spibuf = (SPI->SPIBUF);
      break;
    }
    if (spiflg & SPI_OVRNINTFLG)
    {
      spibuf = (SPI->SPIBUF);
      SPI->SPIFLG &= SPI_OVRNINTFLG;
      continue;
    }
    if (spiflg & SPI_BITERRFLG)
    {                      
      spibuf = 0x0;
      break;
    }
  } 
  while (TRUE);

  return (spibuf & mask);
}

void SPI_enableCS(SPI_InfoHandle const hSPIInfo)
{
  DEVICE_SPIRegs* const   SPI    = (DEVICE_SPIRegs *) hSPIInfo->regs;
  SPI_Config const* const config = hSPIInfo->config;

  if (hSPIInfo->mode != SPI_MODE_3PIN)
    return;

  // Clear GPIO
  SPI->SPIPC5 = 0x1 << hSPIInfo->csNum;
  UTIL_waitLoopAccurate (config->delayAfterCSdown);
}

void SPI_disableCS(SPI_InfoHandle const hSPIInfo)
{
  DEVICE_SPIRegs *SPI = (DEVICE_SPIRegs *) hSPIInfo->regs;
  SPI_Config const* const config = hSPIInfo->config;

  if (hSPIInfo->mode != SPI_MODE_3PIN)
    return;

  UTIL_waitLoopAccurate (config->delayBeforeCSup);

  // Set GPIO
  SPI->SPIPC4 = 0x1 << hSPIInfo->csNum;

  UTIL_waitLoopAccurate (config->delayAfterCSup);
}


// Defining this macro for the build will cause write (flash) ability to be removed
// This can be used for using this driver as read-only for ROM code
#ifndef USE_IN_ROM    


#endif

/************************************************************
* Local Function Definitions                                *
************************************************************/

static Uint32 LOCAL_setupMode(SPI_InfoHandle hSPIInfo)
{
  DEVICE_SPIRegs *SPI = (DEVICE_SPIRegs *) hSPIInfo->regs;
  
  if (hSPIInfo->role == SPI_ROLE_MASTER)
  {
    SPI->SPIGCR1 = 0
        | ( 0 << 24 )
        | ( 0 << 16 )
        | ( 1 << 1 )
        | ( 1 << 0 );
  }
  else if (hSPIInfo->role == SPI_ROLE_SLAVE)
  {
    SPI->SPIGCR1 = 0
        | ( 0 << 24 )
        | ( 0 << 16 )
        | ( 0 << 1 )
        | ( 0 << 0 ); 
  }
  else
  {
    return E_FAIL;
  }
   
  switch (hSPIInfo->mode)
  {
    case SPI_MODE_3PIN:
      SPI->SPIPC0 =  0
        | ( 1 << 11 )   // SOMI
        | ( 1 << 10 )   // SIMO
        | ( 1 << 9 );   // SPI_CLK
      SPI->SPIPC4 |= 0x01;
      SPI->SPIPC1 |= 0x01;  // Make CS a GPIO output 
      break;
    case SPI_MODE_4PIN_CS:
      SPI->SPIPC0 =  0
        | ( 1 << 11 )   // DI
        | ( 1 << 10 )   // DO
        | ( 1 << 9 )    // CLK
        | ( 1 << 0 );   // CS0
      break;
    case SPI_MODE_4PIN_ENA:
    SPI->SPIPC0 =  0
        | ( 1 << 11 )   // DI
        | ( 1 << 10 )   // DO
        | ( 1 << 9 )    // CLK
        | ( 1 << 8 );   // ENA
      break;
    case SPI_MODE_5PIN:
      SPI->SPIPC0 =  0
        | ( 1 << 11 )   // DI
        | ( 1 << 10 )   // DO
        | ( 1 << 9 )    // CLK
        | ( 1 << 8 )    // ENA
        | ( 1 << 0 );   // CS0
      break;
    default: 
      return E_FAIL;
  }
  
  SPI->SPIFMT[0] = 0x0;
  if (hSPIInfo->config->polarity)
    SPI->SPIFMT[0] |= ( 1 << 17 );   // Polarity
  if (hSPIInfo->config->phase)
    SPI->SPIFMT[0] |= ( 1 << 16 );   // Phase
  SPI->SPIFMT[0] |= (hSPIInfo->config->charLen & 0xFF);
  SPI->SPIFMT[0] |= ((hSPIInfo->config->prescalar & 0xFF) << 8);

  // CSHOLD off, FMT[0] used  
  SPI->SPIDAT1 = 0x00;

  // All chip selects go high when no transfer
  SPI->SPIDEF = 0xFF;

  // Disable interrupts
  SPI->SPIINT = 0x00;
  SPI->SPILVL = 0x00;
  
  return E_PASS;
}



/***********************************************************
* End file                                                 *
***********************************************************/

/* --------------------------------------------------------------------------
  HISTORY
    v1.00 - DJA - 19-Aug-2008
      Initial release
-------------------------------------------------------------------------- */

