/*======================================================================================================================
  (c) Copyright OMRON Europe BV 2011
  All Right Reserved
  OMRON Development & Engineering
  -----------------------------------------------------------------------------------------------------------------*//**
  @file

  <short description of the file in one sentence>.

  [Optional extended description of the file. More details]

  [@pre <pre-conditions and Environment restrictions>]
*//*--------------------------------------------------------------------------------------------------------------------
      Unique identifier generated by svn:
      $Header$

======================================================================================================================*/

/*=== Include files ==================================================================================================*/
#include "switch.h"
#include "spi.h"
#include "debug.h"
#include "string.h"
#include "device.h"
#include "util.h"

/*=== Constant definitions (File scope) ==============================================================================*/
#define MAX_NR_SWITCH_PORTS     (4)
#define RESET_ASSERT_TIME_NS    NS_DELAY_TO_LOOPCNT(12E6)    /* > 10 ms - tSR Stable Supply Voltages to Reset High */
#define RESET_DELAY_TIME_NS     NS_DELAY_TO_LOOPCNT(100E3)   /* > 10 us - Power up the board and assert a reset signal.
                                                                          After reset wait 100s, the start switch bit
                                                                          in Register 1 will be set to e0f. Configure
                                                                          the desired settings in the KSZ8864RMN before
                                                                          setting the start register to e1.' */
#define RESET_SIGNAL_BANK       (6)
#define RESET_SIGNAL_PIN        (2)

/*=== Macro definitions (File scope) =================================================================================*/

/*=== Type definitions (File scope) ==================================================================================*/
struct ethsw_net
{
  SPI_InfoHandle hSPIInfo;
};

/*=== Type definitions for events (File scope) =======================================================================*/

/*=== Function declarations (File scope) =============================================================================*/

/*=== Macro-function definitions (File scope) ========================================================================*/

/*=== Data definitions (File scope) ==================================================================================*/

/*=== Local Function definitions (File scope) ========================================================================*/

/*=== Global function definitions (Global system or Local sub-system scope) ==========================================*/
/**
 * Reset the switch and PHY
 */
void switch_reset(void)
{
  DEVICE_setPin(RESET_SIGNAL_BANK, RESET_SIGNAL_PIN, 0);
  UTIL_waitLoopAccurate(RESET_ASSERT_TIME_NS);
  DEVICE_setPin(RESET_SIGNAL_BANK, RESET_SIGNAL_PIN, 1);
  UTIL_waitLoopAccurate(RESET_DELAY_TIME_NS);
}

/**
 * Initialises the switch interface
 * @param[in]   port      The SPI port
 * @param[in]   channel   The SPI channel (CS)
 * @return                Pointer to the switch object
 */
struct ethsw_net* switch_open(T_SPI_PORT const port, unsigned const channel)
{
    static struct ethsw_net ks;
    static struct {
        SPI_Config config;
        SPI_InfoHandle handle;
    } spi = {
        {
            /* Configure the SPI:
             * Use industry standard mode 3 (note that our SPI peripheral phase value is inverted compared
             * to all other industry players) */
            .charLen          = 8,
            .phase            = 1,
            .polarity         = 0,
            .prescalar        = 30,
            .delayAfterCSdown = NS_DELAY_TO_LOOPCNT(12),  // ~14.6 ns > 10 ns (tSLCH SPIS_N Active Set-Up Time)
            .delayBeforeCSup  = NS_DELAY_TO_LOOPCNT(12),    // ~14.6 ns > 10 ns (tCHSH SPIS_N Active Hold Time)
            .delayAfterCSup   = NS_DELAY_TO_LOOPCNT(120),   // ~373 ns > 200 ns (tSHSL SPIS_N Deselect Time)
        },
        NULL
    };

    // Configure reset pin
    DEVICE_setPin(RESET_SIGNAL_BANK, RESET_SIGNAL_PIN, 1);
    DEVICE_setPinDirection(RESET_SIGNAL_BANK, RESET_SIGNAL_PIN, 0);

    ks.hSPIInfo = SPI_open(port,
                           channel,
                           SPI_ROLE_MASTER,
                           SPI_MODE_3PIN,
                           &spi.config);

    return &ks;
}

void ksz8864_writeData(struct ethsw_net*  ks,
                       unsigned const       startReg,
                       void const* const    data,
                       unsigned const       noBytes )
{
  struct
  {
    T_KSZ8864_TX_FRAME_HEAD head;
    UNS8                    data[noBytes];
  } KSZ8864_BYTE_ALIGN(txFrame);

  txFrame.head.command    = KSZ8864_SMI_WRITE_CMD;
  txFrame.head.registerNo = (UNS8) startReg;
  memcpy(txFrame.data, data, noBytes);
  SPI_enableCS(ks->hSPIInfo);
  SPI_sync(ks->hSPIInfo, sizeof(txFrame), &txFrame, 0, NULL);
  SPI_disableCS(ks->hSPIInfo);
}

void ksz8864_readData(struct ethsw_net*     ks,
                      unsigned const        startReg,
                      void* const           data,
                      unsigned const        noBytes )
{
  T_KSZ8864_TX_FRAME_HEAD txFrame =
  {
    KSZ8864_SMI_READ_CMD,
    (UNS8) startReg,
  };
  struct
  {
    T_KSZ8864_TX_FRAME_HEAD syncBytes;
    UNS8                    data[noBytes];
  } KSZ8864_BYTE_ALIGN(rxFrame);

  SPI_enableCS(ks->hSPIInfo);
  SPI_sync(ks->hSPIInfo, sizeof(txFrame), &txFrame, sizeof(rxFrame), &rxFrame);
  SPI_disableCS(ks->hSPIInfo);
  memcpy(data, rxFrame.data, noBytes);
}

/**
 * Writes a switch register
 * @param[in] ks      The switch device
 * @param[in] reg     Register number
 * @param[in] value   Value to write
 */
void ksz8864_writeRegister(struct ethsw_net* ks, unsigned const reg, unsigned const value)
{
  struct
  {
    T_KSZ8864_TX_FRAME_HEAD head;
    UNS8                    value;
  } KSZ8864_BYTE_ALIGN(txFrame) =
  {
    {
      KSZ8864_SMI_WRITE_CMD,
      (UNS8) reg,
    },
    (UNS8) value
  };
  SPI_sync(ks->hSPIInfo, sizeof(txFrame), &txFrame, 0, NULL);
}

/**
 * Reads a switch register
 * @param[in] ks    The switch device
 * @param[in] reg   Register number
 * @return          The register value
 */
unsigned ksz8864_readRegister(struct ethsw_net* ks, unsigned const reg)
{
  T_KSZ8864_TX_FRAME_HEAD txFrame =
  {
    KSZ8864_SMI_READ_CMD,
    (UNS8) reg,
  };
  struct
  {
    T_KSZ8864_TX_FRAME_HEAD syncBytes;
    UNS8                    value;
  } KSZ8864_BYTE_ALIGN(rxFrame);

  SPI_sync(ks->hSPIInfo, sizeof(txFrame), &txFrame, sizeof(rxFrame), &rxFrame);

  return (unsigned)rxFrame.value;
}

/**
 * Dumps all switch registers starting 0x00 and up to 0xFF
 * @param[in] ks    The switch device
 */
void switch_dumpRegs(struct ethsw_net* ks)
{
  unsigned regNo;

  for(regNo = 0; regNo < 0x100; regNo++)
  {
    if((regNo % 16) == 0)
    {
        DEBUG1_printString("\r\n"); DEBUG_printHexChar(regNo); DEBUG1_printString(":  ");
    }
    DEBUG1_printHexChar(ksz8864_readRegister(ks, regNo)); DEBUG1_printString(" ");
  }
  DEBUG1_printString("\r\n");
}

/* END OF FILE */

