/*
 * TI Booting and Flashing Utilities
 *
 * The main project file for the user boot loader
 *
 * Copyright (C) 2009 Texas Instruments Incorporated - http://www.ti.com/
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation version 2.
 *
 * This program is distributed "as is" WITHOUT ANY WARRANTY of any
 * kind, whether express or implied; without even the implied warranty
 * of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 */

/* --------------------------------------------------------------------------
  AUTHOR      : Daniel Allred
 --------------------------------------------------------------------------- */

// General type include
#include "tistdtypes.h"

// This module's header file
#include "ubl.h"

// Device specific CSL
#include "device.h"
#include "assert.h"
#include "led_defs.h"

#include "switch.h"
#include "atags.h"
#include "otags.h"
#include "uboot.h"
#include "string.h"
#include "rebootcmd.h"
#include "macaddrs.h"
#include "leds.h"
#include "verinfo.h"

// Misc. utility function include
#include "util.h"

// Project specific debug functionality
#include "debug.h"
#include "uartboot.h"

#include "spi.h"
#include "spi_mem.h"

#ifdef UBL_NOR
// NOR driver include
#include "nor.h"
#include "norboot.h"
#endif

#ifdef UBL_NAND
// NAND driver include
#include "nand.h"
#include "nandboot.h"
#endif

#ifdef UBL_SPI
// SPI driver include
#include "spi.h"
#endif

#include "memtest.h"

#define SPI_DEVICE          (1)

/************************************************************
* Explicit External Declarations                            *
************************************************************/


/************************************************************
* Local Macro Declarations                                  *
************************************************************/

/**
 * The unique id of the 8051 PHY
 */
#define PHY_UNIQUE_ID   (0x050022U)

/**
 * the unique HW ID of the Micrel 8864 switch
 */
#define KSZ_SWITCH_ID       (0x95U)

/**
 * The start address of the ARM9 vectior-table RAM.
 */
#define ARM_LOCAL_RAM_BASE  (0xFFFF0000U)

/**
 * The size of the internal RAM.
 */
#define ARM_LOCAL_RAM_SIZE  (8U * SIZE_1KB)

/**
 * The start address of the ARM9 core RAM.
 */
#define ARM_LOCAL_RAM_END   (ARM_LOCAL_RAM_BASE + ARM_LOCAL_RAM_SIZE)

/**
 * The start address of internal RAM.
 */
#define IRAM_START_ADR      (0x80000000U)

/**
 * The size of the internal RAM.
 */
#define IRAM_SIZE           (128U * SIZE_1KB)

/**
 * The start address of internal RAM.
 */
#define IRAM_END_ADR        (IRAM_START_ADR + IRAM_SIZE)

/**
 * The start address of the SD-RAM.
 */
#define DRAM_START_ADR      (0xC0000000U)

/**
 * The size of the SD-RAM.
 */
#define DRAM_SIZE           (128U * SIZE_1MB)

/**
 * The end address of the SD-RAM.
 */
#define DRAM_END_ADR        (DRAM_START_ADR + DRAM_SIZE)

/**
 * The start address of the NOR flash.
 */
#define ROM_START_ADR       (0x60000000U)

/**
 * The size of the NOR flash.
 */
#define ROM_SIZE            (64U * SIZE_1MB)

/**
 * The end address of the NOR flash.
 */
#define ROM_END_ADR         (ROM_START_ADR + ROM_SIZE)

/**
 * Whatever's left on the NOR flash.
 */
#define ROM_SIZE_REMAINING  (-1U)

/**
 * The base address (in RAM) which will be used to generate an A-tag list needed by Lnux.
 */
#define ATAGS_DRAM_ADR  (DRAM_START_ADR + 0x100U)

/**
 * In absence of an O-Tag, the kernel can be started with the following O-Tag
 */
#define DEFAULT_LINUX_CMD_ATAG      "root=/dev/ram0 rw rdinit=/linuxrc rwconsole=ttyS0,115200 davinci_wdt.heartbeat=1"
#define DEFAULT_LINUX_MAINT_CMD     "bootmode=fhwt interfaces=eth0=192.168.0.250,255.255.255.0,0.0.0.0"
#define DEFAULT_OTAGS_CMD_ADD       ""

#define LINUX_CMD_LINE_MAX          (1024)

/**
 * The Linux machine type
 */
#define MACH_TYPE_DAVINCI_AM18XX_OMN    4250

/************************************************************
* Local Typedef Declarations                                *
************************************************************/


/************************************************************
* Local Function Declarations                               *
************************************************************/

/************************************************************
* Local Variable Definitions                                *
************************************************************/
#define MTD_DEVICE_TYPE             "davinci-flash.0"
#define BOOTLOADER_START_ADR        (ROM_START_ADR)
#define BOOTLOADER_SIZE             (128 * SIZE_1KB)
#define MAINTENANCE_START_ADR       (BOOTLOADER_START_ADR + BOOTLOADER_SIZE)
#define MAINTENANCE_SIZE            (16 * SIZE_1MB)
#define APPLICATION_START_ADR       (MAINTENANCE_START_ADR + MAINTENANCE_SIZE)
#define APPLICATION_SIZE            (32 * SIZE_1MB)
#define OTAGS_START_ADR             (APPLICATION_START_ADR + APPLICATION_SIZE)
#define OTAGS_SIZE                  (128 * SIZE_1KB)
#define USERDATA_START_ADR          (OTAGS_START_ADR + OTAGS_SIZE)
#define USERDATA_SIZE               (8 * SIZE_1MB)
#define RAW_START_ADR               (USERDATA_START_ADR + USERDATA_SIZE)
#define FACT_BANK_PIN               6, 0
#define ETHSWRES_BANK_PIN           6, 2
#define IBC_XINT0_BANK_PIN          2, 2
#define IBC_XINT1_BANK_PIN          2, 4
#define IBC_SWITCH1_BANK_PIN        2, 15
#define IBC_SWITCH2_BANK_PIN        3, 12
#define IBC_SWITCH3_BANK_PIN        6, 6
#define IBC_SWITCH4_BANK_PIN        6, 5

/* Spinlock counter value per 500 ms. Determined empirically */
#define COUNTS_PER_500MS           3787878

#define VERSION_MAGIC               0x5652534E
#define VERSION_STR_SIZE            128

#define CMDLINE_MAC_PREFIX          " ethsw.am18xx_mac="

static volatile const struct {
    UNS32           magic;
    char const      buffer[VERSION_STR_SIZE];
} VERSION = { VERSION_MAGIC, AUTO_FULL_VERSION};

typedef enum {
    PARTITION_BOOTLOADER=0,
    PARTITION_MAINTENANCE,
    PARTITION_APPLICATION,
    PARTITION_OTAGS,
    PARTITION_USERDATA,
    PARTITION_RAW,
    PARTITION_MAX
} T_PARTITION_NO;

typedef enum {
    TYPE_BOOT_IMAGE=0,
    TYPE_UBOOT_IMAGE,
    TYPE_RAW_DATA,
    TYPE_UBIFS,
} T_TYPE;

typedef struct {
    char const* const   description;
    UNS32 const         start;
    UNS32 const         size;
    T_TYPE const        type;
} T_PARTITION;

typedef enum {
    BOOT_IMAGE_MAINT  = 0,
    BOOT_IMAGE_APPLIC,
    BOOT_IMAGE_MAX
} T_BOOT_IMAGE_TYPE;

typedef enum {
    BOOT_FLAG_NORAML        = (1<<0),
    BOOT_FLAG_FORCE_MAINT   = (1<<1),
} T_BOOT_MODE_FLAGS;

typedef struct {
    T_BOOT_IMAGE_TYPE const     image;
    T_PARTITION const* const    partition;
} T_BOOT_IMAGE;

static T_PARTITION const norPartitions[PARTITION_MAX] = {
    { "Bootloader",     BOOTLOADER_START_ADR,   BOOTLOADER_SIZE,    TYPE_BOOT_IMAGE,  },
    { "Maintenance",    MAINTENANCE_START_ADR,  MAINTENANCE_SIZE,   TYPE_UBOOT_IMAGE, },
    { "Application",    APPLICATION_START_ADR,  APPLICATION_SIZE,   TYPE_UBOOT_IMAGE, },
    { "O-Tags",         OTAGS_START_ADR,        OTAGS_SIZE,         TYPE_RAW_DATA,    },
    { "Userdata",       USERDATA_START_ADR,     USERDATA_SIZE,      TYPE_UBIFS,       },
    { "RawStorage",     RAW_START_ADR,          ROM_SIZE_REMAINING, TYPE_RAW_DATA,    },
};

static T_BOOT_IMAGE const bootImages[BOOT_IMAGE_MAX] = {
    { BOOT_IMAGE_MAINT,  norPartitions + PARTITION_MAINTENANCE },
    { BOOT_IMAGE_APPLIC, norPartitions + PARTITION_APPLICATION },
};

static T_LED const leds[LED_MAX] = {
    { "NS_GREEN",  2, 0, TRUE,  FALSE, },  /* NS_GREEN */
    { "NS_RED",    2, 1, TRUE,  FALSE, },  /* NS_RED */
    { "MS_RED",    2, 5, FALSE, FALSE, },  /* MS_RED */
    { "MS_GREEN",  2, 8, TRUE,  FALSE, },  /* MS_GREEN */
    { "PRPHL",     6, 4, TRUE,  FALSE, },  /* LED_PRPHL */
};

typedef enum {
    ERROR_BOOTLOADER_CRC = 0,
    ERROR_MAINT_CRC,
    ERROR_APPLIC_CRC,
    ERROR_OTAGS_CRC,
    ERROR_MEMORY,
    ERROR_FLASH_IBC,
    ERROR_UNDEF,

    ERROR_MAX,
} T_ERROR;

static const unsigned errorBlinks[] = {
    2,          /* ERROR_BOOTLOADER_CRC */
    3,          /* ERROR_MAINT_CRC      */
    4,          /* ERROR_APPLIC_CRC     */
    5,          /* ERROR_OTAGS_CRC      */
    6,          /* ERROR_MEMORY         */
    7,          /* ERROR_FLASH_IBC      */
    1,          /* ERROR_UNDEF          */
};

/************************************************************
* Global Variable Definitions                               *
************************************************************/

Uint32 gEntryPoint;

extern DEVICE_OperatingPoint gDeviceOpPoint;
/************************************************************
* Global Function Definitions                               *
************************************************************/

/**
 * Generates a complete A-tag list.
 * @param[out] baseTag          Destination memory for the A-tags
 * @param[in]  linuxCmdLine     The Linux cmdline string to add to the A-Tags. If @p NULL, the cmdline A-tag is skipped.
 * @param[in]  linuxCmdLineAdd  Additional string data for the Linux cmdline. If @p NULL, will be skipped.
 * @return     Pointer to the A-Tags list
 */
static struct tag* generateAtagList(struct tag*                 baseTag,
                                    char const* const           linuxCmdLine,
                                    char const* const* const    linuxCmdLineAdd )
{
    struct tag const* kernelCmdTag;
    struct tag* nextTag = atags_addStartTag(    baseTag);
    nextTag             = atags_addMemoryTag(   nextTag, DRAM_START_ADR, (DRAM_END_ADR - DRAM_START_ADR));
    kernelCmdTag        = nextTag;

    if (linuxCmdLine != NULL) {
        nextTag = atags_addCommandlineTag( nextTag,
                                           (char const* const) linuxCmdLine,
                                           (char const* const* const) linuxCmdLineAdd );

    } else {
        nextTag = atags_addCommandlineTag( nextTag, DEFAULT_LINUX_CMD_ATAG, NULL);
        DEBUG1_printString("warning: No Kernel cmdline provided: Using defaults.\n\r");
    }
    DEBUG1_printString("A-TAG: Kernel cmdline = \"");
        DEBUG1_printString(kernelCmdTag->u.cmdline.cmdline);
        DEBUG1_printString("\"\n\r");

    nextTag = atags_addEndTag(nextTag);

    return baseTag;
}

/**
 * Appends the MTD partition information to a string
 * @param[out] dest   The destination string
 * @param[in]  n      The size remaining in the destination string
 */
char const* addPartitions(char* dest, size_t n)
{
    unsigned idx;
    char*    result = dest;

    for(idx = 0U; idx < ARRAY_SIZE(norPartitions); idx++) {
        T_PARTITION const* const partition = norPartitions + idx;
        if(partition->type == TYPE_UBIFS) {
            (void)strncpy(dest, " ubi.mtd=", n - ((size_t)dest - (size_t)result));
            dest = result + strlen(result);
            u2a(dest, idx, 10U, 0, n - ((size_t)dest - (size_t)result));
            dest = result + strlen(result);
        }
    }

    /* Add the partition token */
    (void)strncpy(dest, " mtdparts=" MTD_DEVICE_TYPE ":", n - ((size_t)dest - (size_t)result));
    dest = result + strlen(result);

    /* List all the known partitions */
    for(idx = 0U; idx < ARRAY_SIZE(norPartitions); idx++) {
        T_PARTITION const* const partition = norPartitions + idx;

        /* Close off the previous partition description and start the next partition */
        if(idx != 0U) {
            (void)strncpy(dest, "),", n - ((size_t)dest - (size_t)result));
            dest = result + strlen(result);
        }

        /* Append human-readable size and start the partition description:
         * '#k' for the size in kilobytes
         * '#M' for the size in megabytes
         * '-' for whatever is left of the NOR flash */
        if(partition->size == ROM_SIZE_REMAINING) {
            (void)strncpy(dest, "-(", n - ((size_t)dest - (size_t)result));
            dest = result + strlen(result);
        } else if((partition->size % SIZE_1MB) == 0U) {
            u2a(dest, partition->size / SIZE_1MB, 10U, 0, n - ((size_t)dest - (size_t)result));
            dest = result + strlen(result);
            (void)strncpy(dest, "M(", n - ((size_t)dest - (size_t)result));
            dest = result + strlen(result);
        } else if ((partition->size % SIZE_1KB) == 0U) {
            u2a(dest, partition->size / SIZE_1KB, 10U, 0, n - ((size_t)dest - (size_t)result));
            dest = result + strlen(result);
            (void)strncpy(dest, "k(", n - ((size_t)dest - (size_t)result));
            dest = result + strlen(result);
        } else {
            u2a(dest, partition->size, 10U, 0, n - ((size_t)dest - (size_t)result));
            dest = result + strlen(result);
            (void)strncpy(dest, "(", n - ((size_t)dest - (size_t)result));
            dest = result + strlen(result);
        }

        /* Append the partition description */
        (void)strncpy(dest, partition->description, n - ((size_t)dest - (size_t)result));
        dest = result + strlen(result);
    }
    /* Close off the last partition description */
    (void)strncpy(dest, ")", n - ((size_t)dest - (size_t)result));
    dest = result + strlen(result);

    return result;
}

/**
 *
 */
static char* macaddrs_cmdline(T_OTAG_VALUES const* const otagsList, char* s, size_t n)
{
    char mac[128];
    unsigned idx;
    unsigned offs;

    (void)strncpy(mac, CMDLINE_MAC_PREFIX, sizeof(CMDLINE_MAC_PREFIX));
    offs = sizeof(CMDLINE_MAC_PREFIX)-1;

    for(idx = 0; idx < otagsList->mac.count; idx++) {
        T_OTAG_MAC const* const             address = otagsList->mac.address[idx];
        if(address && idx == 0U) {
            (void)u2a(&mac[offs+0], ((address->macAddrLo >>  0U) & 0xFF), 16, 2, 3);
            (void)u2a(&mac[offs+3], ((address->macAddrLo >>  8U) & 0xFF), 16, 2, 3);
            (void)u2a(&mac[offs+6], ((address->macAddrLo >>  16U) & 0xFF), 16, 2, 3);
            (void)u2a(&mac[offs+9], ((address->macAddrLo >>  24U) & 0xFF), 16, 2, 3);
            (void)u2a(&mac[offs+12], ((address->macAddrHi >>  0U) & 0xFF), 16, 2, 3);
            (void)u2a(&mac[offs+15], ((address->macAddrHi >>  8U) & 0xFF), 16, 2, 3);
            mac[offs+2] = mac[offs+5] = mac[offs+8] = mac[offs+11] = mac[offs+14] = ':';
        }
    }

    return strncpy(s, mac, n);
}

/**
 * Constructs the Linux cmdline
 * @param[in] otagsList     Parsed O-tag data to take into account
 * @param[in] cmdLineAdd    Additional string data to add to the command-line
 * @return    The Linux cmdline to pass to the kernel
 */
char const* generateCmdLine(T_OTAG_VALUES const* const otagsList, char const* const cmdLineAdd)
{
    static char         result[LINUX_CMD_LINE_MAX] = "\0";
    char*               dest                       = result;
    static char const   defaultCmdBase[]           = DEFAULT_LINUX_CMD_ATAG;
    char const* const   cmdBase                    = (otagsList->linuxCmdOverride.string != NULL)
                                                     ? otagsList->linuxCmdOverride.string
                                                     : defaultCmdBase;

    /* Add the base string */
    (void)strncpy(dest, cmdBase, sizeof(result) - ((size_t)dest - (size_t)result));
    dest = result + strlen(result);

    /* Append anything requested by the O-tags */
    if(otagsList->linuxCmdAdd.string != NULL) {
        (void)strncpy(dest, " ", sizeof(result) - ((size_t)dest - (size_t)result));
        dest = result + strlen(result);
        (void)strncpy(dest, otagsList->linuxCmdAdd.string, sizeof(result) - ((size_t)dest - (size_t)result));
        dest = result + strlen(result);
    }

    /* Append the NOR flash partition information */
    (void)addPartitions(dest, sizeof(result) - ((size_t)dest - (size_t)result));
    dest = result + strlen(result);

    /* Append the MAC address */
    (void)macaddrs_cmdline(otagsList, dest, sizeof(result) - ((size_t)dest - (size_t)result));
    dest = result + strlen(result);

    /* Append any additional cmdline data */
    if(cmdLineAdd != NULL) {
        (void)strncpy(dest, " ", sizeof(result) - ((size_t)dest - (size_t)result));
        dest = result + strlen(result);
        (void)strncpy(dest, cmdLineAdd, strlen(cmdLineAdd));
        dest = result + strlen(result);
    }
    return result;
}

static T_BOOT_IMAGE const* getBootImage(T_REBOOTCMD const* const rebootCmd, T_BOOT_MODE_FLAGS const flags)
{
    T_BOOT_IMAGE const* retval;

    if ((flags & BOOT_FLAG_FORCE_MAINT)) {
        retval = bootImages + (unsigned) BOOT_IMAGE_MAINT;
    } else if(DEVICE_getPin(FACT_BANK_PIN) == FALSE) {
        DEBUG1_printString("Factory pin CP2/GP6[0] asserted!\r\n");
        retval = bootImages + (unsigned) BOOT_IMAGE_MAINT;
    } else if( (rebootCmd != NULL) && (((unsigned) rebootCmd->bootImageNo) < ARRAY_SIZE(bootImages)) ) {
        retval = bootImages + rebootCmd->bootImageNo;
    } else {
        retval = bootImages + (unsigned) BOOT_IMAGE_APPLIC;
    }
    DEBUG1_printString("\r\nBooting '");
    DEBUG1_printString(retval->partition->description);
    DEBUG1_printString("'");
    return retval;
}

static void sleep(unsigned half_seconds)
{
    while (half_seconds--) {
        volatile unsigned long count = COUNTS_PER_500MS;
        while (count != 0)
            count--;
    }
}

static void indicateError(T_ERROR const error)
{
    static T_LED const pattern1[LED_MAX] = {
        { "NS_GREEN",  2, 0, TRUE,  FALSE, },  /* NS_GREEN */
        { "NS_RED",    2, 1, TRUE,  FALSE, },  /* NS_RED */
        { "MS_RED",    2, 5, FALSE, FALSE, },  /* MS_RED */
        { "MS_GREEN",  2, 8, TRUE,  FALSE, },  /* MS_GREEN */
        { "PRPHL",     6, 4, TRUE,  FALSE, },  /* LED_PRPHL */
    };

    static T_LED const pattern2[LED_MAX] = {
        { "NS_GREEN",  2, 0, TRUE,  FALSE, },  /* NS_GREEN */
        { "NS_RED",    2, 1, TRUE,  TRUE,  },  /* NS_RED */
        { "MS_RED",    2, 5, FALSE, FALSE, },  /* MS_RED */
        { "MS_GREEN",  2, 8, TRUE,  FALSE, },  /* MS_GREEN */
        { "PRPHL",     6, 4, TRUE,  FALSE, },  /* LED_PRPHL */
    };

    unsigned blinkCount = (error >= ERROR_MAX ? ERROR_MEMORY : errorBlinks[(unsigned)error]);
    while (blinkCount--) {
        leds_init(pattern2, ARRAY_SIZE(pattern2));
        sleep(1);
        leds_init(pattern1, ARRAY_SIZE(pattern1));
        sleep(1);
    }
    sleep(4);
}

static void stopWithError(T_ERROR const error)
{
    while (1) {
        indicateError(error);
    }
}


static T_BOOT_IMAGE const* enterImage(struct ethsw_net* ethsw, T_BOOT_MODE_FLAGS const flag)
{

    T_REBOOTCMD const* const    rebootCmd   = rebootcmd_parse((void*)ARM_LOCAL_RAM_END);
    T_BOOT_IMAGE const* const   bootImage   = getBootImage(rebootCmd, flag);
    T_OTAG_VALUES const* const  otagsList   = otags_parseList((void*)OTAGS_START_ADR, (unsigned) bootImage->image);
    BOOL const                  checkOtags  = (bootImage != NULL && bootImage->image != BOOT_IMAGE_MAINT);
    char const*                 cmdLineAdd  = (rebootCmd != NULL)
                                              ? rebootCmd->cmdlineAdd
                                              : NULL;
    //DEFAULT_LINUX_MAINT_CMD
    if (flag & BOOT_FLAG_FORCE_MAINT)
         cmdLineAdd  = DEFAULT_LINUX_MAINT_CMD;

    if ((checkOtags == TRUE) && ((otagsList == NULL) || (otagsList->errors))) {
        return NULL;
    } else {
        BOOL const                  macAddrsSet = macaddrs_set(ethsw, otagsList);
        char const* const           cmdLine     = generateCmdLine(otagsList, cmdLineAdd);
        struct tag const* const     aTagList    = generateAtagList((void*)ATAGS_DRAM_ADR, cmdLine, NULL);
        T_UBOOT_RESULT const        result      = uboot_bootKernelImage(MACH_TYPE_DAVINCI_AM18XX_OMN,
                                                                        (void const* const) bootImage->partition->start,
                                                                        (unsigned int) aTagList,
                                                                        DRAM_START_ADR, DRAM_SIZE);

        if (macAddrsSet == FALSE) {
            DEBUG1_printString("\r\nerror: failed to configure EMAC mac address from O-Tags\n");
        }

        switch(result) {
            case UBOOT_HEADER_NOT_FOUND: {
                DEBUG1_printString("\r\nerror: uboot image not found");
                break;
            }
            case UBOOT_HEADER_INVALID: {
                DEBUG1_printString("\r\nerror: uboot header CRC fail");
                break;
            }
            case UBOOT_HEADER_TYPE: {
                DEBUG1_printString("\r\nerror: uboot unsupported header type detected");
                break;
            }
            case UBOOT_DEST_INVALID: {
                DEBUG1_printString("\r\nerror: uboot destination address not (or partially not) in RAM");
                break;
            }
            case UBOOT_PAYLOAD_INVALID: {
                DEBUG1_printString("\r\nerror: uboot payload CRC fail");
                break;
            }
            case UBOOT_BOOT_FAILED: {
                DEBUG1_printString("\r\nerror: uboot kernel main() returned");
                break;
            }
            default: {
                DEBUG1_printString("\r\nerror: uboot unexpected return value");
                break;
            }
        }
        DEBUG1_printString(" @ ");
        DEBUG1_printHexInt(bootImage->partition->start);
        DEBUG1_printString("\r\n");

        return bootImage;
    }
    return NULL;
}

// Main entry point
void main(void)
{

    // First thing, put DSP in reset
    DEVICE_disable_DSP();
    DEVICE_init(DEVICE_BOOTMODE_EMU_DEBUG);

    // Set IBC XINTs to input ASAP
    DEVICE_setPinDirection(IBC_XINT0_BANK_PIN, TRUE);
    DEVICE_setPinDirection(IBC_XINT1_BANK_PIN, TRUE);

    // Issue #7492, need to keep the switch in reset as soon as possible
#ifndef BOOT_ETH_SWITCH    
    struct ethsw_net* ethsw = NULL;
    DEVICE_setPin(ETHSWRES_BANK_PIN, 0);
    DEVICE_setPinDirection(ETHSWRES_BANK_PIN, FALSE);
    DEVICE_setPin(ETHSWRES_BANK_PIN, 0);
#endif

    // Initialise the UART
#if defined(DEVICE_UART0_FOR_DEBUG)
    DEVICE_UARTInit(0);
#elif defined(DEVICE_UART1_FOR_DEBUG)
    DEVICE_UARTInit(1);
#elif defined(DEVICE_UART2_FOR_DEBUG)
    DEVICE_UARTInit(2);
#endif

    // Some feedback is in order
    DEBUG1_printString("\r\nO-Boot version [");
    DEBUG1_printString((const char * const)VERSION.buffer);
    DEBUG1_printString("]\r\n");
    DEVICE_setPinDirection(FACT_BANK_PIN, TRUE);

    leds_init(leds, ARRAY_SIZE(leds));

#ifdef BOOT_STRESS_MEM
    mem_stress(DRAM_START_ADR, DRAM_SIZE);
#else
    BOOL const memTestResult = mem_test_short(DRAM_START_ADR, DRAM_SIZE);
    if (memTestResult != TRUE) {
        stopWithError(ERROR_MEMORY);
    }
#endif

#ifdef BOOT_ETH_SWITCH    
    struct ethsw_net* ethsw = switch_open(SPI_DEVICE, SPI_CS_NUM);

    switch_reset();

    if (ethsw == NULL) {
        DEBUG1_printString("error: Opening the Ethernet switch\r\n\tCannot continue!");
    } else {
        int const probeResult = switch_probe(ethsw);

        if (probeResult < 0) {
            DEBUG1_printString("error: Failed probing the Ethernet switch");
            DEBUG1_printString("\r\n\tCannot continue!");
        } else {
            DEBUG1_printString("Starting Ethernet switch...");
            switch_start(ethsw);
        }
        switch_dumpRegs(ethsw);
    }
#endif

    // Set IBC SWITCHs to input ASAP
    DEVICE_setPinDirection(IBC_SWITCH4_BANK_PIN, TRUE);
    DEVICE_setPinDirection(IBC_SWITCH3_BANK_PIN, TRUE);
    DEVICE_setPinDirection(IBC_SWITCH2_BANK_PIN, TRUE);
    DEVICE_setPinDirection(IBC_SWITCH1_BANK_PIN, TRUE);

	if (DEVICE_getPin(IBC_SWITCH4_BANK_PIN))
	    DEBUG1_printString("IBC_SWITCH4_BANK_PIN:on\r\n");
	else
	    DEBUG1_printString("IBC_SWITCH4_BANK_PIN:off\r\n");
#if 0
	T_BOOT_IMAGE const* image = NULL;
	if (DEVICE_getPin(IBC_SWITCH1_BANK_PIN) == TRUE)
	    image = enterImage(ethsw, BOOT_FLAG_FORCE_MAINT);
	else
	    image = enterImage(ethsw, BOOT_FLAG_NORAML);
#else
	char r = 0;
	DEBUG_readChar(&r);
	
	T_BOOT_IMAGE const* image;
	if (DEVICE_getPin(IBC_SWITCH4_BANK_PIN) || r == 'm') {
		image = enterImage(ethsw, BOOT_FLAG_FORCE_MAINT);
	}
	else {
    	image = enterImage(ethsw, BOOT_FLAG_NORAML);
	}
#endif

    if (image == NULL) {
        DEBUG1_printString("error: OTAG CRC, attempting boot maintenance image\r\n");
        indicateError(ERROR_OTAGS_CRC);
        image = enterImage(ethsw, BOOT_FLAG_FORCE_MAINT);
        DEBUG1_printString("error: maintenance image CRC\r\n\tCannot continue!");
        stopWithError(ERROR_MAINT_CRC);
    } else if (image->image == BOOT_IMAGE_MAINT) {
        DEBUG1_printString("error: maintenance image CRC\r\n\tCannot continue!");
        stopWithError(ERROR_MAINT_CRC);
    } else if (image->image == BOOT_IMAGE_APPLIC) {
        DEBUG1_printString("error: applic CRC, attempting boot maintenance image\r\n");
        indicateError(ERROR_APPLIC_CRC);
        image = enterImage(ethsw, BOOT_FLAG_FORCE_MAINT);
        DEBUG1_printString("error: maintenance image CRC\r\n\tCannot continue!");
        stopWithError(ERROR_MAINT_CRC);
    } else {
        DEBUG1_printString("error: \r\n\tCannot continue!");
        stopWithError(ERROR_UNDEF);
    }

    stopWithError(ERROR_UNDEF);
}

/************************************************************
* End file                                                  *
************************************************************/
