/*======================================================================================================================
  (c) Copyright OMRON Europe BV 2011
  All Right Reserved
  OMRON Development & Engineering
  -----------------------------------------------------------------------------------------------------------------*//**
  @file

  Implements boot support for U-Boot images

*//*--------------------------------------------------------------------------------------------------------------------
      Unique identifier generated by svn:
      $Header$

======================================================================================================================*/

/*=== Include files ==================================================================================================*/
#include "debug.h"
#include "crc32.h"
#include "string.h"
#include "uboot_header.h"
#include "uboot.h"

/*=== Constant definitions (File scope) ==============================================================================*/

/*=== Macro definitions (File scope) =================================================================================*/

/*=== Type definitions (File scope) ==================================================================================*/
typedef void (*T_FUNC_KERNEL_MAIN)(int zero, int arch, unsigned int params);

/*=== Type definitions for events (File scope) =======================================================================*/

/*=== Function declarations (File scope) =============================================================================*/

/*=== Macro-function definitions (File scope) ========================================================================*/

/*=== Data definitions (File scope) ==================================================================================*/

/*=== Local Function definitions (File scope) ========================================================================*/

/**
 * Checks the CRC-32 of an U-boot header.
 * @pre -
 * @param header A pointer to the U-boot header;
 * @return @p TRUE if it is a valid header.
 */
static T_UBOOT_IMG_HEADER const* findHeader(void const* const image)
{
    T_UBOOT_IMG_HEADER const* const header = (T_UBOOT_IMG_HEADER const* const) image;
    return ((READ_UNS32_BIG_ENDIAN(header->magic) == IH_MAGIC) ? header : NULL);
}

/**
 * Checks the CRC-32 of an U-boot header.
 * @pre -
 * @param header A pointer to the U-boot header;
 * @return @p TRUE if it is a valid header.
 */
static T_UBOOT_IMG_HEADER const* checkHeader(T_UBOOT_IMG_HEADER const* const header)
{
  UNS32 const crc32 = crc32_calculateSubstituted(0U, header, sizeof(T_UBOOT_IMG_HEADER),
                                                 &header->hcrc, 0U, sizeof(header->hcrc));
  return ((crc32 == READ_UNS32_BIG_ENDIAN(header->hcrc)) ? header : NULL);
}

/**
 * Parses and boots an ARM Linux U-Boot image at @p image
 * @param[in] machineType   Linux ARM machine-type
 * @param[in] image         Location of the U-Boot image in addressable memory
 * @param[in] parms         Parameters passed to the Linux kernel entry function
 * @param[in] addressRAM    The platform start RAM address used to validate the U-Boot copy destination
 * @param[in] sizeRAM       The platform RAM size used to validate the U-Boot copy destination
 * @return
 */
T_UBOOT_RESULT uboot_bootKernelImage(int const machineType,
                                     void const* const image,
                                     unsigned int const parms,
                                     unsigned int const addressRAM,
                                     unsigned int const sizeRAM )
{
    T_UBOOT_IMG_HEADER const* const header = findHeader(image);

    if(header == NULL) {
        DEBUG1_printString("uboot: error: no valid header @ ");
        DEBUG1_printHexInt((Uint32 const) image);
        DEBUG1_printString("\r\n");
        return UBOOT_HEADER_NOT_FOUND;
    } else if(header != checkHeader(header)) {
        DEBUG1_printString("uboot: error: header CRC fail @ ");
        DEBUG1_printHexInt((Uint32 const) image);
        DEBUG1_printString("\r\n");
        return UBOOT_HEADER_INVALID;
    } else if((header->arch  != IH_ARCH_ARM)    ||
              (header->type  != IH_TYPE_KERNEL) ||
              (header->os    != IH_OS_LINUX)       ) {
        DEBUG1_printString("uboot: error: unsupported header type @ ");
        DEBUG1_printHexInt((Uint32 const) image);
        DEBUG1_printString("\r\n");
        return UBOOT_HEADER_TYPE;
    } else {
        void const* const           srcData     = header + 1;
        UNS32 const                 srcSize     =                       READ_UNS32_BIG_ENDIAN(header->size);
        void* const                 destData    = (void* const)         READ_UNS32_BIG_ENDIAN(header->load);
        UNS32 const                 destCrc32   =                       READ_UNS32_BIG_ENDIAN(header->dcrc);
        T_FUNC_KERNEL_MAIN const    entryPoint  = (T_FUNC_KERNEL_MAIN)  READ_UNS32_BIG_ENDIAN(header->ep);

        DEBUG1_printString("uboot: found '");
        DEBUG1_printString((char const* const) header->name);
        DEBUG1_printString("'\r\n");

        if(destData == srcData) {
            DEBUG1_printString("uboot: XIP image (Skip relocation & CRC)\r\n");
        } else if( ( ((unsigned int) destData) < addressRAM ) ||
                   ( (((unsigned int) destData) + srcSize) > (addressRAM + sizeRAM) ) ) {
            DEBUG1_printString("\r\nuboot: error: Destination (");
            DEBUG1_printHexInt((Uint32 const) destData);
            DEBUG1_printString(") doesn't fit in RAM\r\n");
            return UBOOT_DEST_INVALID;
        } else {
            UNS32 crc32;

            DEBUG1_printString("uboot: Relocate image: ");
            DEBUG1_printHexInt((Uint32 const) srcData);
            DEBUG1_printString(" -> ");
            DEBUG1_printHexInt((Uint32 const) destData);

            DEBUG1_printString("\r\n{");
            crc32 = crc32_memcpy(0U, destData, srcData, srcSize);
            DEBUG1_printString("\r\n}");

            DEBUG1_printString("\r\n");
            if(crc32 != destCrc32) {
                DEBUG1_printString("uboot: error: payload CRC fail @ ");
                DEBUG1_printHexInt((Uint32 const) destData  );
                DEBUG1_printString("\r\n");
                return UBOOT_PAYLOAD_INVALID;
            }
        }
        entryPoint(0, machineType, parms);
    }
    return UBOOT_BOOT_FAILED;
}

/* END OF FILE */
