/*======================================================================================================================
  (c) Copyright OMRON Europe BV 2011
  All Right Reserved
  OMRON Development & Engineering
  -----------------------------------------------------------------------------------------------------------------*//**
  @file
 
  Linux device driver for the Micrel KSZ8864 switch
 
  It is specifically designed by and for Omron to retrieve the link status (up/down, speed, etc). It is not a full
  switch driver for network access.
  @pre -
*//*--------------------------------------------------------------------------------------------------------------------
      Unique identifier generated by svn:
      $Header$

======================================================================================================================*/
/*=== Include files ==================================================================================================*/
#include <linux/module.h>
#include <linux/moduleparam.h>
#include <linux/kernel.h>
#include <linux/interrupt.h>
#include <linux/irq.h>
#include <linux/sched.h>
#include <linux/spi/spi.h>

#include <asm/uaccess.h>
#include <asm/errno.h>
#include <linux/types.h>
#include <linux/fs.h>
#include <linux/init.h>
#include <linux/module.h>
#include <linux/cdev.h>
#include <linux/gpio.h>
#include <linux/timer.h>
#include <linux/netdevice.h>		/* call_netdevice_notifiers */
#include <linux/rtnetlink.h>		/* rtnl_lock, rtnl_unlock */
#include <linux/phy.h>
#include <net/net_namespace.h>
#include <linux/proc_fs.h>
#include <linux/skbuff.h>
#include <linux/if_ether.h>			/* ETH_ZLEN */
#include <linux/seq_file.h>

#include <asm/gpio.h>

#include "ethsw.h"
#include "ksz8864.h"
#include "ksz8051.h"

/*=== Constant definitions (File scope) ==============================================================================*/
#define ETHSW_TRACE_ENABLE
//#define ETHSW_TRACE_DATA
//#define ETHSW_TRACE_TAG
#define PRIx64 "llx"

/*=== Macro definitions (File scope) =================================================================================*/
/**
 * Different macro's used to report driver messages to the user.
 */
#define KSERR(fmt, args...) printk(KERN_ALERT "ethsw [%s,%d]: " fmt, __FUNCTION__, __LINE__, ##args)
#define KSMSG(fmt, args...) printk(KERN_INFO  "ethsw [%s,%d]: " fmt, __FUNCTION__, __LINE__, ##args)

#ifdef ETHSW_TRACE_ENABLE
#define KSDBG(fmt, args...) printk(KERN_DEBUG "ethsw [%s,%d]: " fmt, __FUNCTION__, __LINE__, ##args)

#define KSDBGHEX(prefix,data,len)										\
	{													\
		unsigned char buff[16*3+1];									\
		unsigned char* p = (unsigned char*)(data);							\
		unsigned char* pend = &p[(len)];								\
		while (p != pend) {										\
			unsigned col, offs = 0;									\
			for (col = 0; (p != pend) && (col < 16); col++)	{					\
				offs += snprintf(buff + offs, sizeof(buff) - offs - 1, "%02X ", *p);		\
				p++;										\
			}											\
			buff[offs] = 0;										\
			printk(KERN_DEBUG "ethsw [%s,%d]: %s%s\n", __FUNCTION__, __LINE__, prefix, buff);	\
		}												\
	}

#ifdef ETHSW_TRACE_DATA
#define KSDBG_DATA(fmt, args...)       KSDBG(fmt, ##args)
#define KSDBGHEX_DATA(prefix,data,len) KSDBGHEX(prefix,data,len)
#else
#define KSDBG_DATA(fmt, args...)
#define KSDBGHEX_DATA(prefix,data,len)
#endif

#ifdef ETHSW_TRACE_TAG
#define KSDBG_TAG(fmt, args...)       KSDBG(fmt, ##args)
#else
#define KSDBG_TAG(fmt, args...)
#endif

#else
#define KSDBG(fmt, args...)
#define KSDBGHEX(prefix,data,len)
#endif

#define BUG_ON_NOT_LOCKED(lock) if(!mutex_is_locked(&lock))	{           \
	KSERR("Resource should be locked");                                 \
	BUG();                                                              \
}

#define LOCK_RESOURCE(lock) {                                           \
	mutex_lock(&lock);                                                  \
}

#define RELEASE_RESOURCE(lock) {                                        \
	mutex_unlock(&lock);                                                \
}

#define MIIDATA(phyAddr, regNo)  (void*)(((phyAddr) << 16) | ((regNo) & 0xFFFF))

#define SMAC_ENTRY_IDX_PORT1  (1)
#define SMAC_ENTRY_IDX_PORT2  (2)
#define SMAC_ENTRY_IDX_PORT3  (3)

#define TAILTAG_SIZE  (1)

#define NUM_PORTS_REPORTED       (ETHSW_EXT_PORT_MAX)
#define PORT_STAT_SIZE           (2)
#define PORT_DELIMETER_SIZE      (1)
#define CHARS_PER_PORT           (PORT_STAT_SIZE + PORT_DELIMETER_SIZE)
#define MAX_CHARS_AT_END         (1) /* New-line */
#define DATA_SIZE                ((PORT_STAT_SIZE * NUM_PORTS_REPORTED) + (PORT_DELIMETER_SIZE * (NUM_PORTS_REPORTED - 1)) + MAX_CHARS_AT_END)

/*=== Type definitions (File scope) ==================================================================================*/
/**
 * Receiver control settings.
 * Representation of the settings needs to control the receive filtering
 * such as the multicast hash-filter and the receive register settings. This
 * is used to make the job of working out if the receive settings change and
 * then issuing the new settings to the worker that will send the necessary
 * commands.
 */
struct ethsw_rxctrl {
	u16 mchash[4]; /**< Multicast hash-table data */
	u16 rxcr1; /**< KS_RXCR1 register setting */
	u16 rxcr2; /**< KS_RXCR2 register setting */
};

/**
 * Tx header.
 * A dual representation of the tx header data to allow
 * access to individual bytes, and to allow 16bit accesses
 * with 16bit alignment.
 */
union ethsw_tx_hdr {
	u8 txb[6]; /**< The header as bytes */
	__le16 txw[3]; /**< The header as 16bit, little-endian words */
};

struct link_status {
	unsigned int link;
	unsigned int speed;
	unsigned int duplex;
};

typedef enum {
	ETHSW_EXT_PORT_1 = 0,
	ETHSW_EXT_PORT_2,
	ETHSW_EXT_PORT_3,
	ETHSW_EXT_PORT_MAX
} ext_port;

/**
 * Main structure holding the different variables for communicating with and using
 * the KSZ8864 and attached PHY.
 * We align the buffers we may use for rx/tx to ensure that if the SPI driver
 * wants to DMA map them, it will not have any problems with data the driver
 * modifies.
 */
struct ethsw_net {

	/* Interrupt work-queue related */
	unsigned int            ethsw_wq_count;
	struct work_struct      ethsw_isr_wq;
	unsigned int            phy_wq_count;
	struct work_struct      phy_isr_wq;

	/* File-access operations related*/
	int                     read_flag;      /**< Port link status file operations wait queue flag */
	wait_queue_head_t       read_queue;     /**< Port link status file operations wait queue */

	/* ksz8864-related resources and lock */
	struct mutex            ksz8864_lock;	/**< Lock to ensure that the switch transactions are not interrupted */
	struct spi_device*      spidev;      	/**< The spi device we're bound to. */
	u8*                     rxd;         	/**< Space for receiving SPI data, in DMA-able space. */
	u8*                     txd;         	/**< Space for transmitting SPI data, in DMA-able space. */
	struct spi_message      spi_msg;     	/**< pre-setup SPI transfer with one message, @spi_xfer */
	struct spi_transfer     spi_xfer;

	/* mii-related resources and lock */
	struct mutex            mii_lock;       /**< Lock to ensure that MII access not done concurrently */
	struct mii_bus*         mii_bus;        /**< The MII bus attached to the ksz8864 and PHY MDC/MDIO */
	struct phy_device*      phydev;         /**< Port 3 PHY) */

	/* Device-related resources and lock */
	struct mutex            dev_lock;       /**< Lock to ensure that this struct is not concurrently accessed */
	struct net_device*      net_dev;        /**< Net device to which switch is attached */

	T_KSZ8864_STATIC_MAC_ENTRY smac_entries[KSZ8864_SMAC_TABLE_SIZE];  /**< MAC address to static table mapping */
	struct link_status      link_statuses[ETHSW_EXT_PORT_MAX];         /**< Link status tracking */
	unsigned int            prev_link_statuses[ETHSW_EXT_PORT_MAX];    /**< Link status change tracking */
	
	/* procfs related */
	struct proc_dir_entry*  proc_dir;
	struct proc_dir_entry*  regs_dir;
	struct proc_dir_entry*  mii_dir;
	struct proc_dir_entry*  ctrl_dir;
	struct proc_dir_entry*  mib_dir;
	struct proc_dir_entry*  sroute_dir;
	struct proc_dir_entry*  droute_dir;
	char*                   reset_result;      /**< Reset result message */
	char*                   dmac_flush_result; /**< Dynamic MAC table flush result message */
	bool                    tail_tag_enabled;
};
struct ks_proc_entry {
	char*            name;
	void*            data;
	struct file_operations op;
};

/*=== Type definitions for events (File scope) =======================================================================*/

/*=== Function declarations (File scope) =============================================================================*/
static int /*__devinit */ethsw_probe(struct spi_device *spi);
static int /*__devexit */ethsw_remove(struct spi_device *spi);

static int user_ethsw_open(struct inode *, struct file *fp);
static ssize_t user_ethsw_read(struct file *fp, char __user *buff, size_t count, loff_t *offp);

static int ks_proc_reg_read(struct seq_file *m, void *v);
static ssize_t ks_proc_reg_write(struct file *, const char __user *, size_t, loff_t *);
static int ks_proc_reg_open(struct inode *inode, struct file *file);
static int ks_proc_mii_read(struct seq_file *m, void *v);
static ssize_t ks_proc_mii_write(struct file *, const char __user *, size_t, loff_t *);
static int ks_proc_mii_open(struct inode *inode, struct file *file);
static int ks_proc_ctl_reset_read(struct seq_file *m, void *v);
static ssize_t ks_proc_ctl_reset_write(struct file *, const char __user *, size_t, loff_t *);
static int ks_proc_ctl_reset_open(struct inode *inode, struct file *file);
static int ks_proc_ctl_fldmac_read(struct seq_file *m, void *v);
static ssize_t ks_proc_ctl_fldmac_write(struct file *, const char __user *, size_t, loff_t *);
static int ks_proc_ctl_fldmac_open(struct inode *inode, struct file *file);
static int ks_proc_ctl_tailtag_read(struct seq_file *m, void *v);
static ssize_t ks_proc_ctl_tailtag_write(struct file *, const char __user *, size_t, loff_t *);
static int ks_proc_ctl_tailtag_open(struct inode *inode, struct file *file);
static int ks_proc_mib_read(struct seq_file *m, void *v);
static int ks_proc_mib_open(struct inode *inode, struct file *file);
static int ks_proc_sroute_read(struct seq_file *m, void *v);
static ssize_t ks_proc_sroute_write(struct file *, const char __user *, size_t, loff_t *);
static int ks_proc_sroute_open(struct inode *inode, struct file *file);
static int ks_proc_droute_read(struct seq_file *m, void *v);
static int ks_proc_droute_open(struct inode *inode, struct file *file);
static void ethsw_isr_tasklet(struct work_struct *work);
static void phy_isr_tasklet(struct work_struct *work);

/*=== Macro-function definitions (File scope) ========================================================================*/
#define ks_info(_ks, _msg...) dev_info(&(_ks)->spidev->dev, _msg)
#define ks_warn(_ks, _msg...) dev_warn(&(_ks)->spidev->dev, _msg)
#define ks_dbg(_ks, _msg...) dev_dbg(&(_ks)->spidev->dev, _msg)
#define ks_err(_ks, _msg...) dev_err(&(_ks)->spidev->dev, _msg)

#define MAC_CHAR_ARRAY_TO_LONGLONG(arr)		((unsigned long long)(arr)[0] << 40) |		\
											((unsigned long long)(arr)[1] << 32) |		\
											((unsigned long long)(arr)[2] << 24) |		\
											((unsigned long long)(arr)[3] << 16) |		\
											((unsigned long long)(arr)[4] <<  8) |		\
											((unsigned long long)(arr)[5])

#define MAC_LONGLONG_TO_CHAR_ARRAY(arr,mac)	(arr)[0] = (unsigned int)(mac >> 40) & 0xFF; 	\
											(arr)[1] = (unsigned int)(mac >> 32) & 0xFF;	\
											(arr)[2] = (unsigned int)(mac >> 24) & 0xFF;	\
											(arr)[3] = (unsigned int)(mac >> 16) & 0xFF;	\
											(arr)[4] = (unsigned int)(mac >>  8) & 0xFF;	\
											(arr)[5] = (unsigned int)(mac)       & 0xFF;

/*=== data definitions (File scope) ==================================================================================*/
static char* am18xx_mac = "";
module_param(am18xx_mac, charp, 0000);

static struct cdev*           dev;

static struct ethsw_net       ethsw;

static const struct of_device_id ksz8864_spi_of_match[] = {
        { .compatible = "micrel,ksz8864ethsw" },
        { .compatible = "micrel,ksz8051" },
        { },
 };
MODULE_DEVICE_TABLE(of, ksz8864_spi_of_match);

static struct spi_driver      ethsw_driver = {
	.driver = {
		.name = "ethsw",
		.owner = THIS_MODULE,
		.of_match_table = of_match_ptr(ksz8864_spi_of_match),
	},
	.probe = ethsw_probe,
	.remove = /*__devexit_p(ethsw_remove)*/ethsw_remove,//KDEBUG
};

static struct file_operations user_ethsw_file_ops = {
	.owner = THIS_MODULE,
	.read = user_ethsw_read,
	.open = user_ethsw_open,
};

static struct ks_proc_entry ks_proc_reg_entries[] = {
  { "chipId0",                (void*)KSZ8864_SMI_CHIP_ID0,          { .owner = THIS_MODULE, .open = ks_proc_reg_open, .write = NULL,              .read = seq_read, .llseek = seq_lseek, .release = single_release,},},
  { "revIdStart",             (void*)KSZ8864_SMI_REV_ID_START,      { .owner = THIS_MODULE, .open = ks_proc_reg_open, .write = ks_proc_reg_write, .read = seq_read, .llseek = seq_lseek, .release = single_release,},},
  { "control0",               (void*)KSZ8864_SMI_CONTROL_0_GLOBAL,  { .owner = THIS_MODULE, .open = ks_proc_reg_open, .write = ks_proc_reg_write, .read = seq_read, .llseek = seq_lseek, .release = single_release,},},
  { "control1",               (void*)KSZ8864_SMI_CONTROL_1_GLOBAL ,  { .owner = THIS_MODULE, .open = ks_proc_reg_open, .write = ks_proc_reg_write, .read = seq_read, .llseek = seq_lseek, .release = single_release,}},
  { "control2",               (void*)KSZ8864_SMI_CONTROL_2_GLOBAL ,  { .owner = THIS_MODULE, .open = ks_proc_reg_open, .write = ks_proc_reg_write, .read = seq_read, .llseek = seq_lseek, .release = single_release,}},
  { "control3",               (void*)KSZ8864_SMI_CONTROL_3_GLOBAL ,  { .owner = THIS_MODULE, .open = ks_proc_reg_open, .write = ks_proc_reg_write, .read = seq_read, .llseek = seq_lseek, .release = single_release,}},
  { "control4",               (void*)KSZ8864_SMI_CONTROL_4_GLOBAL ,  { .owner = THIS_MODULE, .open = ks_proc_reg_open, .write = ks_proc_reg_write, .read = seq_read, .llseek = seq_lseek, .release = single_release,}},
  { "control5",               (void*)KSZ8864_SMI_CONTROL_5_GLOBAL ,  { .owner = THIS_MODULE, .open = ks_proc_reg_open, .write = ks_proc_reg_write, .read = seq_read, .llseek = seq_lseek, .release = single_release,}},
  { "control6",               (void*)KSZ8864_SMI_CONTROL_6_GLOBAL ,          { .owner = THIS_MODULE, .open = ks_proc_reg_open, .write = NULL,              .read = seq_read, .llseek = seq_lseek, .release = single_release,}},
  { "control7",               (void*)KSZ8864_SMI_CONTROL_7_GLOBAL ,          { .owner = THIS_MODULE, .open = ks_proc_reg_open, .write = NULL,              .read = seq_read, .llseek = seq_lseek, .release = single_release,}},
  { "control8",               (void*)KSZ8864_SMI_CONTROL_8_GLOBAL ,          { .owner = THIS_MODULE, .open = ks_proc_reg_open, .write = NULL,              .read = seq_read, .llseek = seq_lseek, .release = single_release,}},
  { "control9",               (void*)KSZ8864_SMI_CONTROL_9_GLOBAL ,  { .owner = THIS_MODULE, .open = ks_proc_reg_open, .write = ks_proc_reg_write, .read = seq_read, .llseek = seq_lseek, .release = single_release,}},
  { "control10",              (void*)KSZ8864_SMI_CONTROL_10_GLOBAL ,  { .owner = THIS_MODULE, .open = ks_proc_reg_open, .write = ks_proc_reg_write, .read = seq_read, .llseek = seq_lseek, .release = single_release,}},
  { "control11",              (void*)KSZ8864_SMI_CONTROL_11_GLOBAL ,          { .owner = THIS_MODULE, .open = ks_proc_reg_open, .write = NULL,              .read = seq_read, .llseek = seq_lseek, .release = single_release,}},
  { "powerControl1",          (void*)KSZ8864_SMI_POWER_CONTROL_1_GLOBAL ,  { .owner = THIS_MODULE, .open = ks_proc_reg_open, .write = ks_proc_reg_write, .read = seq_read, .llseek = seq_lseek, .release = single_release,}},
  { "powerControl2",          (void*)KSZ8864_SMI_POWER_CONTROL_2_GLOBAL ,  { .owner = THIS_MODULE, .open = ks_proc_reg_open, .write = ks_proc_reg_write, .read = seq_read, .llseek = seq_lseek, .release = single_release,}},
  { "interruptStatus",        (void*)KSZ8864_SMI_INTERRUPT_STATUS ,          { .owner = THIS_MODULE, .open = ks_proc_reg_open, .write = NULL,              .read = seq_read, .llseek = seq_lseek, .release = single_release,}},
  { "interruptMask",          (void*)KSZ8864_SMI_INTERRUPT_MASK ,  { .owner = THIS_MODULE, .open = ks_proc_reg_open, .write = ks_proc_reg_write, .read = seq_read, .llseek = seq_lseek, .release = single_release,}},
  { "control12",              (void*)KSZ8864_SMI_CONTROL_12_GLOBAL ,  { .owner = THIS_MODULE, .open = ks_proc_reg_open, .write = ks_proc_reg_write, .read = seq_read, .llseek = seq_lseek, .release = single_release,}},
  { "control13",              (void*)KSZ8864_SMI_CONTROL_13_GLOBAL ,  { .owner = THIS_MODULE, .open = ks_proc_reg_open, .write = ks_proc_reg_write, .read = seq_read, .llseek = seq_lseek, .release = single_release,}},
  { "control14",              (void*)KSZ8864_SMI_CONTROL_14_GLOBAL ,  { .owner = THIS_MODULE, .open = ks_proc_reg_open, .write = ks_proc_reg_write, .read = seq_read, .llseek = seq_lseek, .release = single_release,}},
  { "control15",              (void*)KSZ8864_SMI_CONTROL_15_GLOBAL ,  { .owner = THIS_MODULE, .open = ks_proc_reg_open, .write = ks_proc_reg_write, .read = seq_read, .llseek = seq_lseek, .release = single_release,}},
  { "control16",              (void*)KSZ8864_SMI_CONTROL_16_GLOBAL ,  { .owner = THIS_MODULE, .open = ks_proc_reg_open, .write = ks_proc_reg_write, .read = seq_read, .llseek = seq_lseek, .release = single_release,}},
  { "control17",              (void*)KSZ8864_SMI_CONTROL_17_GLOBAL ,  { .owner = THIS_MODULE, .open = ks_proc_reg_open, .write = ks_proc_reg_write, .read = seq_read, .llseek = seq_lseek, .release = single_release,}},
  { "control18",              (void*)KSZ8864_SMI_CONTROL_18_GLOBAL ,  { .owner = THIS_MODULE, .open = ks_proc_reg_open, .write = ks_proc_reg_write, .read = seq_read, .llseek = seq_lseek, .release = single_release,}},
  { "control19",              (void*)KSZ8864_SMI_CONTROL_19_GLOBAL ,  { .owner = THIS_MODULE, .open = ks_proc_reg_open, .write = ks_proc_reg_write, .read = seq_read, .llseek = seq_lseek, .release = single_release,}},
  { "tosPriorityControl0",    (void*)KSZ8864_SMI_TOS_PRIORITY_CONTROL_0 ,  { .owner = THIS_MODULE, .open = ks_proc_reg_open, .write = ks_proc_reg_write, .read = seq_read, .llseek = seq_lseek, .release = single_release,}},
  { "tosPriorityControl1",    (void*)KSZ8864_SMI_TOS_PRIORITY_CONTROL_1 ,  { .owner = THIS_MODULE, .open = ks_proc_reg_open, .write = ks_proc_reg_write, .read = seq_read, .llseek = seq_lseek, .release = single_release,}},
  { "tosPriorityControl2",    (void*)KSZ8864_SMI_TOS_PRIORITY_CONTROL_2 ,  { .owner = THIS_MODULE, .open = ks_proc_reg_open, .write = ks_proc_reg_write, .read = seq_read, .llseek = seq_lseek, .release = single_release,}},
  { "tosPriorityControl3",    (void*)KSZ8864_SMI_TOS_PRIORITY_CONTROL_3 ,  { .owner = THIS_MODULE, .open = ks_proc_reg_open, .write = ks_proc_reg_write, .read = seq_read, .llseek = seq_lseek, .release = single_release,}},
  { "tosPriorityControl4",    (void*)KSZ8864_SMI_TOS_PRIORITY_CONTROL_4 ,  { .owner = THIS_MODULE, .open = ks_proc_reg_open, .write = ks_proc_reg_write, .read = seq_read, .llseek = seq_lseek, .release = single_release,}},
  { "tosPriorityControl5",    (void*)KSZ8864_SMI_TOS_PRIORITY_CONTROL_5 ,  { .owner = THIS_MODULE, .open = ks_proc_reg_open, .write = ks_proc_reg_write, .read = seq_read, .llseek = seq_lseek, .release = single_release,}},
  { "tosPriorityControl6",    (void*)KSZ8864_SMI_TOS_PRIORITY_CONTROL_6 ,  { .owner = THIS_MODULE, .open = ks_proc_reg_open, .write = ks_proc_reg_write, .read = seq_read, .llseek = seq_lseek, .release = single_release,}},
  { "tosPriorityControl7",    (void*)KSZ8864_SMI_TOS_PRIORITY_CONTROL_7 ,  { .owner = THIS_MODULE, .open = ks_proc_reg_open, .write = ks_proc_reg_write, .read = seq_read, .llseek = seq_lseek, .release = single_release,}},
  { "tosPriorityControl8",    (void*)KSZ8864_SMI_TOS_PRIORITY_CONTROL_8 ,  { .owner = THIS_MODULE, .open = ks_proc_reg_open, .write = ks_proc_reg_write, .read = seq_read, .llseek = seq_lseek, .release = single_release,}},
  { "tosPriorityControl9",    (void*)KSZ8864_SMI_TOS_PRIORITY_CONTROL_9 ,  { .owner = THIS_MODULE, .open = ks_proc_reg_open, .write = ks_proc_reg_write, .read = seq_read, .llseek = seq_lseek, .release = single_release,}},
  { "tosPriorityControl10",   (void*)KSZ8864_SMI_TOS_PRIORITY_CONTROL_10 ,  { .owner = THIS_MODULE, .open = ks_proc_reg_open, .write = ks_proc_reg_write, .read = seq_read, .llseek = seq_lseek, .release = single_release,}},
  { "tosPriorityControl11",   (void*)KSZ8864_SMI_TOS_PRIORITY_CONTROL_11 ,  { .owner = THIS_MODULE, .open = ks_proc_reg_open, .write = ks_proc_reg_write, .read = seq_read, .llseek = seq_lseek, .release = single_release,}},
  { "tosPriorityControl12",   (void*)KSZ8864_SMI_TOS_PRIORITY_CONTROL_12 ,  { .owner = THIS_MODULE, .open = ks_proc_reg_open, .write = ks_proc_reg_write, .read = seq_read, .llseek = seq_lseek, .release = single_release,}},
  { "tosPriorityControl13",   (void*)KSZ8864_SMI_TOS_PRIORITY_CONTROL_13 ,  { .owner = THIS_MODULE, .open = ks_proc_reg_open, .write = ks_proc_reg_write, .read = seq_read, .llseek = seq_lseek, .release = single_release,}},
  { "tosPriorityControl14",   (void*)KSZ8864_SMI_TOS_PRIORITY_CONTROL_14 ,  { .owner = THIS_MODULE, .open = ks_proc_reg_open, .write = ks_proc_reg_write, .read = seq_read, .llseek = seq_lseek, .release = single_release,}},
  { "tosPriorityControl15",   (void*)KSZ8864_SMI_TOS_PRIORITY_CONTROL_15 ,  { .owner = THIS_MODULE, .open = ks_proc_reg_open, .write = ks_proc_reg_write, .read = seq_read, .llseek = seq_lseek, .release = single_release,}},
  { "macAddress0",            (void*)KSZ8864_SMI_MAC_ADDRESS_0 ,  { .owner = THIS_MODULE, .open = ks_proc_reg_open, .write = ks_proc_reg_write, .read = seq_read, .llseek = seq_lseek, .release = single_release,}},
  { "macAddress1",            (void*)KSZ8864_SMI_MAC_ADDRESS_1 ,  { .owner = THIS_MODULE, .open = ks_proc_reg_open, .write = ks_proc_reg_write, .read = seq_read, .llseek = seq_lseek, .release = single_release,}},
  { "macAddress2",            (void*)KSZ8864_SMI_MAC_ADDRESS_2 ,  { .owner = THIS_MODULE, .open = ks_proc_reg_open, .write = ks_proc_reg_write, .read = seq_read, .llseek = seq_lseek, .release = single_release,}},
  { "macAddress3",            (void*)KSZ8864_SMI_MAC_ADDRESS_3 ,  { .owner = THIS_MODULE, .open = ks_proc_reg_open, .write = ks_proc_reg_write, .read = seq_read, .llseek = seq_lseek, .release = single_release,}},
  { "macAddress4",            (void*)KSZ8864_SMI_MAC_ADDRESS_4 ,  { .owner = THIS_MODULE, .open = ks_proc_reg_open, .write = ks_proc_reg_write, .read = seq_read, .llseek = seq_lseek, .release = single_release,}},
  { "macAddress5",            (void*)KSZ8864_SMI_MAC_ADDRESS_5 ,  { .owner = THIS_MODULE, .open = ks_proc_reg_open, .write = ks_proc_reg_write, .read = seq_read, .llseek = seq_lseek, .release = single_release,}},
  { "inderectAccessControl0", (void*)KSZ8864_SMI_INDERECT_ACCESS_CONTROL_0 ,  { .owner = THIS_MODULE, .open = ks_proc_reg_open, .write = ks_proc_reg_write, .read = seq_read, .llseek = seq_lseek, .release = single_release,}},
  { "inderectAccessControl1", (void*)KSZ8864_SMI_INDERECT_ACCESS_CONTROL_1 ,  { .owner = THIS_MODULE, .open = ks_proc_reg_open, .write = ks_proc_reg_write, .read = seq_read, .llseek = seq_lseek, .release = single_release,}},
  { "indirectData8",          (void*)KSZ8864_SMI_INDIRECT_DATA_8 ,  { .owner = THIS_MODULE, .open = ks_proc_reg_open, .write = ks_proc_reg_write, .read = seq_read, .llseek = seq_lseek, .release = single_release,}},
  { "indirectData7",          (void*)KSZ8864_SMI_INDIRECT_DATA_7 ,  { .owner = THIS_MODULE, .open = ks_proc_reg_open, .write = ks_proc_reg_write, .read = seq_read, .llseek = seq_lseek, .release = single_release,}},
  { "indirectData6",          (void*)KSZ8864_SMI_INDIRECT_DATA_6 ,  { .owner = THIS_MODULE, .open = ks_proc_reg_open, .write = ks_proc_reg_write, .read = seq_read, .llseek = seq_lseek, .release = single_release,}},
  { "indirectData5",          (void*)KSZ8864_SMI_INDIRECT_DATA_5 ,  { .owner = THIS_MODULE, .open = ks_proc_reg_open, .write = ks_proc_reg_write, .read = seq_read, .llseek = seq_lseek, .release = single_release,}},
  { "indirectData4",          (void*)KSZ8864_SMI_INDIRECT_DATA_4 ,  { .owner = THIS_MODULE, .open = ks_proc_reg_open, .write = ks_proc_reg_write, .read = seq_read, .llseek = seq_lseek, .release = single_release,}},
  { "indirectData3",          (void*)KSZ8864_SMI_INDIRECT_DATA_3 ,  { .owner = THIS_MODULE, .open = ks_proc_reg_open, .write = ks_proc_reg_write, .read = seq_read, .llseek = seq_lseek, .release = single_release,}},
  { "indirectData2",          (void*)KSZ8864_SMI_INDIRECT_DATA_2 ,  { .owner = THIS_MODULE, .open = ks_proc_reg_open, .write = ks_proc_reg_write, .read = seq_read, .llseek = seq_lseek, .release = single_release,}},
  { "indirectData1",          (void*)KSZ8864_SMI_INDIRECT_DATA_1 ,  { .owner = THIS_MODULE, .open = ks_proc_reg_open, .write = ks_proc_reg_write, .read = seq_read, .llseek = seq_lseek, .release = single_release,}},
  { "indirectData0",          (void*)KSZ8864_SMI_INDIRECT_DATA_0,  { .owner = THIS_MODULE, .open = ks_proc_reg_open, .write = ks_proc_reg_write, .read = seq_read, .llseek = seq_lseek, .release = single_release,}},

  { "p1_control0",            (void*)KSZ8864_SMI_PORT1_OFFSET+KSZ8864_SMI_PORTX_CONTROL_0 ,  { .owner = THIS_MODULE, .open = ks_proc_reg_open, .write = ks_proc_reg_write, .read = seq_read, .llseek = seq_lseek, .release = single_release,}},
  { "p1_control1",            (void*)KSZ8864_SMI_PORT1_OFFSET+KSZ8864_SMI_PORTX_CONTROL_1 ,  { .owner = THIS_MODULE, .open = ks_proc_reg_open, .write = ks_proc_reg_write, .read = seq_read, .llseek = seq_lseek, .release = single_release,}},
  { "p1_control2",            (void*)KSZ8864_SMI_PORT1_OFFSET+KSZ8864_SMI_PORTX_CONTROL_2 ,  { .owner = THIS_MODULE, .open = ks_proc_reg_open, .write = ks_proc_reg_write, .read = seq_read, .llseek = seq_lseek, .release = single_release,}},
  { "p1_control3",            (void*)KSZ8864_SMI_PORT1_OFFSET+KSZ8864_SMI_PORTX_CONTROL_3 ,  { .owner = THIS_MODULE, .open = ks_proc_reg_open, .write = ks_proc_reg_write, .read = seq_read, .llseek = seq_lseek, .release = single_release,}},
  { "p1_control4",            (void*)KSZ8864_SMI_PORT1_OFFSET+KSZ8864_SMI_PORTX_CONTROL_4 ,  { .owner = THIS_MODULE, .open = ks_proc_reg_open, .write = ks_proc_reg_write, .read = seq_read, .llseek = seq_lseek, .release = single_release,}},
  { "p1_status0",             (void*)KSZ8864_SMI_PORT1_OFFSET+KSZ8864_SMI_PORTX_STATUS_0 ,  { .owner = THIS_MODULE, .open = ks_proc_reg_open, .write = ks_proc_reg_write, .read = seq_read, .llseek = seq_lseek, .release = single_release,}},
  { "p1_phyCtrlStat",         (void*)KSZ8864_SMI_PORT1_OFFSET+KSZ8864_SMI_PORTX_PHY_CTRL_STAT ,  { .owner = THIS_MODULE, .open = ks_proc_reg_open, .write = ks_proc_reg_write, .read = seq_read, .llseek = seq_lseek, .release = single_release,}},
  { "p1_control5",            (void*)KSZ8864_SMI_PORT1_OFFSET+KSZ8864_SMI_PORTX_CONTROL_5 ,  { .owner = THIS_MODULE, .open = ks_proc_reg_open, .write = ks_proc_reg_write, .read = seq_read, .llseek = seq_lseek, .release = single_release,}},
  { "p1_control6",            (void*)KSZ8864_SMI_PORT1_OFFSET+KSZ8864_SMI_PORTX_CONTROL_6 ,  { .owner = THIS_MODULE, .open = ks_proc_reg_open, .write = ks_proc_reg_write, .read = seq_read, .llseek = seq_lseek, .release = single_release,}},
  { "p1_status1",             (void*)KSZ8864_SMI_PORT1_OFFSET+KSZ8864_SMI_PORTX_STATUS_1 ,          { .owner = THIS_MODULE, .open = ks_proc_reg_open, .write = NULL,              .read = seq_read, .llseek = seq_lseek, .release = single_release,}},
  { "p1_control7status2",     (void*)KSZ8864_SMI_PORT1_OFFSET+KSZ8864_SMI_PORTX_CONTROL_7_STATUS_2 ,  { .owner = THIS_MODULE, .open = ks_proc_reg_open, .write = ks_proc_reg_write, .read = seq_read, .llseek = seq_lseek, .release = single_release,}},
  { "p1_control8",            (void*)KSZ8864_SMI_PORT1_OFFSET+KSZ8864_SMI_PORTX_CONTROL_8 ,  { .owner = THIS_MODULE, .open = ks_proc_reg_open, .write = ks_proc_reg_write, .read = seq_read, .llseek = seq_lseek, .release = single_release,}},
  { "p1_control9",            (void*)KSZ8864_SMI_PORT1_OFFSET+KSZ8864_SMI_PORTX_CONTROL_9 ,  { .owner = THIS_MODULE, .open = ks_proc_reg_open, .write = ks_proc_reg_write, .read = seq_read, .llseek = seq_lseek, .release = single_release,}},
  { "p1_control10",           (void*)KSZ8864_SMI_PORT1_OFFSET+KSZ8864_SMI_PORTX_CONTROL_10 ,  { .owner = THIS_MODULE, .open = ks_proc_reg_open, .write = ks_proc_reg_write, .read = seq_read, .llseek = seq_lseek, .release = single_release,}},
  { "p1_control11",           (void*)KSZ8864_SMI_PORT1_OFFSET+KSZ8864_SMI_PORTX_CONTROL_11 ,  { .owner = THIS_MODULE, .open = ks_proc_reg_open, .write = ks_proc_reg_write, .read = seq_read, .llseek = seq_lseek, .release = single_release,}},
  { "p1_control12",           (void*)KSZ8864_SMI_PORT1_OFFSET+KSZ8864_SMI_PORTX_CONTROL_12 ,  { .owner = THIS_MODULE, .open = ks_proc_reg_open, .write = ks_proc_reg_write, .read = seq_read, .llseek = seq_lseek, .release = single_release,}},
  { "p1_control13",           (void*)KSZ8864_SMI_PORT1_OFFSET+KSZ8864_SMI_PORTX_CONTROL_13 ,  { .owner = THIS_MODULE, .open = ks_proc_reg_open, .write = ks_proc_reg_write, .read = seq_read, .llseek = seq_lseek, .release = single_release,}},
  { "p1_rateControl",         (void*)KSZ8864_SMI_PORT1_OFFSET+KSZ8864_SMI_PORTX_RATE_CONTROL ,  { .owner = THIS_MODULE, .open = ks_proc_reg_open, .write = ks_proc_reg_write, .read = seq_read, .llseek = seq_lseek, .release = single_release,}},
  { "p1_ingressControl1",     (void*)KSZ8864_SMI_PORT1_OFFSET+KSZ8864_SMI_PORTX_INGRESS_CONTROL_1 ,  { .owner = THIS_MODULE, .open = ks_proc_reg_open, .write = ks_proc_reg_write, .read = seq_read, .llseek = seq_lseek, .release = single_release,}},
  { "p1_ingressControl2",     (void*)KSZ8864_SMI_PORT1_OFFSET+KSZ8864_SMI_PORTX_INGRESS_CONTROL_2 ,  { .owner = THIS_MODULE, .open = ks_proc_reg_open, .write = ks_proc_reg_write, .read = seq_read, .llseek = seq_lseek, .release = single_release,}},
  { "p1_ingressControl3",     (void*)KSZ8864_SMI_PORT1_OFFSET+KSZ8864_SMI_PORTX_INGRESS_CONTROL_3 ,  { .owner = THIS_MODULE, .open = ks_proc_reg_open, .write = ks_proc_reg_write, .read = seq_read, .llseek = seq_lseek, .release = single_release,}},
  { "p1_ingressControl4",     (void*)KSZ8864_SMI_PORT1_OFFSET+KSZ8864_SMI_PORTX_INGRESS_CONTROL_4 ,  { .owner = THIS_MODULE, .open = ks_proc_reg_open, .write = ks_proc_reg_write, .read = seq_read, .llseek = seq_lseek, .release = single_release,}},
  { "p1_egressControl1",      (void*)KSZ8864_SMI_PORT1_OFFSET+KSZ8864_SMI_PORTX_EGRESS_CONTROL_1 ,  { .owner = THIS_MODULE, .open = ks_proc_reg_open, .write = ks_proc_reg_write, .read = seq_read, .llseek = seq_lseek, .release = single_release,}},
  { "p1_egressControl2",      (void*)KSZ8864_SMI_PORT1_OFFSET+KSZ8864_SMI_PORTX_EGRESS_CONTROL_2 ,  { .owner = THIS_MODULE, .open = ks_proc_reg_open, .write = ks_proc_reg_write, .read = seq_read, .llseek = seq_lseek, .release = single_release,}},
  { "p1_egressControl3",      (void*)KSZ8864_SMI_PORT1_OFFSET+KSZ8864_SMI_PORTX_EGRESS_CONTROL_3 ,  { .owner = THIS_MODULE, .open = ks_proc_reg_open, .write = ks_proc_reg_write, .read = seq_read, .llseek = seq_lseek, .release = single_release,}},
  { "p1_egressControl4",      (void*)KSZ8864_SMI_PORT1_OFFSET+KSZ8864_SMI_PORTX_EGRESS_CONTROL_4 ,  { .owner = THIS_MODULE, .open = ks_proc_reg_open, .write = ks_proc_reg_write, .read = seq_read, .llseek = seq_lseek, .release = single_release,}},

  { "p2_control0",            (void*)KSZ8864_SMI_PORT2_OFFSET+KSZ8864_SMI_PORTX_CONTROL_0 ,  { .owner = THIS_MODULE, .open = ks_proc_reg_open, .write = ks_proc_reg_write, .read = seq_read, .llseek = seq_lseek, .release = single_release,}},
  { "p2_control1",            (void*)KSZ8864_SMI_PORT2_OFFSET+KSZ8864_SMI_PORTX_CONTROL_1 ,  { .owner = THIS_MODULE, .open = ks_proc_reg_open, .write = ks_proc_reg_write, .read = seq_read, .llseek = seq_lseek, .release = single_release,}},
  { "p2_control2",            (void*)KSZ8864_SMI_PORT2_OFFSET+KSZ8864_SMI_PORTX_CONTROL_2 ,  { .owner = THIS_MODULE, .open = ks_proc_reg_open, .write = ks_proc_reg_write, .read = seq_read, .llseek = seq_lseek, .release = single_release,}},
  { "p2_control3",            (void*)KSZ8864_SMI_PORT2_OFFSET+KSZ8864_SMI_PORTX_CONTROL_3 ,  { .owner = THIS_MODULE, .open = ks_proc_reg_open, .write = ks_proc_reg_write, .read = seq_read, .llseek = seq_lseek, .release = single_release,}},
  { "p2_control4",            (void*)KSZ8864_SMI_PORT2_OFFSET+KSZ8864_SMI_PORTX_CONTROL_4 ,  { .owner = THIS_MODULE, .open = ks_proc_reg_open, .write = ks_proc_reg_write, .read = seq_read, .llseek = seq_lseek, .release = single_release,}},
  { "p2_status0",             (void*)KSZ8864_SMI_PORT2_OFFSET+KSZ8864_SMI_PORTX_STATUS_0 ,  { .owner = THIS_MODULE, .open = ks_proc_reg_open, .write = ks_proc_reg_write, .read = seq_read, .llseek = seq_lseek, .release = single_release,}},
  { "p2_phyCtrlStat",         (void*)KSZ8864_SMI_PORT2_OFFSET+KSZ8864_SMI_PORTX_PHY_CTRL_STAT ,  { .owner = THIS_MODULE, .open = ks_proc_reg_open, .write = ks_proc_reg_write, .read = seq_read, .llseek = seq_lseek, .release = single_release,}},
  { "p2_control5",            (void*)KSZ8864_SMI_PORT2_OFFSET+KSZ8864_SMI_PORTX_CONTROL_5 ,  { .owner = THIS_MODULE, .open = ks_proc_reg_open, .write = ks_proc_reg_write, .read = seq_read, .llseek = seq_lseek, .release = single_release,}},
  { "p2_control6",            (void*)KSZ8864_SMI_PORT2_OFFSET+KSZ8864_SMI_PORTX_CONTROL_6 ,  { .owner = THIS_MODULE, .open = ks_proc_reg_open, .write = ks_proc_reg_write, .read = seq_read, .llseek = seq_lseek, .release = single_release,}},
  { "p2_status1",             (void*)KSZ8864_SMI_PORT2_OFFSET+KSZ8864_SMI_PORTX_STATUS_1 ,          { .owner = THIS_MODULE, .open = ks_proc_reg_open, .write = NULL,              .read = seq_read, .llseek = seq_lseek, .release = single_release,}},
  { "p2_control7status2",     (void*)KSZ8864_SMI_PORT2_OFFSET+KSZ8864_SMI_PORTX_CONTROL_7_STATUS_2 ,  { .owner = THIS_MODULE, .open = ks_proc_reg_open, .write = ks_proc_reg_write, .read = seq_read, .llseek = seq_lseek, .release = single_release,}},
  { "p2_control8",            (void*)KSZ8864_SMI_PORT2_OFFSET+KSZ8864_SMI_PORTX_CONTROL_8 ,  { .owner = THIS_MODULE, .open = ks_proc_reg_open, .write = ks_proc_reg_write, .read = seq_read, .llseek = seq_lseek, .release = single_release,}},
  { "p2_control9",            (void*)KSZ8864_SMI_PORT2_OFFSET+KSZ8864_SMI_PORTX_CONTROL_9 ,  { .owner = THIS_MODULE, .open = ks_proc_reg_open, .write = ks_proc_reg_write, .read = seq_read, .llseek = seq_lseek, .release = single_release,}},
  { "p2_control10",           (void*)KSZ8864_SMI_PORT2_OFFSET+KSZ8864_SMI_PORTX_CONTROL_10 ,  { .owner = THIS_MODULE, .open = ks_proc_reg_open, .write = ks_proc_reg_write, .read = seq_read, .llseek = seq_lseek, .release = single_release,}},
  { "p2_control11",           (void*)KSZ8864_SMI_PORT2_OFFSET+KSZ8864_SMI_PORTX_CONTROL_11 ,  { .owner = THIS_MODULE, .open = ks_proc_reg_open, .write = ks_proc_reg_write, .read = seq_read, .llseek = seq_lseek, .release = single_release,}},
  { "p2_control12",           (void*)KSZ8864_SMI_PORT2_OFFSET+KSZ8864_SMI_PORTX_CONTROL_12 ,  { .owner = THIS_MODULE, .open = ks_proc_reg_open, .write = ks_proc_reg_write, .read = seq_read, .llseek = seq_lseek, .release = single_release,}},
  { "p2_control13",           (void*)KSZ8864_SMI_PORT2_OFFSET+KSZ8864_SMI_PORTX_CONTROL_13 ,  { .owner = THIS_MODULE, .open = ks_proc_reg_open, .write = ks_proc_reg_write, .read = seq_read, .llseek = seq_lseek, .release = single_release,}},
  { "p2_rateControl",         (void*)KSZ8864_SMI_PORT2_OFFSET+KSZ8864_SMI_PORTX_RATE_CONTROL ,  { .owner = THIS_MODULE, .open = ks_proc_reg_open, .write = ks_proc_reg_write, .read = seq_read, .llseek = seq_lseek, .release = single_release,}},
  { "p2_ingressControl1",     (void*)KSZ8864_SMI_PORT2_OFFSET+KSZ8864_SMI_PORTX_INGRESS_CONTROL_1 ,  { .owner = THIS_MODULE, .open = ks_proc_reg_open, .write = ks_proc_reg_write, .read = seq_read, .llseek = seq_lseek, .release = single_release,}},
  { "p2_ingressControl2",     (void*)KSZ8864_SMI_PORT2_OFFSET+KSZ8864_SMI_PORTX_INGRESS_CONTROL_2 ,  { .owner = THIS_MODULE, .open = ks_proc_reg_open, .write = ks_proc_reg_write, .read = seq_read, .llseek = seq_lseek, .release = single_release,}},
  { "p2_ingressControl3",     (void*)KSZ8864_SMI_PORT2_OFFSET+KSZ8864_SMI_PORTX_INGRESS_CONTROL_3 ,  { .owner = THIS_MODULE, .open = ks_proc_reg_open, .write = ks_proc_reg_write, .read = seq_read, .llseek = seq_lseek, .release = single_release,}},
  { "p2_ingressControl4",     (void*)KSZ8864_SMI_PORT2_OFFSET+KSZ8864_SMI_PORTX_INGRESS_CONTROL_4 ,  { .owner = THIS_MODULE, .open = ks_proc_reg_open, .write = ks_proc_reg_write, .read = seq_read, .llseek = seq_lseek, .release = single_release,}},
  { "p2_egressControl1",      (void*)KSZ8864_SMI_PORT2_OFFSET+KSZ8864_SMI_PORTX_EGRESS_CONTROL_1 ,  { .owner = THIS_MODULE, .open = ks_proc_reg_open, .write = ks_proc_reg_write, .read = seq_read, .llseek = seq_lseek, .release = single_release,}},
  { "p2_egressControl2",      (void*)KSZ8864_SMI_PORT2_OFFSET+KSZ8864_SMI_PORTX_EGRESS_CONTROL_2 ,  { .owner = THIS_MODULE, .open = ks_proc_reg_open, .write = ks_proc_reg_write, .read = seq_read, .llseek = seq_lseek, .release = single_release,}},
  { "p2_egressControl3",      (void*)KSZ8864_SMI_PORT2_OFFSET+KSZ8864_SMI_PORTX_EGRESS_CONTROL_3 ,  { .owner = THIS_MODULE, .open = ks_proc_reg_open, .write = ks_proc_reg_write, .read = seq_read, .llseek = seq_lseek, .release = single_release,}},
  { "p2_egressControl4",      (void*)KSZ8864_SMI_PORT2_OFFSET+KSZ8864_SMI_PORTX_EGRESS_CONTROL_4 ,  { .owner = THIS_MODULE, .open = ks_proc_reg_open, .write = ks_proc_reg_write, .read = seq_read, .llseek = seq_lseek, .release = single_release,}},

  { "p3_control0",            (void*)KSZ8864_SMI_PORT3_OFFSET+KSZ8864_SMI_PORTX_CONTROL_0 ,  { .owner = THIS_MODULE, .open = ks_proc_reg_open, .write = ks_proc_reg_write, .read = seq_read, .llseek = seq_lseek, .release = single_release,}},
  { "p3_control1",            (void*)KSZ8864_SMI_PORT3_OFFSET+KSZ8864_SMI_PORTX_CONTROL_1 ,  { .owner = THIS_MODULE, .open = ks_proc_reg_open, .write = ks_proc_reg_write, .read = seq_read, .llseek = seq_lseek, .release = single_release,}},
  { "p3_control2",            (void*)KSZ8864_SMI_PORT3_OFFSET+KSZ8864_SMI_PORTX_CONTROL_2 ,  { .owner = THIS_MODULE, .open = ks_proc_reg_open, .write = ks_proc_reg_write, .read = seq_read, .llseek = seq_lseek, .release = single_release,}},
  { "p3_control3",            (void*)KSZ8864_SMI_PORT3_OFFSET+KSZ8864_SMI_PORTX_CONTROL_3 ,  { .owner = THIS_MODULE, .open = ks_proc_reg_open, .write = ks_proc_reg_write, .read = seq_read, .llseek = seq_lseek, .release = single_release,}},
  { "p3_control4",            (void*)KSZ8864_SMI_PORT3_OFFSET+KSZ8864_SMI_PORTX_CONTROL_4 ,  { .owner = THIS_MODULE, .open = ks_proc_reg_open, .write = ks_proc_reg_write, .read = seq_read, .llseek = seq_lseek, .release = single_release,}},
  { "p3_status0",             (void*)KSZ8864_SMI_PORT3_OFFSET+KSZ8864_SMI_PORTX_STATUS_0 ,  { .owner = THIS_MODULE, .open = ks_proc_reg_open, .write = ks_proc_reg_write, .read = seq_read, .llseek = seq_lseek, .release = single_release,}},
  { "p3_control6",            (void*)KSZ8864_SMI_PORT3_OFFSET+KSZ8864_SMI_PORTX_CONTROL_6 ,  { .owner = THIS_MODULE, .open = ks_proc_reg_open, .write = ks_proc_reg_write, .read = seq_read, .llseek = seq_lseek, .release = single_release,}},
  { "p3_control8",            (void*)KSZ8864_SMI_PORT3_OFFSET+KSZ8864_SMI_PORTX_CONTROL_8 ,  { .owner = THIS_MODULE, .open = ks_proc_reg_open, .write = ks_proc_reg_write, .read = seq_read, .llseek = seq_lseek, .release = single_release,}},
  { "p3_control9",            (void*)KSZ8864_SMI_PORT3_OFFSET+KSZ8864_SMI_PORTX_CONTROL_9 ,  { .owner = THIS_MODULE, .open = ks_proc_reg_open, .write = ks_proc_reg_write, .read = seq_read, .llseek = seq_lseek, .release = single_release,}},
  { "p3_control10",           (void*)KSZ8864_SMI_PORT3_OFFSET+KSZ8864_SMI_PORTX_CONTROL_10 ,  { .owner = THIS_MODULE, .open = ks_proc_reg_open, .write = ks_proc_reg_write, .read = seq_read, .llseek = seq_lseek, .release = single_release,}},
  { "p3_control11",           (void*)KSZ8864_SMI_PORT3_OFFSET+KSZ8864_SMI_PORTX_CONTROL_11 ,  { .owner = THIS_MODULE, .open = ks_proc_reg_open, .write = ks_proc_reg_write, .read = seq_read, .llseek = seq_lseek, .release = single_release,}},
  { "p3_control12",           (void*)KSZ8864_SMI_PORT3_OFFSET+KSZ8864_SMI_PORTX_CONTROL_12 ,  { .owner = THIS_MODULE, .open = ks_proc_reg_open, .write = ks_proc_reg_write, .read = seq_read, .llseek = seq_lseek, .release = single_release,}},
  { "p3_control13",           (void*)KSZ8864_SMI_PORT3_OFFSET+KSZ8864_SMI_PORTX_CONTROL_13 ,  { .owner = THIS_MODULE, .open = ks_proc_reg_open, .write = ks_proc_reg_write, .read = seq_read, .llseek = seq_lseek, .release = single_release,}},
  { "p3_rateControl",         (void*)KSZ8864_SMI_PORT3_OFFSET+KSZ8864_SMI_PORTX_RATE_CONTROL ,  { .owner = THIS_MODULE, .open = ks_proc_reg_open, .write = ks_proc_reg_write, .read = seq_read, .llseek = seq_lseek, .release = single_release,}},
  { "p3_ingressControl1",     (void*)KSZ8864_SMI_PORT3_OFFSET+KSZ8864_SMI_PORTX_INGRESS_CONTROL_1 ,  { .owner = THIS_MODULE, .open = ks_proc_reg_open, .write = ks_proc_reg_write, .read = seq_read, .llseek = seq_lseek, .release = single_release,}},
  { "p3_ingressControl2",     (void*)KSZ8864_SMI_PORT3_OFFSET+KSZ8864_SMI_PORTX_INGRESS_CONTROL_2 ,  { .owner = THIS_MODULE, .open = ks_proc_reg_open, .write = ks_proc_reg_write, .read = seq_read, .llseek = seq_lseek, .release = single_release,}},
  { "p3_ingressControl3",     (void*)KSZ8864_SMI_PORT3_OFFSET+KSZ8864_SMI_PORTX_INGRESS_CONTROL_3 ,  { .owner = THIS_MODULE, .open = ks_proc_reg_open, .write = ks_proc_reg_write, .read = seq_read, .llseek = seq_lseek, .release = single_release,}},
  { "p3_ingressControl4",     (void*)KSZ8864_SMI_PORT3_OFFSET+KSZ8864_SMI_PORTX_INGRESS_CONTROL_4 ,  { .owner = THIS_MODULE, .open = ks_proc_reg_open, .write = ks_proc_reg_write, .read = seq_read, .llseek = seq_lseek, .release = single_release,}},
  { "p3_egressControl1",      (void*)KSZ8864_SMI_PORT3_OFFSET+KSZ8864_SMI_PORTX_EGRESS_CONTROL_1 ,  { .owner = THIS_MODULE, .open = ks_proc_reg_open, .write = ks_proc_reg_write, .read = seq_read, .llseek = seq_lseek, .release = single_release,}},
  { "p3_egressControl2",      (void*)KSZ8864_SMI_PORT3_OFFSET+KSZ8864_SMI_PORTX_EGRESS_CONTROL_2 ,  { .owner = THIS_MODULE, .open = ks_proc_reg_open, .write = ks_proc_reg_write, .read = seq_read, .llseek = seq_lseek, .release = single_release,}},
  { "p3_egressControl3",      (void*)KSZ8864_SMI_PORT3_OFFSET+KSZ8864_SMI_PORTX_EGRESS_CONTROL_3 ,  { .owner = THIS_MODULE, .open = ks_proc_reg_open, .write = ks_proc_reg_write, .read = seq_read, .llseek = seq_lseek, .release = single_release,}},
  { "p3_egressControl4",      (void*)KSZ8864_SMI_PORT3_OFFSET+KSZ8864_SMI_PORTX_EGRESS_CONTROL_4 ,  { .owner = THIS_MODULE, .open = ks_proc_reg_open, .write = ks_proc_reg_write, .read = seq_read, .llseek = seq_lseek, .release = single_release,}},
  { "p3_MiiControl1",         (void*)KSZ8864_SMI_PORT3_MII_CONTROL_1 ,  { .owner = THIS_MODULE, .open = ks_proc_reg_open, .write = ks_proc_reg_write, .read = seq_read, .llseek = seq_lseek, .release = single_release,}},
  { "p3_MiiControl2",         (void*)KSZ8864_SMI_PORT3_MII_CONTROL_2 ,  { .owner = THIS_MODULE, .open = ks_proc_reg_open, .write = ks_proc_reg_write, .read = seq_read, .llseek = seq_lseek, .release = single_release,}},

  { "p4_control0",            (void*)KSZ8864_SMI_PORT4_OFFSET+KSZ8864_SMI_PORTX_CONTROL_0 ,  { .owner = THIS_MODULE, .open = ks_proc_reg_open, .write = ks_proc_reg_write, .read = seq_read, .llseek = seq_lseek, .release = single_release,}},
  { "p4_control1",            (void*)KSZ8864_SMI_PORT4_OFFSET+KSZ8864_SMI_PORTX_CONTROL_1 ,  { .owner = THIS_MODULE, .open = ks_proc_reg_open, .write = ks_proc_reg_write, .read = seq_read, .llseek = seq_lseek, .release = single_release,}},
  { "p4_control2",            (void*)KSZ8864_SMI_PORT4_OFFSET+KSZ8864_SMI_PORTX_CONTROL_2 ,  { .owner = THIS_MODULE, .open = ks_proc_reg_open, .write = ks_proc_reg_write, .read = seq_read, .llseek = seq_lseek, .release = single_release,}},
  { "p4_control3",            (void*)KSZ8864_SMI_PORT4_OFFSET+KSZ8864_SMI_PORTX_CONTROL_3 ,  { .owner = THIS_MODULE, .open = ks_proc_reg_open, .write = ks_proc_reg_write, .read = seq_read, .llseek = seq_lseek, .release = single_release,}},
  { "p4_control4",            (void*)KSZ8864_SMI_PORT4_OFFSET+KSZ8864_SMI_PORTX_CONTROL_4 ,  { .owner = THIS_MODULE, .open = ks_proc_reg_open, .write = ks_proc_reg_write, .read = seq_read, .llseek = seq_lseek, .release = single_release,}},
  { "p4_status0",             (void*)KSZ8864_SMI_PORT4_OFFSET+KSZ8864_SMI_PORTX_STATUS_0 ,  { .owner = THIS_MODULE, .open = ks_proc_reg_open, .write = ks_proc_reg_write, .read = seq_read, .llseek = seq_lseek, .release = single_release,}},
  { "p4_control6",            (void*)KSZ8864_SMI_PORT4_OFFSET+KSZ8864_SMI_PORTX_CONTROL_6 ,  { .owner = THIS_MODULE, .open = ks_proc_reg_open, .write = ks_proc_reg_write, .read = seq_read, .llseek = seq_lseek, .release = single_release,}},
  { "p4_control8",            (void*)KSZ8864_SMI_PORT4_OFFSET+KSZ8864_SMI_PORTX_CONTROL_8 ,  { .owner = THIS_MODULE, .open = ks_proc_reg_open, .write = ks_proc_reg_write, .read = seq_read, .llseek = seq_lseek, .release = single_release,}},
  { "p4_control9",            (void*)KSZ8864_SMI_PORT4_OFFSET+KSZ8864_SMI_PORTX_CONTROL_9 ,  { .owner = THIS_MODULE, .open = ks_proc_reg_open, .write = ks_proc_reg_write, .read = seq_read, .llseek = seq_lseek, .release = single_release,}},
  { "p4_control10",           (void*)KSZ8864_SMI_PORT4_OFFSET+KSZ8864_SMI_PORTX_CONTROL_10 ,  { .owner = THIS_MODULE, .open = ks_proc_reg_open, .write = ks_proc_reg_write, .read = seq_read, .llseek = seq_lseek, .release = single_release,}},
  { "p4_control11",           (void*)KSZ8864_SMI_PORT4_OFFSET+KSZ8864_SMI_PORTX_CONTROL_11 ,  { .owner = THIS_MODULE, .open = ks_proc_reg_open, .write = ks_proc_reg_write, .read = seq_read, .llseek = seq_lseek, .release = single_release,}},
  { "p4_control12",           (void*)KSZ8864_SMI_PORT4_OFFSET+KSZ8864_SMI_PORTX_CONTROL_12 ,  { .owner = THIS_MODULE, .open = ks_proc_reg_open, .write = ks_proc_reg_write, .read = seq_read, .llseek = seq_lseek, .release = single_release,}},
  { "p4_control13",           (void*)KSZ8864_SMI_PORT4_OFFSET+KSZ8864_SMI_PORTX_CONTROL_13 ,  { .owner = THIS_MODULE, .open = ks_proc_reg_open, .write = ks_proc_reg_write, .read = seq_read, .llseek = seq_lseek, .release = single_release,}},
  { "p4_rateControl",         (void*)KSZ8864_SMI_PORT4_OFFSET+KSZ8864_SMI_PORTX_RATE_CONTROL ,  { .owner = THIS_MODULE, .open = ks_proc_reg_open, .write = ks_proc_reg_write, .read = seq_read, .llseek = seq_lseek, .release = single_release,}},
  { "p4_ingressControl1",     (void*)KSZ8864_SMI_PORT4_OFFSET+KSZ8864_SMI_PORTX_INGRESS_CONTROL_1 ,  { .owner = THIS_MODULE, .open = ks_proc_reg_open, .write = ks_proc_reg_write, .read = seq_read, .llseek = seq_lseek, .release = single_release,}},
  { "p4_ingressControl2",     (void*)KSZ8864_SMI_PORT4_OFFSET+KSZ8864_SMI_PORTX_INGRESS_CONTROL_2 ,  { .owner = THIS_MODULE, .open = ks_proc_reg_open, .write = ks_proc_reg_write, .read = seq_read, .llseek = seq_lseek, .release = single_release,}},
  { "p4_ingressControl3",     (void*)KSZ8864_SMI_PORT4_OFFSET+KSZ8864_SMI_PORTX_INGRESS_CONTROL_3 ,  { .owner = THIS_MODULE, .open = ks_proc_reg_open, .write = ks_proc_reg_write, .read = seq_read, .llseek = seq_lseek, .release = single_release,}},
  { "p4_ingressControl4",     (void*)KSZ8864_SMI_PORT4_OFFSET+KSZ8864_SMI_PORTX_INGRESS_CONTROL_4 ,  { .owner = THIS_MODULE, .open = ks_proc_reg_open, .write = ks_proc_reg_write, .read = seq_read, .llseek = seq_lseek, .release = single_release,}},
  { "p4_egressControl1",      (void*)KSZ8864_SMI_PORT4_OFFSET+KSZ8864_SMI_PORTX_EGRESS_CONTROL_1 ,  { .owner = THIS_MODULE, .open = ks_proc_reg_open, .write = ks_proc_reg_write, .read = seq_read, .llseek = seq_lseek, .release = single_release,}},
  { "p4_egressControl2",      (void*)KSZ8864_SMI_PORT4_OFFSET+KSZ8864_SMI_PORTX_EGRESS_CONTROL_2 ,  { .owner = THIS_MODULE, .open = ks_proc_reg_open, .write = ks_proc_reg_write, .read = seq_read, .llseek = seq_lseek, .release = single_release,}},
  { "p4_egressControl3",      (void*)KSZ8864_SMI_PORT4_OFFSET+KSZ8864_SMI_PORTX_EGRESS_CONTROL_3 ,  { .owner = THIS_MODULE, .open = ks_proc_reg_open, .write = ks_proc_reg_write, .read = seq_read, .llseek = seq_lseek, .release = single_release,}},
  { "p4_egressControl4",      (void*)KSZ8864_SMI_PORT4_OFFSET+KSZ8864_SMI_PORTX_EGRESS_CONTROL_4 ,  { .owner = THIS_MODULE, .open = ks_proc_reg_open, .write = ks_proc_reg_write, .read = seq_read, .llseek = seq_lseek, .release = single_release,}},
  { "p4_MiiControl",          (void*)KSZ8864_SMI_PORT4_MII_CONTROL ,  { .owner = THIS_MODULE, .open = ks_proc_reg_open, .write = ks_proc_reg_write, .read = seq_read, .llseek = seq_lseek, .release = single_release,}},
};
static struct ks_proc_entry ks_proc_mii_entries[] = {
  { "p1_control",             MIIDATA(KSZ8864_MII_PORT1_ADDR, KSZ8864_MII_CONTROL),  { .owner = THIS_MODULE, .open = ks_proc_mii_open, .write = ks_proc_mii_write, .read = seq_read, .llseek = seq_lseek, .release = single_release,}, },
  { "p1_status",              MIIDATA(KSZ8864_MII_PORT1_ADDR, KSZ8864_MII_STATUS),  { .owner = THIS_MODULE, .open = ks_proc_mii_open, .write = NULL, .read = seq_read, .llseek = seq_lseek, .release = single_release,}, },
  { "p1_phyIdHi",             MIIDATA(KSZ8864_MII_PORT1_ADDR, KSZ8864_MII_PHYID_HI) ,  { .owner = THIS_MODULE, .open = ks_proc_mii_open, .write = NULL, .read = seq_read, .llseek = seq_lseek, .release = single_release,}, },
  { "p1_phyIdLo",             MIIDATA(KSZ8864_MII_PORT1_ADDR, KSZ8864_MII_PHYID_LO) ,  { .owner = THIS_MODULE, .open = ks_proc_mii_open, .write = NULL, .read = seq_read, .llseek = seq_lseek, .release = single_release,}, },
  { "p1_anAdvertise",         MIIDATA(KSZ8864_MII_PORT1_ADDR, KSZ8864_MII_ADVERTISE) ,  { .owner = THIS_MODULE, .open = ks_proc_mii_open, .write = ks_proc_mii_write, .read = seq_read, .llseek = seq_lseek, .release = single_release,}, },
  { "p1_anPartnerAbility",    MIIDATA(KSZ8864_MII_PORT1_ADDR, KSZ8864_MII_PARTNER_ABILITY) ,  { .owner = THIS_MODULE, .open = ks_proc_mii_open, .write = NULL, .read = seq_read, .llseek = seq_lseek, .release = single_release,}, },
  { "p1_specialCtrlStat",     MIIDATA(KSZ8864_MII_PORT1_ADDR, KSZ8864_MII_SPECIAL_CONTROL_STATUS)  ,  { .owner = THIS_MODULE, .open = ks_proc_mii_open, .write = ks_proc_mii_write, .read = seq_read, .llseek = seq_lseek, .release = single_release,}, },

  { "p2_control",             MIIDATA(KSZ8864_MII_PORT2_ADDR, KSZ8864_MII_CONTROL) ,  { .owner = THIS_MODULE, .open = ks_proc_mii_open, .write = ks_proc_mii_write, .read = seq_read, .llseek = seq_lseek, .release = single_release,},  },
  { "p2_status",              MIIDATA(KSZ8864_MII_PORT2_ADDR, KSZ8864_MII_STATUS) ,  { .owner = THIS_MODULE, .open = ks_proc_mii_open, .write = NULL, .read = seq_read, .llseek = seq_lseek, .release = single_release,}, },
  { "p2_phyIdHi",             MIIDATA(KSZ8864_MII_PORT2_ADDR, KSZ8864_MII_PHYID_HI) ,  { .owner = THIS_MODULE, .open = ks_proc_mii_open, .write = NULL, .read = seq_read, .llseek = seq_lseek, .release = single_release,}, },
  { "p2_phyIdLo",             MIIDATA(KSZ8864_MII_PORT2_ADDR, KSZ8864_MII_PHYID_LO) ,  { .owner = THIS_MODULE, .open = ks_proc_mii_open, .write = NULL, .read = seq_read, .llseek = seq_lseek, .release = single_release,}, },
  { "p2_anAdversize",         MIIDATA(KSZ8864_MII_PORT2_ADDR, KSZ8864_MII_ADVERTISE) ,  { .owner = THIS_MODULE, .open = ks_proc_mii_open, .write = ks_proc_mii_write, .read = seq_read, .llseek = seq_lseek, .release = single_release,},  },
  { "p2_anPartnerAbility",    MIIDATA(KSZ8864_MII_PORT2_ADDR, KSZ8864_MII_PARTNER_ABILITY) ,  { .owner = THIS_MODULE, .open = ks_proc_mii_open, .write = NULL, .read = seq_read, .llseek = seq_lseek, .release = single_release,}, },
  { "p2_specialCtrlStat",     MIIDATA(KSZ8864_MII_PORT2_ADDR, KSZ8864_MII_SPECIAL_CONTROL_STATUS)  ,  { .owner = THIS_MODULE, .open = ks_proc_mii_open, .write = ks_proc_mii_write, .read = seq_read, .llseek = seq_lseek, .release = single_release,}, },

  { "p3_control",             MIIDATA(PHY_ADDR, KSZ8051_BASIC_CONTROL) ,  { .owner = THIS_MODULE, .open = ks_proc_mii_open, .write = ks_proc_mii_write, .read = seq_read, .llseek = seq_lseek, .release = single_release,},  },
  { "p3_status",              MIIDATA(PHY_ADDR, KSZ8051_BASIC_STATUS) ,  { .owner = THIS_MODULE, .open = ks_proc_mii_open, .write = NULL, .read = seq_read, .llseek = seq_lseek, .release = single_release,}, },
  { "p3_phyIdHi",             MIIDATA(PHY_ADDR, KSZ8051_PHY_ID_1) ,  { .owner = THIS_MODULE, .open = ks_proc_mii_open, .write = NULL, .read = seq_read, .llseek = seq_lseek, .release = single_release,}, },
  { "p3_phyIdLo",             MIIDATA(PHY_ADDR, KSZ8051_PHY_ID_2) ,  { .owner = THIS_MODULE, .open = ks_proc_mii_open, .write = NULL, .read = seq_read, .llseek = seq_lseek, .release = single_release,}, },
  { "p3_anAdvertise",         MIIDATA(PHY_ADDR, KSZ8051_AUTONEG_AVERTISE) ,  { .owner = THIS_MODULE, .open = ks_proc_mii_open, .write = ks_proc_mii_write, .read = seq_read, .llseek = seq_lseek, .release = single_release,},  },
  { "p3_anPartnerAbility",    MIIDATA(PHY_ADDR, KSZ8051_AUTONEG_LINK_PARTNER) ,  { .owner = THIS_MODULE, .open = ks_proc_mii_open, .write = NULL, .read = seq_read, .llseek = seq_lseek, .release = single_release,}, },
  { "p3_anExpansion",         MIIDATA(PHY_ADDR, KSZ8051_AUTONEG_EXPANSION) ,  { .owner = THIS_MODULE, .open = ks_proc_mii_open, .write = NULL, .read = seq_read, .llseek = seq_lseek, .release = single_release,}, },
  { "p3_anNextPage",          MIIDATA(PHY_ADDR, KSZ8051_AUTONEG_NEXT_PAGE) ,  { .owner = THIS_MODULE, .open = ks_proc_mii_open, .write = ks_proc_mii_write, .read = seq_read, .llseek = seq_lseek, .release = single_release,},  },
  { "p3_anPartnerNextPage",   MIIDATA(PHY_ADDR, KSZ8051_AUTONEG_LINK_PARTNER_NEXT_PAGE) ,  { .owner = THIS_MODULE, .open = ks_proc_mii_open, .write = NULL, .read = seq_read, .llseek = seq_lseek, .release = single_release,}, },
  { "p3_afeControl1",         MIIDATA(PHY_ADDR, KSZ8051_AFE_CONTROL_1) ,  { .owner = THIS_MODULE, .open = ks_proc_mii_open, .write = ks_proc_mii_write, .read = seq_read, .llseek = seq_lseek, .release = single_release,},  },
  { "p3_rxErrorCount",        MIIDATA(PHY_ADDR, KSZ8051_RXER_COUNTER) ,  { .owner = THIS_MODULE, .open = ks_proc_mii_open, .write = ks_proc_mii_write, .read = seq_read, .llseek = seq_lseek, .release = single_release,},  },
  { "p3_modeStrapOverride",   MIIDATA(PHY_ADDR, KSZ8051_MODESTRAP_OVERRIDE) ,  { .owner = THIS_MODULE, .open = ks_proc_mii_open, .write = ks_proc_mii_write, .read = seq_read, .llseek = seq_lseek, .release = single_release,},  },
  { "p3_modeStrapStatus",     MIIDATA(PHY_ADDR, KSZ8051_MODESTRAP_STATUS) ,  { .owner = THIS_MODULE, .open = ks_proc_mii_open, .write = NULL, .read = seq_read, .llseek = seq_lseek, .release = single_release,}, },
  { "p3_expandedControl",     MIIDATA(PHY_ADDR, KSZ8051_EXPANDED_CONTROL) ,  { .owner = THIS_MODULE, .open = ks_proc_mii_open, .write = ks_proc_mii_write, .read = seq_read, .llseek = seq_lseek, .release = single_release,},  },
  { "p3_irqControlStatus",    MIIDATA(PHY_ADDR, KSZ8051_IRQ_STATUS_CONTROL)  ,  { .owner = THIS_MODULE, .open = ks_proc_mii_open, .write = ks_proc_mii_write, .read = seq_read, .llseek = seq_lseek, .release = single_release,}, },
  { "p3_linkMdControlStatus", MIIDATA(PHY_ADDR, KSZ8051_LINKMD_CONTROL_STATUS)  ,  { .owner = THIS_MODULE, .open = ks_proc_mii_open, .write = ks_proc_mii_write, .read = seq_read, .llseek = seq_lseek, .release = single_release,}, },
  { "p3_phyControl1",         MIIDATA(PHY_ADDR, KSZ8051_PHY_CONTROL_1)  ,  { .owner = THIS_MODULE, .open = ks_proc_mii_open, .write = ks_proc_mii_write, .read = seq_read, .llseek = seq_lseek, .release = single_release,}, },
  { "p3_phyControl2",         MIIDATA(PHY_ADDR, KSZ8051_PHY_CONTROL_2) ,  { .owner = THIS_MODULE, .open = ks_proc_mii_open, .write = ks_proc_mii_write, .read = seq_read, .llseek = seq_lseek, .release = single_release,},  },
};

 static struct ks_proc_entry ks_proc_ctl_entries[] = {
  { "reset",                  NULL ,  { .owner = THIS_MODULE, .open = ks_proc_ctl_reset_open, .write = ks_proc_ctl_reset_write, .read = seq_read, .llseek = seq_lseek, .release = single_release,}, },
  { "flushDynMACs",           NULL ,  { .owner = THIS_MODULE, .open = ks_proc_ctl_fldmac_open, .write = ks_proc_ctl_fldmac_write, .read = seq_read, .llseek = seq_lseek, .release = single_release,}, },
  { "tailTag",                NULL ,  { .owner = THIS_MODULE, .open = ks_proc_ctl_tailtag_open, .write = ks_proc_ctl_tailtag_write, .read = seq_read, .llseek = seq_lseek, .release = single_release,}, },
};

static struct ks_proc_entry ks_proc_mib_entries[] = {
  { "p1_rxLoPriorityBs",      (void*)KSZ8864_MIB_PORT1_OFFSET+KSZ8864_MIB_PORTX_RX_LO_PRIORITY_BYTES,  { .owner = THIS_MODULE, .open = ks_proc_mib_open, .write = NULL, .read = seq_read, .llseek = seq_lseek, .release = single_release,}, },
  { "p1_rxHiPriorityBs",      (void*)KSZ8864_MIB_PORT1_OFFSET+KSZ8864_MIB_PORTX_RX_HI_PRIORITY_BYTES,  { .owner = THIS_MODULE, .open = ks_proc_mib_open, .write = NULL, .read = seq_read, .llseek = seq_lseek, .release = single_release,}, },
  { "p1_rxUndersizePkts",     (void*)KSZ8864_MIB_PORT1_OFFSET+KSZ8864_MIB_PORTX_RX_UNDERSIZE_PACKETS,  { .owner = THIS_MODULE, .open = ks_proc_mib_open, .write = NULL, .read = seq_read, .llseek = seq_lseek, .release = single_release,}, },
  { "p1_rxFragments",         (void*)KSZ8864_MIB_PORT1_OFFSET+KSZ8864_MIB_PORTX_RX_FRAGMENTS,  { .owner = THIS_MODULE, .open = ks_proc_mib_open, .write = NULL, .read = seq_read, .llseek = seq_lseek, .release = single_release,}, },
  { "p1_rxOversizePkts",      (void*)KSZ8864_MIB_PORT1_OFFSET+KSZ8864_MIB_PORTX_RX_OVERSIZE_PACKETS,  { .owner = THIS_MODULE, .open = ks_proc_mib_open, .write = NULL, .read = seq_read, .llseek = seq_lseek, .release = single_release,}, },
  { "p1_rxJabbers",           (void*)KSZ8864_MIB_PORT1_OFFSET+KSZ8864_MIB_PORTX_RX_JABBERS,  { .owner = THIS_MODULE, .open = ks_proc_mib_open, .write = NULL, .read = seq_read, .llseek = seq_lseek, .release = single_release,}, },
  { "p1_rxSymbolErrs",        (void*)KSZ8864_MIB_PORT1_OFFSET+KSZ8864_MIB_PORTX_RX_SYMBOL_ERRORS,  { .owner = THIS_MODULE, .open = ks_proc_mib_open, .write = NULL, .read = seq_read, .llseek = seq_lseek, .release = single_release,}, },
  { "p1_rxCrcErrs",           (void*)KSZ8864_MIB_PORT1_OFFSET+KSZ8864_MIB_PORTX_RX_CRC_ERRORS,  { .owner = THIS_MODULE, .open = ks_proc_mib_open, .write = NULL, .read = seq_read, .llseek = seq_lseek, .release = single_release,}, },
  { "p1_rxAlignmentErrs",     (void*)KSZ8864_MIB_PORT1_OFFSET+KSZ8864_MIB_PORTX_RX_ALIGNMENT_ERRORS,  { .owner = THIS_MODULE, .open = ks_proc_mib_open, .write = NULL, .read = seq_read, .llseek = seq_lseek, .release = single_release,}, },
  { "p1_rxControl8808Pkts",   (void*)KSZ8864_MIB_PORT1_OFFSET+KSZ8864_MIB_PORTX_RX_CONTROL_8808_PACKETS,  { .owner = THIS_MODULE, .open = ks_proc_mib_open, .write = NULL, .read = seq_read, .llseek = seq_lseek, .release = single_release,}, },
  { "p1_rxPausePkts",         (void*)KSZ8864_MIB_PORT1_OFFSET+KSZ8864_MIB_PORTX_RX_PAUSE_PACKETS,  { .owner = THIS_MODULE, .open = ks_proc_mib_open, .write = NULL, .read = seq_read, .llseek = seq_lseek, .release = single_release,}, },
  { "p1_rxBroadcastPkts",     (void*)KSZ8864_MIB_PORT1_OFFSET+KSZ8864_MIB_PORTX_RX_BROADCAST_PACKETS,  { .owner = THIS_MODULE, .open = ks_proc_mib_open, .write = NULL, .read = seq_read, .llseek = seq_lseek, .release = single_release,}, },
  { "p1_rxMulticastPkts",     (void*)KSZ8864_MIB_PORT1_OFFSET+KSZ8864_MIB_PORTX_RX_MULTICAST_PACKETS,  { .owner = THIS_MODULE, .open = ks_proc_mib_open, .write = NULL, .read = seq_read, .llseek = seq_lseek, .release = single_release,}, },
  { "p1_rxUnicastPkts",       (void*)KSZ8864_MIB_PORT1_OFFSET+KSZ8864_MIB_PORTX_RX_UNICAST_PACKETS,  { .owner = THIS_MODULE, .open = ks_proc_mib_open, .write = NULL, .read = seq_read, .llseek = seq_lseek, .release = single_release,}, },
  { "p1_rx64BsPkts",          (void*)KSZ8864_MIB_PORT1_OFFSET+KSZ8864_MIB_PORTX_RX_64_BYTE_PACKETS,  { .owner = THIS_MODULE, .open = ks_proc_mib_open, .write = NULL, .read = seq_read, .llseek = seq_lseek, .release = single_release,}, },
  { "p1_rx65-127BsPkts",      (void*)KSZ8864_MIB_PORT1_OFFSET+KSZ8864_MIB_PORTX_RX_65_TO_127_BYTE_PACKETS,  { .owner = THIS_MODULE, .open = ks_proc_mib_open, .write = NULL, .read = seq_read, .llseek = seq_lseek, .release = single_release,}, },
  { "p1_rx128-255BsPkts",     (void*)KSZ8864_MIB_PORT1_OFFSET+KSZ8864_MIB_PORTX_RX_128_TO_255_BYTE_PACKETS,  { .owner = THIS_MODULE, .open = ks_proc_mib_open, .write = NULL, .read = seq_read, .llseek = seq_lseek, .release = single_release,}, },
  { "p1_rx256-511BsPkts",     (void*)KSZ8864_MIB_PORT1_OFFSET+KSZ8864_MIB_PORTX_RX_256_TO_511_BYTE_PACKETS,  { .owner = THIS_MODULE, .open = ks_proc_mib_open, .write = NULL, .read = seq_read, .llseek = seq_lseek, .release = single_release,}, },
  { "p1_rx512-1023BsPkts",    (void*)KSZ8864_MIB_PORT1_OFFSET+KSZ8864_MIB_PORTX_RX_512_TO_1023_BYTE_PACKETS,  { .owner = THIS_MODULE, .open = ks_proc_mib_open, .write = NULL, .read = seq_read, .llseek = seq_lseek, .release = single_release,}, },
  { "p1_rx1024-1522BsPkts",   (void*)KSZ8864_MIB_PORT1_OFFSET+KSZ8864_MIB_PORTX_RX_1024_TO_1522_BYTE_PACKETS,  { .owner = THIS_MODULE, .open = ks_proc_mib_open, .write = NULL, .read = seq_read, .llseek = seq_lseek, .release = single_release,}, },
  { "p1_txLoPriorityBs",      (void*)KSZ8864_MIB_PORT1_OFFSET+KSZ8864_MIB_PORTX_TX_LO_PRIORITY_BYTES,  { .owner = THIS_MODULE, .open = ks_proc_mib_open, .write = NULL, .read = seq_read, .llseek = seq_lseek, .release = single_release,}, },
  { "p1_txHiPriorityBs",      (void*)KSZ8864_MIB_PORT1_OFFSET+KSZ8864_MIB_PORTX_TX_HI_PRIORITY_BYTES,  { .owner = THIS_MODULE, .open = ks_proc_mib_open, .write = NULL, .read = seq_read, .llseek = seq_lseek, .release = single_release,}, },
  { "p1_txLateCols",          (void*)KSZ8864_MIB_PORT1_OFFSET+KSZ8864_MIB_PORTX_TX_LATE_COLLISIONS,  { .owner = THIS_MODULE, .open = ks_proc_mib_open, .write = NULL, .read = seq_read, .llseek = seq_lseek, .release = single_release,}, },
  { "p1_txPausePkts",         (void*)KSZ8864_MIB_PORT1_OFFSET+KSZ8864_MIB_PORTX_TX_PAUSE_PACKETS,  { .owner = THIS_MODULE, .open = ks_proc_mib_open, .write = NULL, .read = seq_read, .llseek = seq_lseek, .release = single_release,}, },
  { "p1_txBroadcastPkts",     (void*)KSZ8864_MIB_PORT1_OFFSET+KSZ8864_MIB_PORTX_TX_BROADCAST_PACKETS,  { .owner = THIS_MODULE, .open = ks_proc_mib_open, .write = NULL, .read = seq_read, .llseek = seq_lseek, .release = single_release,}, },
  { "p1_txMulticastPkts",     (void*)KSZ8864_MIB_PORT1_OFFSET+KSZ8864_MIB_PORTX_TX_MULTICAST_PACKETS,  { .owner = THIS_MODULE, .open = ks_proc_mib_open, .write = NULL, .read = seq_read, .llseek = seq_lseek, .release = single_release,}, },
  { "p1_txUnicastPkts",       (void*)KSZ8864_MIB_PORT1_OFFSET+KSZ8864_MIB_PORTX_TX_UNICAST_PACKETS,  { .owner = THIS_MODULE, .open = ks_proc_mib_open, .write = NULL, .read = seq_read, .llseek = seq_lseek, .release = single_release,}, },
  { "p1_txDeferredPkts",      (void*)KSZ8864_MIB_PORT1_OFFSET+KSZ8864_MIB_PORTX_TX_DEFERRED_PACKETS,  { .owner = THIS_MODULE, .open = ks_proc_mib_open, .write = NULL, .read = seq_read, .llseek = seq_lseek, .release = single_release,}, },
  { "p1_txTotalCols",         (void*)KSZ8864_MIB_PORT1_OFFSET+KSZ8864_MIB_PORTX_TX_TOTAL_COLLISIONS,  { .owner = THIS_MODULE, .open = ks_proc_mib_open, .write = NULL, .read = seq_read, .llseek = seq_lseek, .release = single_release,}, },
  { "p1_txExcessiveCols",     (void*)KSZ8864_MIB_PORT1_OFFSET+KSZ8864_MIB_PORTX_TX_EXCESSIVE_COLLISIONS,  { .owner = THIS_MODULE, .open = ks_proc_mib_open, .write = NULL, .read = seq_read, .llseek = seq_lseek, .release = single_release,}, },
  { "p1_txSingleCols",        (void*)KSZ8864_MIB_PORT1_OFFSET+KSZ8864_MIB_PORTX_TX_SINGLE_COLLISIONS,  { .owner = THIS_MODULE, .open = ks_proc_mib_open, .write = NULL, .read = seq_read, .llseek = seq_lseek, .release = single_release,}, },
  { "p1_txMultipleCols",      (void*)KSZ8864_MIB_PORT1_OFFSET+KSZ8864_MIB_PORTX_TX_MULTIPLE_COLLISIONS,  { .owner = THIS_MODULE, .open = ks_proc_mib_open, .write = NULL, .read = seq_read, .llseek = seq_lseek, .release = single_release,}, },

  { "p2_rxLoPriorityBs",      (void*)KSZ8864_MIB_PORT2_OFFSET+KSZ8864_MIB_PORTX_RX_LO_PRIORITY_BYTES,  { .owner = THIS_MODULE, .open = ks_proc_mib_open, .write = NULL, .read = seq_read, .llseek = seq_lseek, .release = single_release,}, },
  { "p2_rxHiPriorityBs",      (void*)KSZ8864_MIB_PORT2_OFFSET+KSZ8864_MIB_PORTX_RX_HI_PRIORITY_BYTES,  { .owner = THIS_MODULE, .open = ks_proc_mib_open, .write = NULL, .read = seq_read, .llseek = seq_lseek, .release = single_release,}, },
  { "p2_rxUndersizePkts",     (void*)KSZ8864_MIB_PORT2_OFFSET+KSZ8864_MIB_PORTX_RX_UNDERSIZE_PACKETS,  { .owner = THIS_MODULE, .open = ks_proc_mib_open, .write = NULL, .read = seq_read, .llseek = seq_lseek, .release = single_release,}, },
  { "p2_rxFragments",         (void*)KSZ8864_MIB_PORT2_OFFSET+KSZ8864_MIB_PORTX_RX_FRAGMENTS,  { .owner = THIS_MODULE, .open = ks_proc_mib_open, .write = NULL, .read = seq_read, .llseek = seq_lseek, .release = single_release,}, },
  { "p2_rxOversizePkts",      (void*)KSZ8864_MIB_PORT2_OFFSET+KSZ8864_MIB_PORTX_RX_OVERSIZE_PACKETS,  { .owner = THIS_MODULE, .open = ks_proc_mib_open, .write = NULL, .read = seq_read, .llseek = seq_lseek, .release = single_release,}, },
  { "p2_rxJabbers",           (void*)KSZ8864_MIB_PORT2_OFFSET+KSZ8864_MIB_PORTX_RX_JABBERS,  { .owner = THIS_MODULE, .open = ks_proc_mib_open, .write = NULL, .read = seq_read, .llseek = seq_lseek, .release = single_release,}, },
  { "p2_rxSymbolErrs",        (void*)KSZ8864_MIB_PORT2_OFFSET+KSZ8864_MIB_PORTX_RX_SYMBOL_ERRORS,  { .owner = THIS_MODULE, .open = ks_proc_mib_open, .write = NULL, .read = seq_read, .llseek = seq_lseek, .release = single_release,}, },
  { "p2_rxCrcErrs",           (void*)KSZ8864_MIB_PORT2_OFFSET+KSZ8864_MIB_PORTX_RX_CRC_ERRORS,  { .owner = THIS_MODULE, .open = ks_proc_mib_open, .write = NULL, .read = seq_read, .llseek = seq_lseek, .release = single_release,}, },
  { "p2_rxAlignmentErrs",     (void*)KSZ8864_MIB_PORT2_OFFSET+KSZ8864_MIB_PORTX_RX_ALIGNMENT_ERRORS,  { .owner = THIS_MODULE, .open = ks_proc_mib_open, .write = NULL, .read = seq_read, .llseek = seq_lseek, .release = single_release,}, },
  { "p2_rxControl8808Pkts",   (void*)KSZ8864_MIB_PORT2_OFFSET+KSZ8864_MIB_PORTX_RX_CONTROL_8808_PACKETS,  { .owner = THIS_MODULE, .open = ks_proc_mib_open, .write = NULL, .read = seq_read, .llseek = seq_lseek, .release = single_release,}, },
  { "p2_rxPausePkts",         (void*)KSZ8864_MIB_PORT2_OFFSET+KSZ8864_MIB_PORTX_RX_PAUSE_PACKETS,  { .owner = THIS_MODULE, .open = ks_proc_mib_open, .write = NULL, .read = seq_read, .llseek = seq_lseek, .release = single_release,}, },
  { "p2_rxBroadcastPkts",     (void*)KSZ8864_MIB_PORT2_OFFSET+KSZ8864_MIB_PORTX_RX_BROADCAST_PACKETS,  { .owner = THIS_MODULE, .open = ks_proc_mib_open, .write = NULL, .read = seq_read, .llseek = seq_lseek, .release = single_release,}, },
  { "p2_rxMulticastPkts",     (void*)KSZ8864_MIB_PORT2_OFFSET+KSZ8864_MIB_PORTX_RX_MULTICAST_PACKETS,  { .owner = THIS_MODULE, .open = ks_proc_mib_open, .write = NULL, .read = seq_read, .llseek = seq_lseek, .release = single_release,}, },
  { "p2_rxUnicastPkts",       (void*)KSZ8864_MIB_PORT2_OFFSET+KSZ8864_MIB_PORTX_RX_UNICAST_PACKETS,  { .owner = THIS_MODULE, .open = ks_proc_mib_open, .write = NULL, .read = seq_read, .llseek = seq_lseek, .release = single_release,}, },
  { "p2_rx64BsPkts",          (void*)KSZ8864_MIB_PORT2_OFFSET+KSZ8864_MIB_PORTX_RX_64_BYTE_PACKETS,  { .owner = THIS_MODULE, .open = ks_proc_mib_open, .write = NULL, .read = seq_read, .llseek = seq_lseek, .release = single_release,}, },
  { "p2_rx65-127BsPkts",      (void*)KSZ8864_MIB_PORT2_OFFSET+KSZ8864_MIB_PORTX_RX_65_TO_127_BYTE_PACKETS,  { .owner = THIS_MODULE, .open = ks_proc_mib_open, .write = NULL, .read = seq_read, .llseek = seq_lseek, .release = single_release,}, },
  { "p2_rx128-255BsPkts",     (void*)KSZ8864_MIB_PORT2_OFFSET+KSZ8864_MIB_PORTX_RX_128_TO_255_BYTE_PACKETS,  { .owner = THIS_MODULE, .open = ks_proc_mib_open, .write = NULL, .read = seq_read, .llseek = seq_lseek, .release = single_release,}, },
  { "p2_rx256-511BsPkts",     (void*)KSZ8864_MIB_PORT2_OFFSET+KSZ8864_MIB_PORTX_RX_256_TO_511_BYTE_PACKETS,  { .owner = THIS_MODULE, .open = ks_proc_mib_open, .write = NULL, .read = seq_read, .llseek = seq_lseek, .release = single_release,}, },
  { "p2_rx512-1023BsPkts",    (void*)KSZ8864_MIB_PORT2_OFFSET+KSZ8864_MIB_PORTX_RX_512_TO_1023_BYTE_PACKETS,  { .owner = THIS_MODULE, .open = ks_proc_mib_open, .write = NULL, .read = seq_read, .llseek = seq_lseek, .release = single_release,}, },
  { "p2_rx1024-1522BsPkts",   (void*)KSZ8864_MIB_PORT2_OFFSET+KSZ8864_MIB_PORTX_RX_1024_TO_1522_BYTE_PACKETS,  { .owner = THIS_MODULE, .open = ks_proc_mib_open, .write = NULL, .read = seq_read, .llseek = seq_lseek, .release = single_release,}, },
  { "p2_txLoPriorityBs",      (void*)KSZ8864_MIB_PORT2_OFFSET+KSZ8864_MIB_PORTX_TX_LO_PRIORITY_BYTES,  { .owner = THIS_MODULE, .open = ks_proc_mib_open, .write = NULL, .read = seq_read, .llseek = seq_lseek, .release = single_release,}, },
  { "p2_txHiPriorityBs",      (void*)KSZ8864_MIB_PORT2_OFFSET+KSZ8864_MIB_PORTX_TX_HI_PRIORITY_BYTES,  { .owner = THIS_MODULE, .open = ks_proc_mib_open, .write = NULL, .read = seq_read, .llseek = seq_lseek, .release = single_release,}, },
  { "p2_txLateCols",          (void*)KSZ8864_MIB_PORT2_OFFSET+KSZ8864_MIB_PORTX_TX_LATE_COLLISIONS,  { .owner = THIS_MODULE, .open = ks_proc_mib_open, .write = NULL, .read = seq_read, .llseek = seq_lseek, .release = single_release,}, },
  { "p2_txPausePkts",         (void*)KSZ8864_MIB_PORT2_OFFSET+KSZ8864_MIB_PORTX_TX_PAUSE_PACKETS,  { .owner = THIS_MODULE, .open = ks_proc_mib_open, .write = NULL, .read = seq_read, .llseek = seq_lseek, .release = single_release,}, },
  { "p2_txBroadcastPkts",     (void*)KSZ8864_MIB_PORT2_OFFSET+KSZ8864_MIB_PORTX_TX_BROADCAST_PACKETS,  { .owner = THIS_MODULE, .open = ks_proc_mib_open, .write = NULL, .read = seq_read, .llseek = seq_lseek, .release = single_release,}, },
  { "p2_txMulticastPkts",     (void*)KSZ8864_MIB_PORT2_OFFSET+KSZ8864_MIB_PORTX_TX_MULTICAST_PACKETS,  { .owner = THIS_MODULE, .open = ks_proc_mib_open, .write = NULL, .read = seq_read, .llseek = seq_lseek, .release = single_release,}, },
  { "p2_txUnicastPkts",       (void*)KSZ8864_MIB_PORT2_OFFSET+KSZ8864_MIB_PORTX_TX_UNICAST_PACKETS,  { .owner = THIS_MODULE, .open = ks_proc_mib_open, .write = NULL, .read = seq_read, .llseek = seq_lseek, .release = single_release,}, },
  { "p2_txDeferredPkts",      (void*)KSZ8864_MIB_PORT2_OFFSET+KSZ8864_MIB_PORTX_TX_DEFERRED_PACKETS,  { .owner = THIS_MODULE, .open = ks_proc_mib_open, .write = NULL, .read = seq_read, .llseek = seq_lseek, .release = single_release,}, },
  { "p2_txTotalCols",         (void*)KSZ8864_MIB_PORT2_OFFSET+KSZ8864_MIB_PORTX_TX_TOTAL_COLLISIONS,  { .owner = THIS_MODULE, .open = ks_proc_mib_open, .write = NULL, .read = seq_read, .llseek = seq_lseek, .release = single_release,}, },
  { "p2_txExcessiveCols",     (void*)KSZ8864_MIB_PORT2_OFFSET+KSZ8864_MIB_PORTX_TX_EXCESSIVE_COLLISIONS,  { .owner = THIS_MODULE, .open = ks_proc_mib_open, .write = NULL, .read = seq_read, .llseek = seq_lseek, .release = single_release,}, },
  { "p2_txSingleCols",        (void*)KSZ8864_MIB_PORT2_OFFSET+KSZ8864_MIB_PORTX_TX_SINGLE_COLLISIONS,  { .owner = THIS_MODULE, .open = ks_proc_mib_open, .write = NULL, .read = seq_read, .llseek = seq_lseek, .release = single_release,}, },
  { "p2_txMultipleCols",      (void*)KSZ8864_MIB_PORT2_OFFSET+KSZ8864_MIB_PORTX_TX_MULTIPLE_COLLISIONS,  { .owner = THIS_MODULE, .open = ks_proc_mib_open, .write = NULL, .read = seq_read, .llseek = seq_lseek, .release = single_release,}, },

  { "p3_rxLoPriorityBs",      (void*)KSZ8864_MIB_PORT3_OFFSET+KSZ8864_MIB_PORTX_RX_LO_PRIORITY_BYTES,  { .owner = THIS_MODULE, .open = ks_proc_mib_open, .write = NULL, .read = seq_read, .llseek = seq_lseek, .release = single_release,}, },
  { "p3_rxHiPriorityBs",      (void*)KSZ8864_MIB_PORT3_OFFSET+KSZ8864_MIB_PORTX_RX_HI_PRIORITY_BYTES,  { .owner = THIS_MODULE, .open = ks_proc_mib_open, .write = NULL, .read = seq_read, .llseek = seq_lseek, .release = single_release,}, },
  { "p3_rxUndersizePkts",     (void*)KSZ8864_MIB_PORT3_OFFSET+KSZ8864_MIB_PORTX_RX_UNDERSIZE_PACKETS,  { .owner = THIS_MODULE, .open = ks_proc_mib_open, .write = NULL, .read = seq_read, .llseek = seq_lseek, .release = single_release,}, },
  { "p3_rxFragments",         (void*)KSZ8864_MIB_PORT3_OFFSET+KSZ8864_MIB_PORTX_RX_FRAGMENTS,  { .owner = THIS_MODULE, .open = ks_proc_mib_open, .write = NULL, .read = seq_read, .llseek = seq_lseek, .release = single_release,}, },
  { "p3_rxOversizePkts",      (void*)KSZ8864_MIB_PORT3_OFFSET+KSZ8864_MIB_PORTX_RX_OVERSIZE_PACKETS,  { .owner = THIS_MODULE, .open = ks_proc_mib_open, .write = NULL, .read = seq_read, .llseek = seq_lseek, .release = single_release,}, },
  { "p3_rxJabbers",           (void*)KSZ8864_MIB_PORT3_OFFSET+KSZ8864_MIB_PORTX_RX_JABBERS,  { .owner = THIS_MODULE, .open = ks_proc_mib_open, .write = NULL, .read = seq_read, .llseek = seq_lseek, .release = single_release,}, },
  { "p3_rxSymbolErrs",        (void*)KSZ8864_MIB_PORT3_OFFSET+KSZ8864_MIB_PORTX_RX_SYMBOL_ERRORS,  { .owner = THIS_MODULE, .open = ks_proc_mib_open, .write = NULL, .read = seq_read, .llseek = seq_lseek, .release = single_release,}, },
  { "p3_rxCrcErrs",           (void*)KSZ8864_MIB_PORT3_OFFSET+KSZ8864_MIB_PORTX_RX_CRC_ERRORS,  { .owner = THIS_MODULE, .open = ks_proc_mib_open, .write = NULL, .read = seq_read, .llseek = seq_lseek, .release = single_release,}, },
  { "p3_rxAlignmentErrs",     (void*)KSZ8864_MIB_PORT3_OFFSET+KSZ8864_MIB_PORTX_RX_ALIGNMENT_ERRORS,  { .owner = THIS_MODULE, .open = ks_proc_mib_open, .write = NULL, .read = seq_read, .llseek = seq_lseek, .release = single_release,}, },
  { "p3_rxControl8808Pkts",   (void*)KSZ8864_MIB_PORT3_OFFSET+KSZ8864_MIB_PORTX_RX_CONTROL_8808_PACKETS,  { .owner = THIS_MODULE, .open = ks_proc_mib_open, .write = NULL, .read = seq_read, .llseek = seq_lseek, .release = single_release,}, },
  { "p3_rxPausePkts",         (void*)KSZ8864_MIB_PORT3_OFFSET+KSZ8864_MIB_PORTX_RX_PAUSE_PACKETS,  { .owner = THIS_MODULE, .open = ks_proc_mib_open, .write = NULL, .read = seq_read, .llseek = seq_lseek, .release = single_release,}, },
  { "p3_rxBroadcastPkts",     (void*)KSZ8864_MIB_PORT3_OFFSET+KSZ8864_MIB_PORTX_RX_BROADCAST_PACKETS,  { .owner = THIS_MODULE, .open = ks_proc_mib_open, .write = NULL, .read = seq_read, .llseek = seq_lseek, .release = single_release,}, },
  { "p3_rxMulticastPkts",     (void*)KSZ8864_MIB_PORT3_OFFSET+KSZ8864_MIB_PORTX_RX_MULTICAST_PACKETS,  { .owner = THIS_MODULE, .open = ks_proc_mib_open, .write = NULL, .read = seq_read, .llseek = seq_lseek, .release = single_release,}, },
  { "p3_rxUnicastPkts",       (void*)KSZ8864_MIB_PORT3_OFFSET+KSZ8864_MIB_PORTX_RX_UNICAST_PACKETS,  { .owner = THIS_MODULE, .open = ks_proc_mib_open, .write = NULL, .read = seq_read, .llseek = seq_lseek, .release = single_release,}, },
  { "p3_rx64BsPkts",          (void*)KSZ8864_MIB_PORT3_OFFSET+KSZ8864_MIB_PORTX_RX_64_BYTE_PACKETS,  { .owner = THIS_MODULE, .open = ks_proc_mib_open, .write = NULL, .read = seq_read, .llseek = seq_lseek, .release = single_release,}, },
  { "p3_rx65-127BsPkts",      (void*)KSZ8864_MIB_PORT3_OFFSET+KSZ8864_MIB_PORTX_RX_65_TO_127_BYTE_PACKETS,  { .owner = THIS_MODULE, .open = ks_proc_mib_open, .write = NULL, .read = seq_read, .llseek = seq_lseek, .release = single_release,}, },
  { "p3_rx128-255BsPkts",     (void*)KSZ8864_MIB_PORT3_OFFSET+KSZ8864_MIB_PORTX_RX_128_TO_255_BYTE_PACKETS,  { .owner = THIS_MODULE, .open = ks_proc_mib_open, .write = NULL, .read = seq_read, .llseek = seq_lseek, .release = single_release,}, },
  { "p3_rx256-511BsPkts",     (void*)KSZ8864_MIB_PORT3_OFFSET+KSZ8864_MIB_PORTX_RX_256_TO_511_BYTE_PACKETS,  { .owner = THIS_MODULE, .open = ks_proc_mib_open, .write = NULL, .read = seq_read, .llseek = seq_lseek, .release = single_release,}, },
  { "p3_rx512-1023BsPkts",    (void*)KSZ8864_MIB_PORT3_OFFSET+KSZ8864_MIB_PORTX_RX_512_TO_1023_BYTE_PACKETS,  { .owner = THIS_MODULE, .open = ks_proc_mib_open, .write = NULL, .read = seq_read, .llseek = seq_lseek, .release = single_release,}, },
  { "p3_rx1024-1522BsPkts",   (void*)KSZ8864_MIB_PORT3_OFFSET+KSZ8864_MIB_PORTX_RX_1024_TO_1522_BYTE_PACKETS,  { .owner = THIS_MODULE, .open = ks_proc_mib_open, .write = NULL, .read = seq_read, .llseek = seq_lseek, .release = single_release,}, },
  { "p3_txLoPriorityBs",      (void*)KSZ8864_MIB_PORT3_OFFSET+KSZ8864_MIB_PORTX_TX_LO_PRIORITY_BYTES,  { .owner = THIS_MODULE, .open = ks_proc_mib_open, .write = NULL, .read = seq_read, .llseek = seq_lseek, .release = single_release,}, },
  { "p3_txHiPriorityBs",      (void*)KSZ8864_MIB_PORT3_OFFSET+KSZ8864_MIB_PORTX_TX_HI_PRIORITY_BYTES,  { .owner = THIS_MODULE, .open = ks_proc_mib_open, .write = NULL, .read = seq_read, .llseek = seq_lseek, .release = single_release,}, },
  { "p3_txLateCols",          (void*)KSZ8864_MIB_PORT3_OFFSET+KSZ8864_MIB_PORTX_TX_LATE_COLLISIONS,  { .owner = THIS_MODULE, .open = ks_proc_mib_open, .write = NULL, .read = seq_read, .llseek = seq_lseek, .release = single_release,}, },
  { "p3_txPausePkts",         (void*)KSZ8864_MIB_PORT3_OFFSET+KSZ8864_MIB_PORTX_TX_PAUSE_PACKETS,  { .owner = THIS_MODULE, .open = ks_proc_mib_open, .write = NULL, .read = seq_read, .llseek = seq_lseek, .release = single_release,}, },
  { "p3_txBroadcastPkts",     (void*)KSZ8864_MIB_PORT3_OFFSET+KSZ8864_MIB_PORTX_TX_BROADCAST_PACKETS,  { .owner = THIS_MODULE, .open = ks_proc_mib_open, .write = NULL, .read = seq_read, .llseek = seq_lseek, .release = single_release,}, },
  { "p3_txMulticastPkts",     (void*)KSZ8864_MIB_PORT3_OFFSET+KSZ8864_MIB_PORTX_TX_MULTICAST_PACKETS,  { .owner = THIS_MODULE, .open = ks_proc_mib_open, .write = NULL, .read = seq_read, .llseek = seq_lseek, .release = single_release,}, },
  { "p3_txUnicastPkts",       (void*)KSZ8864_MIB_PORT3_OFFSET+KSZ8864_MIB_PORTX_TX_UNICAST_PACKETS,  { .owner = THIS_MODULE, .open = ks_proc_mib_open, .write = NULL, .read = seq_read, .llseek = seq_lseek, .release = single_release,}, },
  { "p3_txDeferredPkts",      (void*)KSZ8864_MIB_PORT3_OFFSET+KSZ8864_MIB_PORTX_TX_DEFERRED_PACKETS,  { .owner = THIS_MODULE, .open = ks_proc_mib_open, .write = NULL, .read = seq_read, .llseek = seq_lseek, .release = single_release,}, },
  { "p3_txTotalCols",         (void*)KSZ8864_MIB_PORT3_OFFSET+KSZ8864_MIB_PORTX_TX_TOTAL_COLLISIONS,  { .owner = THIS_MODULE, .open = ks_proc_mib_open, .write = NULL, .read = seq_read, .llseek = seq_lseek, .release = single_release,}, },
  { "p3_txExcessiveCols",     (void*)KSZ8864_MIB_PORT3_OFFSET+KSZ8864_MIB_PORTX_TX_EXCESSIVE_COLLISIONS,  { .owner = THIS_MODULE, .open = ks_proc_mib_open, .write = NULL, .read = seq_read, .llseek = seq_lseek, .release = single_release,}, },
  { "p3_txSingleCols",        (void*)KSZ8864_MIB_PORT3_OFFSET+KSZ8864_MIB_PORTX_TX_SINGLE_COLLISIONS,  { .owner = THIS_MODULE, .open = ks_proc_mib_open, .write = NULL, .read = seq_read, .llseek = seq_lseek, .release = single_release,}, },
  { "p3_txMultipleCols",      (void*)KSZ8864_MIB_PORT3_OFFSET+KSZ8864_MIB_PORTX_TX_MULTIPLE_COLLISIONS,  { .owner = THIS_MODULE, .open = ks_proc_mib_open, .write = NULL, .read = seq_read, .llseek = seq_lseek, .release = single_release,}, },

  { "p4_rxLoPriorityBs",      (void*)KSZ8864_MIB_PORT4_OFFSET+KSZ8864_MIB_PORTX_RX_LO_PRIORITY_BYTES,  { .owner = THIS_MODULE, .open = ks_proc_mib_open, .write = NULL, .read = seq_read, .llseek = seq_lseek, .release = single_release,}, },
  { "p4_rxHiPriorityBs",      (void*)KSZ8864_MIB_PORT4_OFFSET+KSZ8864_MIB_PORTX_RX_HI_PRIORITY_BYTES,  { .owner = THIS_MODULE, .open = ks_proc_mib_open, .write = NULL, .read = seq_read, .llseek = seq_lseek, .release = single_release,}, },
  { "p4_rxUndersizePkts",     (void*)KSZ8864_MIB_PORT4_OFFSET+KSZ8864_MIB_PORTX_RX_UNDERSIZE_PACKETS,  { .owner = THIS_MODULE, .open = ks_proc_mib_open, .write = NULL, .read = seq_read, .llseek = seq_lseek, .release = single_release,}, },
  { "p4_rxFragments",         (void*)KSZ8864_MIB_PORT4_OFFSET+KSZ8864_MIB_PORTX_RX_FRAGMENTS,  { .owner = THIS_MODULE, .open = ks_proc_mib_open, .write = NULL, .read = seq_read, .llseek = seq_lseek, .release = single_release,}, },
  { "p4_rxOversizePkts",      (void*)KSZ8864_MIB_PORT4_OFFSET+KSZ8864_MIB_PORTX_RX_OVERSIZE_PACKETS,  { .owner = THIS_MODULE, .open = ks_proc_mib_open, .write = NULL, .read = seq_read, .llseek = seq_lseek, .release = single_release,}, },
  { "p4_rxJabbers",           (void*)KSZ8864_MIB_PORT4_OFFSET+KSZ8864_MIB_PORTX_RX_JABBERS,  { .owner = THIS_MODULE, .open = ks_proc_mib_open, .write = NULL, .read = seq_read, .llseek = seq_lseek, .release = single_release,}, },
  { "p4_rxSymbolErrs",        (void*)KSZ8864_MIB_PORT4_OFFSET+KSZ8864_MIB_PORTX_RX_SYMBOL_ERRORS,  { .owner = THIS_MODULE, .open = ks_proc_mib_open, .write = NULL, .read = seq_read, .llseek = seq_lseek, .release = single_release,}, },
  { "p4_rxCrcErrs",           (void*)KSZ8864_MIB_PORT4_OFFSET+KSZ8864_MIB_PORTX_RX_CRC_ERRORS,  { .owner = THIS_MODULE, .open = ks_proc_mib_open, .write = NULL, .read = seq_read, .llseek = seq_lseek, .release = single_release,}, },
  { "p4_rxAlignmentErrs",     (void*)KSZ8864_MIB_PORT4_OFFSET+KSZ8864_MIB_PORTX_RX_ALIGNMENT_ERRORS,  { .owner = THIS_MODULE, .open = ks_proc_mib_open, .write = NULL, .read = seq_read, .llseek = seq_lseek, .release = single_release,}, },
  { "p4_rxControl8808Pkts",   (void*)KSZ8864_MIB_PORT4_OFFSET+KSZ8864_MIB_PORTX_RX_CONTROL_8808_PACKETS,  { .owner = THIS_MODULE, .open = ks_proc_mib_open, .write = NULL, .read = seq_read, .llseek = seq_lseek, .release = single_release,}, },
  { "p4_rxPausePkts",         (void*)KSZ8864_MIB_PORT4_OFFSET+KSZ8864_MIB_PORTX_RX_PAUSE_PACKETS,  { .owner = THIS_MODULE, .open = ks_proc_mib_open, .write = NULL, .read = seq_read, .llseek = seq_lseek, .release = single_release,}, },
  { "p4_rxBroadcastPkts",     (void*)KSZ8864_MIB_PORT4_OFFSET+KSZ8864_MIB_PORTX_RX_BROADCAST_PACKETS,  { .owner = THIS_MODULE, .open = ks_proc_mib_open, .write = NULL, .read = seq_read, .llseek = seq_lseek, .release = single_release,}, },
  { "p4_rxMulticastPkts",     (void*)KSZ8864_MIB_PORT4_OFFSET+KSZ8864_MIB_PORTX_RX_MULTICAST_PACKETS,  { .owner = THIS_MODULE, .open = ks_proc_mib_open, .write = NULL, .read = seq_read, .llseek = seq_lseek, .release = single_release,}, },
  { "p4_rxUnicastPkts",       (void*)KSZ8864_MIB_PORT4_OFFSET+KSZ8864_MIB_PORTX_RX_UNICAST_PACKETS,  { .owner = THIS_MODULE, .open = ks_proc_mib_open, .write = NULL, .read = seq_read, .llseek = seq_lseek, .release = single_release,}, },
  { "p4_rx64BsPkts",          (void*)KSZ8864_MIB_PORT4_OFFSET+KSZ8864_MIB_PORTX_RX_64_BYTE_PACKETS,  { .owner = THIS_MODULE, .open = ks_proc_mib_open, .write = NULL, .read = seq_read, .llseek = seq_lseek, .release = single_release,}, },
  { "p4_rx65-127BsPkts",      (void*)KSZ8864_MIB_PORT4_OFFSET+KSZ8864_MIB_PORTX_RX_65_TO_127_BYTE_PACKETS,  { .owner = THIS_MODULE, .open = ks_proc_mib_open, .write = NULL, .read = seq_read, .llseek = seq_lseek, .release = single_release,}, },
  { "p4_rx128-255BsPkts",     (void*)KSZ8864_MIB_PORT4_OFFSET+KSZ8864_MIB_PORTX_RX_128_TO_255_BYTE_PACKETS,  { .owner = THIS_MODULE, .open = ks_proc_mib_open, .write = NULL, .read = seq_read, .llseek = seq_lseek, .release = single_release,}, },
  { "p4_rx256-511BsPkts",     (void*)KSZ8864_MIB_PORT4_OFFSET+KSZ8864_MIB_PORTX_RX_256_TO_511_BYTE_PACKETS,  { .owner = THIS_MODULE, .open = ks_proc_mib_open, .write = NULL, .read = seq_read, .llseek = seq_lseek, .release = single_release,}, },
  { "p4_rx512-1023BsPkts",    (void*)KSZ8864_MIB_PORT4_OFFSET+KSZ8864_MIB_PORTX_RX_512_TO_1023_BYTE_PACKETS,  { .owner = THIS_MODULE, .open = ks_proc_mib_open, .write = NULL, .read = seq_read, .llseek = seq_lseek, .release = single_release,}, },
  { "p4_rx1024-1522BsPkts",   (void*)KSZ8864_MIB_PORT4_OFFSET+KSZ8864_MIB_PORTX_RX_1024_TO_1522_BYTE_PACKETS,  { .owner = THIS_MODULE, .open = ks_proc_mib_open, .write = NULL, .read = seq_read, .llseek = seq_lseek, .release = single_release,}, },
  { "p4_txLoPriorityBs",      (void*)KSZ8864_MIB_PORT4_OFFSET+KSZ8864_MIB_PORTX_TX_LO_PRIORITY_BYTES,  { .owner = THIS_MODULE, .open = ks_proc_mib_open, .write = NULL, .read = seq_read, .llseek = seq_lseek, .release = single_release,}, },
  { "p4_txHiPriorityBs",      (void*)KSZ8864_MIB_PORT4_OFFSET+KSZ8864_MIB_PORTX_TX_HI_PRIORITY_BYTES,  { .owner = THIS_MODULE, .open = ks_proc_mib_open, .write = NULL, .read = seq_read, .llseek = seq_lseek, .release = single_release,}, },
  { "p4_txLateCols",          (void*)KSZ8864_MIB_PORT4_OFFSET+KSZ8864_MIB_PORTX_TX_LATE_COLLISIONS,  { .owner = THIS_MODULE, .open = ks_proc_mib_open, .write = NULL, .read = seq_read, .llseek = seq_lseek, .release = single_release,}, },
  { "p4_txPausePkts",         (void*)KSZ8864_MIB_PORT4_OFFSET+KSZ8864_MIB_PORTX_TX_PAUSE_PACKETS,  { .owner = THIS_MODULE, .open = ks_proc_mib_open, .write = NULL, .read = seq_read, .llseek = seq_lseek, .release = single_release,}, },
  { "p4_txBroadcastPkts",     (void*)KSZ8864_MIB_PORT4_OFFSET+KSZ8864_MIB_PORTX_TX_BROADCAST_PACKETS,  { .owner = THIS_MODULE, .open = ks_proc_mib_open, .write = NULL, .read = seq_read, .llseek = seq_lseek, .release = single_release,}, },
  { "p4_txMulticastPkts",     (void*)KSZ8864_MIB_PORT4_OFFSET+KSZ8864_MIB_PORTX_TX_MULTICAST_PACKETS,  { .owner = THIS_MODULE, .open = ks_proc_mib_open, .write = NULL, .read = seq_read, .llseek = seq_lseek, .release = single_release,}, },
  { "p4_txUnicastPkts",       (void*)KSZ8864_MIB_PORT4_OFFSET+KSZ8864_MIB_PORTX_TX_UNICAST_PACKETS,  { .owner = THIS_MODULE, .open = ks_proc_mib_open, .write = NULL, .read = seq_read, .llseek = seq_lseek, .release = single_release,}, },
  { "p4_txDeferredPkts",      (void*)KSZ8864_MIB_PORT4_OFFSET+KSZ8864_MIB_PORTX_TX_DEFERRED_PACKETS,  { .owner = THIS_MODULE, .open = ks_proc_mib_open, .write = NULL, .read = seq_read, .llseek = seq_lseek, .release = single_release,}, },
  { "p4_txTotalCols",         (void*)KSZ8864_MIB_PORT4_OFFSET+KSZ8864_MIB_PORTX_TX_TOTAL_COLLISIONS,  { .owner = THIS_MODULE, .open = ks_proc_mib_open, .write = NULL, .read = seq_read, .llseek = seq_lseek, .release = single_release,}, },
  { "p4_txExcessiveCols",     (void*)KSZ8864_MIB_PORT4_OFFSET+KSZ8864_MIB_PORTX_TX_EXCESSIVE_COLLISIONS,  { .owner = THIS_MODULE, .open = ks_proc_mib_open, .write = NULL, .read = seq_read, .llseek = seq_lseek, .release = single_release,}, },
  { "p4_txSingleCols",        (void*)KSZ8864_MIB_PORT4_OFFSET+KSZ8864_MIB_PORTX_TX_SINGLE_COLLISIONS,  { .owner = THIS_MODULE, .open = ks_proc_mib_open, .write = NULL, .read = seq_read, .llseek = seq_lseek, .release = single_release,}, },
  { "p4_txMultipleCols",      (void*)KSZ8864_MIB_PORT4_OFFSET+KSZ8864_MIB_PORTX_TX_MULTIPLE_COLLISIONS,  { .owner = THIS_MODULE, .open = ks_proc_mib_open, .write = NULL, .read = seq_read, .llseek = seq_lseek, .release = single_release,}, },

  { "p1_txDroppedPkts",       (void*)KSZ8864_MIB_PACKETS_DROPPED_PORT1_TX,  { .owner = THIS_MODULE, .open = ks_proc_mib_open, .write = NULL, .read = seq_read, .llseek = seq_lseek, .release = single_release,}, },
  { "p2_txDroppedPkts",       (void*)KSZ8864_MIB_PACKETS_DROPPED_PORT2_TX,  { .owner = THIS_MODULE, .open = ks_proc_mib_open, .write = NULL, .read = seq_read, .llseek = seq_lseek, .release = single_release,}, },
  { "p3_txDroppedPkts",       (void*)KSZ8864_MIB_PACKETS_DROPPED_PORT3_TX,  { .owner = THIS_MODULE, .open = ks_proc_mib_open, .write = NULL, .read = seq_read, .llseek = seq_lseek, .release = single_release,}, },
  { "p4_txDroppedPkts",       (void*)KSZ8864_MIB_PACKETS_DROPPED_PORT4_TX,  { .owner = THIS_MODULE, .open = ks_proc_mib_open, .write = NULL, .read = seq_read, .llseek = seq_lseek, .release = single_release,}, },
  { "p1_rxDroppedPkts",       (void*)KSZ8864_MIB_PACKETS_DROPPED_PORT1_RX,  { .owner = THIS_MODULE, .open = ks_proc_mib_open, .write = NULL, .read = seq_read, .llseek = seq_lseek, .release = single_release,}, },
  { "p2_rxDroppedPkts",       (void*)KSZ8864_MIB_PACKETS_DROPPED_PORT2_RX,  { .owner = THIS_MODULE, .open = ks_proc_mib_open, .write = NULL, .read = seq_read, .llseek = seq_lseek, .release = single_release,}, },
  { "p3_rxDroppedPkts",       (void*)KSZ8864_MIB_PACKETS_DROPPED_PORT3_RX,  { .owner = THIS_MODULE, .open = ks_proc_mib_open, .write = NULL, .read = seq_read, .llseek = seq_lseek, .release = single_release,}, },
  { "p4_rxDroppedPkts",       (void*)KSZ8864_MIB_PACKETS_DROPPED_PORT4_RX,  { .owner = THIS_MODULE, .open = ks_proc_mib_open, .write = NULL, .read = seq_read, .llseek = seq_lseek, .release = single_release,}, },
};

static struct ks_proc_entry ks_proc_entries[] = {
  { "sroute",                 NULL ,  { .owner = THIS_MODULE, .open = ks_proc_sroute_open, .write = ks_proc_sroute_write, .read = seq_read, .llseek = seq_lseek, .release = single_release,}, },
  { "droute",                 NULL ,  { .owner = THIS_MODULE, .open = ks_proc_droute_open, .write = NULL, .read = seq_read, .llseek = seq_lseek, .release = single_release,}, },
};

/*=== Local Function definitions (File scope) ========================================================================*/
/** Reset the switch */
static void reset_switch(void)
{
    gpio_direction_output(ETH_RESET_PIN, 0);
	msleep_interruptible(ETH_RESET_TIME_MS);
    gpio_direction_input(ETH_RESET_PIN);
}

/**
 * Write an SMI register to switch
 * @param[in,out] ks   The chip state
 * @param[in]     reg  The register address
 * @param[in]     val  The value to write
 * @note          Will assert if @p ks->ks8864_lock is not locked
 */
void ksz8864_writeRegister(struct ethsw_net *ks, unsigned const reg, unsigned const val)
{
	struct spi_transfer *xfer = &ks->spi_xfer;
	struct spi_message *msg = &ks->spi_msg;
	u8 txb[3];
	int ret;

	BUG_ON_NOT_LOCKED(ks->ksz8864_lock);

	txb[0] = KSZ8864_SMI_WRITE_CMD;
	txb[1] = reg;
	txb[2] = val;

	xfer->tx_buf = txb;
	xfer->rx_buf = NULL;
	xfer->len = 3;

	ret = spi_sync(ks->spidev, msg);
	if (ret < 0)
		ks_err(ks, "spi_sync() failed on 8-bit wr\n");
}

/**
 * Read a number of SMI registers from switch
 * @param[in,out] ks            The device state
 * @param[in]     startReg      The register number to start reading from
 * @param[out]    data          Read destination buffer
 * @param[in]     noBytes       No bytes to read
 * @note          Will assert if @p ks->ks8864_lock is not locked
 */
void ksz8864_readData(struct ethsw_net *ks, unsigned startReg, void* const data, unsigned noBytes)
{
	struct spi_transfer* xfer = &ks->spi_xfer;
	struct spi_message* msg = &ks->spi_msg;
	u8* const rxb = data;
	u8* trx = ks->rxd;
	u8* txb = ks->txd;
	int ret;
	unsigned i = 0;
	unsigned x;

	BUG_ON_NOT_LOCKED(ks->ksz8864_lock);

	/* Check bounds */
	if (noBytes > (ETHSW_SMI_RX_BUF_SIZE - 2))
		ks_err(ks, "Read: requested more data from switch then it would fit the 'ETHSW_SMI_RX_BUF_SIZE' buf.\n");

	txb[i++] = KSZ8864_SMI_READ_CMD;
	txb[i++] = startReg;
	for (x = 0; x < noBytes; x++)
		txb[i++] = 0; /* dummy */

	xfer->tx_buf = txb;
	xfer->rx_buf = trx;
	xfer->len = i;

	ret = spi_sync(ks->spidev, msg);
	if (ret < 0)
		ks_err(ks, "read: spi_sync() failed on read reg\n");
	else
		memcpy(rxb, trx + 2, noBytes);
}

/**
 * Write a number of SMI registers to switch
 * @param[in,out] ks        The device state
 * @param[in]     startReg  The register number to start reading from
 * @param[out]    data      Read destination buffer
 * @param[in]     noBytes   No bytes to read
 * @note          Will assert if @p ks->ks8864_lock is not locked
 */
void ksz8864_writeData(struct ethsw_net *ks, unsigned startReg, void const* const data, unsigned noBytes)
{
	struct spi_transfer* xfer = &ks->spi_xfer;
	struct spi_message* msg = &ks->spi_msg;
	u8* txb = ks->txd;
	int ret;
	unsigned i = 0;

	BUG_ON_NOT_LOCKED(ks->ksz8864_lock);

	/* Check bounds */
	if (noBytes > (ETHSW_SMI_TX_BUF_SIZE - 2))
		ks_err(ks, "Read: requested more data from switch then it would fit the 'ETHSW_SMI_RX_BUF_SIZE' buf.\n");

	txb[i++] = KSZ8864_SMI_WRITE_CMD;
	txb[i++] = startReg;
	memcpy(txb + i, data, noBytes);
	i += noBytes;

	xfer->tx_buf = txb;
	xfer->rx_buf = NULL;
	xfer->len = i;

	ret = spi_sync(ks->spidev, msg);
	if (ret < 0)
		ks_err(ks, "read: spi_sync() failed on read reg\n");
}

/**
 * Write an SMI register to switch
 * @param[in,out] ks        The device state
 * @param[in]     reg       The register number
 * @return        The value of the register
 * @note          Will assert if @p ks->ks8864_lock is not locked
 */
unsigned ksz8864_readRegister(struct ethsw_net *ks, unsigned reg)
{
	u8 rxb;
  /* BUG_ON_NOT_LOCKED done in ksz8864_readData */
	ksz8864_readData(ks, reg, &rxb, sizeof(rxb));
	return (unsigned) rxb;
}

/**
 * Reads a PHY register from the port 3 PHY
 * @ks[in]    The chip information
 * @reg[in]   The register number
 * @retval    The register value
 * @note          Will assert if @p ks->mii_lock is not locked
 */
int p3phy_read(struct ethsw_net *ks, u16 regnum)
{
	BUG_ON_NOT_LOCKED(ks->mii_lock);

	return mdiobus_read(ks->mii_bus, PHY_ADDR, regnum);
}

/**
 * Writes a PHY register on the port 3 PHY
 * @ks[in]    The chip information
 * @reg[in]   The register number
 * @val[in]   The register value
 * @note          Will assert if @p ks->mii_lock is not locked
 */
int p3phy_write(struct ethsw_net *ks, u16 regnum, u16 val)
{
	BUG_ON_NOT_LOCKED(ks->mii_lock);

	return mdiobus_write(ks->mii_bus, PHY_ADDR, regnum, val);
}

#ifdef HAS_EXT_PHY
/**
 * Bit set / clear operations for phy and switch registers
 * @ks[in]    The chip information
 * @reg[in]   The register number
 * @bits[in]  The bit-mask to set / clear
 */
static void phyRegSetBits(struct ethsw_net *ks, unsigned reg, unsigned bits) {
	int value = p3phy_read(ks, reg);
	value |= bits;
	p3phy_write(ks, reg, value);
}

static void phyRegClearBits(struct ethsw_net *ks, unsigned reg, unsigned bits) {
	int value = p3phy_read(ks, reg);
	value &= ~(bits);
	p3phy_write(ks, reg, value);
}
#endif

static void kszRegSetBits8(struct ethsw_net *ks, unsigned reg, unsigned bits) {
	int value = ksz8864_readRegister(ks, reg);
	value |= bits;
	ksz8864_writeRegister(ks, reg, value);
}

static void kszRegClearBits8(struct ethsw_net *ks, unsigned reg, unsigned bits) {
	int value = ksz8864_readRegister(ks, reg);
	value &= ~(bits);
	ksz8864_writeRegister(ks, reg, value);
}

#ifdef HAS_MAC
/**
 * Callback for packets received by the MAC peripheral.
 * This callback will remove a tail tag which indicates the incoming switch port,
 * when tailtagging is enabled.
 * @param [in,out] skb     skbuff containing a network packet received by the network adapter.
 */
static void onRxMac(struct sk_buff *skb, void* callbackObj)
{
	struct ethsw_net *ks = (struct ethsw_net*)callbackObj;

	if (ks != NULL && ks->tail_tag_enabled)
	{
		/* Store tail tag */
		unsigned char tailTag = skb->data[skb->len - TAILTAG_SIZE];
		if (unlikely(! KSZ8864_TAIL_TAG_EGRESS_IS_VALID(tailTag))) {
			KSERR("Invalid tail tag encountered: 0x%02x\n", tailTag);
		} else {
			skb->cb[0] = tailTag & KSZ8864_TAIL_TAG_EGRESS_PORT_MASK;
		}

		/* Remove tail tag from skb */
		skb_trim(skb, skb->len - TAILTAG_SIZE);
	}
}

/**
 * Callback for packets about to be transmitted by the MAC peripheral.
 * This callback will add a tail tag to the packet to control the outgoing switch port,
 * when tailtagging is enabled.
 * @param [in,out] skb     skbuff containing a network packet to be sent by the network adapter.
 */
static void onTxMac(struct sk_buff *skb, void* callbackObj)
{
	struct ethsw_net *ks = (struct ethsw_net*)callbackObj;

	KSDBG_DATA("TX len %u\n", skb->len);
	KSDBGHEX_DATA("TX ", skb->data, skb->len);

	if (ks != NULL && ks->tail_tag_enabled)
	{
		/*
		 * We have to make sure that the trailer ends up as the very
		 * last byte of the packet.  This means that we have to pad
		 * the packet to the minimum Ethernet frame size (ETH_ZLEN), if necessary,
		 * before adding the trailer.
		 */

		unsigned char* tail;
		unsigned char tailTag;

		/* Assure we can fit the tail tag and optional padding to minimum packet size
		   in the skb. */
		if (skb_padto(skb, TAILTAG_SIZE + max((unsigned)ETH_ZLEN, skb->len)) != 0) {
			KSERR("Failed to pad ethernet packet to required minimum size\n");
			return;
		}

		if (skb->len < ETH_ZLEN)
		{
			/* Pad the packet to the minimum Ethernet frame size */
			(void)skb_put(skb, ETH_ZLEN - skb->len);
		}

		if ((skb->cb[0] < 1) || (skb->cb[0] >= KSZ8864_MAX_PORTS))
		{
			/* No or invalid port specified, set tailtag to be ignored by switch */
			tailTag = KSZ8864_TAIL_TAG_INGRESS_IGNORE;
		}
		else
		{
			/* Outgoing port specified */
			tailTag = KSZ8864_TAIL_TAG_INGRESS_TO_PORT(skb->cb[0]) | KSZ8864_TAIL_TAG_INGRESS_QUEUE_WHATEVER;
		}

		/* Extend skb with tailtag */
		tail = skb_put(skb,TAILTAG_SIZE);
		tail[0] = tailTag;
		KSDBG_TAG("TX tailtag 0x%02x\n", tail[0]);
	}

	KSDBG_DATA("TX len %u\n", skb->len);
	KSDBGHEX_DATA("TX ", skb->data, skb->len);
}
#endif

/**
 * Enable or disable tail tagging functionality.
 * @param [in,out] ks     The device instance
 * @param [in]     enable @p true enables tail tagging, @p false disables it.
 * @return Result code.
 */
static int tailTagEnable(struct ethsw_net *ks, const bool enable)
{
	int res = 0;

	/* BUG_ON_NOT_LOCKED is done in kszRegSetBits8/kszRegClearBits8 */
	BUG_ON_NOT_LOCKED(ks->dev_lock);

	if (enable) {
		kszRegSetBits8(ks, KSZ8864_SMI_CONTROL_10_GLOBAL, KSZ8864_SMI_CONTROL_10_GLOBAL_TAIL_TAG_ENABLE);
	} else {
		kszRegClearBits8(ks, KSZ8864_SMI_CONTROL_10_GLOBAL, KSZ8864_SMI_CONTROL_10_GLOBAL_TAIL_TAG_ENABLE);
	}

	ks->tail_tag_enabled = enable;

	KSDBG("Tail tagging %s\n", enable ? "enabled" : "disabled");
	return res;
}

/**
 * Handles reading of a switch register
 * @param [out]    page   Where the data goes
 * @param [in,out] start
 * @param [in]     off    Read offset
 * @param [in]     count  Number of bytes to read
 * @param [in,out] eof    End-of-file flag
 * @param [in,out] data   Data pointer registered with this function
 * @return                Number of bytes read
 */
//static size_t ks_proc_reg_read(struct file *file, char __user *buffer, size_t count, loff_t *ppos)
static int ks_proc_reg_read(struct seq_file *m, void *v)
{
	unsigned const            reg = (unsigned)m->private;
	struct ethsw_net* const   ks  = &ethsw;

	if(m->read_pos == 0) {
		LOCK_RESOURCE(ks->ksz8864_lock);
		seq_printf(m, "%u\n", ksz8864_readRegister(ks, reg));
		RELEASE_RESOURCE(ks->ksz8864_lock);
	}
	return 0;
}

/**
 * Handles writing to a switch register
 * @param [in]     file   The assocated file
 * @param [in]     buffer The userspace buffer containing the write data
 * @param [in]     count  The buffer size
 * @param [in,out] data   Data pointer registered with this function
 * @return                Number of bytes written
 */
static ssize_t ks_proc_reg_write(struct file *file, const char __user *buffer, size_t count, loff_t *ppos)
{
	unsigned const            reg = (unsigned)PDE_DATA(file_inode(file));
	struct ethsw_net* const   ks  = &ethsw;
	int                       value;
	char                      buff[KS_PROC_MAX_WRITE_SIZE + 1];

	/* limit buffer size */
	count = min(KS_PROC_MAX_WRITE_SIZE, count);

	/* write data to the buffer */
	if ( copy_from_user(buff, buffer, count) ) {
		return -EFAULT;
	}
	buff[count] = '\0';

	if(sscanf(buff, "%u", &value) != 1) {
		return -EINVAL;
	}
	LOCK_RESOURCE(ks->ksz8864_lock);
	ksz8864_writeRegister(ks, reg, value);
	RELEASE_RESOURCE(ks->ksz8864_lock);

	return count;
}

static int ks_proc_reg_open(struct inode *inode, struct file *file)
{
	return single_open(file, ks_proc_reg_read, PDE_DATA(inode));
}

/**
 * Handles reading of an MII register
 * @param [out]    page   Where the data goes
 * @param [in,out] start
 * @param [in]     off    Read offset
 * @param [in]     count  Number of bytes to read
 * @param [in,out] eof    End-of-file flag
 * @param [in,out] data   Data pointer registered with this function
 * @return                Number of bytes read
 */
//static size_t ks_proc_mii_read(struct file *file, char __user *buffer, size_t count, loff_t *ppos)
static int ks_proc_mii_read(struct seq_file *m, void *v)
{
	unsigned const            phyAddr = ((unsigned)m->private) >> 16;
	unsigned const            regNo   = ((unsigned)m->private) & 0xFFFF;
	struct ethsw_net* const   ks      = &ethsw;
	size_t                    len     = 0;

	if(m->read_pos == 0) {
		LOCK_RESOURCE(ks->mii_lock);
		seq_printf(m, "%u\n", mdiobus_read(ks->mii_bus, phyAddr, regNo));
		RELEASE_RESOURCE(ks->mii_lock);
	}
	return len;
}

/**
 * Handles writing to an MII  register
 * @param [in]     file   The assocated file
 * @param [in]     buffer The userspace buffer containing the write data
 * @param [in]     count  The buffer size
 * @param [in,out] data   Data pointer registered with this function
 * @return                Number of bytes written
 */
static ssize_t ks_proc_mii_write(struct file *file, const char __user *buffer, size_t count, loff_t *ppos)
{
	unsigned const            phyAddr = ((unsigned)PDE_DATA(file_inode(file))) >> 16;
	unsigned const            regNo   = ((unsigned)PDE_DATA(file_inode(file))) & 0xFFFF;
	struct ethsw_net* const   ks      = &ethsw;
	int                       value;
	char                      buf[KS_PROC_MAX_WRITE_SIZE + 1];

	/* limit buffer size */
	count = min(KS_PROC_MAX_WRITE_SIZE, count);

	/* write data to the buffer */
	if ( copy_from_user(buf, buffer, count) ) {
		return -EFAULT;
	}
	buf[count] = '\0';

	if(sscanf(buf, "%u", &value) != 1) {
		return -EINVAL;
	}
	LOCK_RESOURCE(ks->mii_lock);
	mdiobus_write(ks->mii_bus, phyAddr, regNo, value);
	RELEASE_RESOURCE(ks->mii_lock);

	return count;
}

static int ks_proc_mii_open(struct inode *inode, struct file *file)
{
	return single_open(file, ks_proc_mii_read, PDE_DATA(inode));
}

/**
 * Handles read requests on the reset control procfs entry
 * @param [out]    page   Where the data goes
 * @param [in,out] start
 * @param [in]     off    Read offset
 * @param [in]     count  Number of bytes to read
 * @param [in,out] eof    End-of-file flag
 * @param [in,out] data   Data pointer registered with this function
 * @return                Number of bytes read
 */
static int ks_proc_ctl_reset_read(struct seq_file *m, void *v)
{
	struct ethsw_net* const ks  = &ethsw;

	if(m->read_pos == 0) {
		LOCK_RESOURCE(ks->dev_lock);
		if(ks->reset_result != NULL) {
			seq_printf(m, "%s\n", ks->reset_result);
		} else {
			seq_printf(m, "%s\n", ETH_RESET_IDLE);
		}
		ks->reset_result = NULL;
		RELEASE_RESOURCE(ks->dev_lock);
	}
	return 0;
}



/**
 * Handles switch / PHY reset requests via the procfs control interface
 * @param [in]     file   The assocated file
 * @param [in]     buffer The userspace buffer containing the write data
 * @param [in]     count  The buffer size
 * @param [in,out] data   Data pointer registered with this function
 * @return                Number of bytes written
 */
static ssize_t ks_proc_ctl_reset_write(struct file *file, const char __user *buffer, size_t count, loff_t *ppos)
{
	struct ethsw_net* const ks    = &ethsw;
	static char const       key[] = ETH_RESET_KEY;
	char                    buf[KS_PROC_MAX_WRITE_SIZE + 1];

	/* limit buffer size */
	count = min(KS_PROC_MAX_WRITE_SIZE, count);

	/* write data to the buffer */
	if ( copy_from_user(buf, buffer, count) ) {
		return -EFAULT;
	}
	buf[count] = '\0';

	if(strncmp(buf, key, sizeof(key)-1) != 0) {
		LOCK_RESOURCE(ks->dev_lock);
		ks->reset_result = ETH_RESET_BAD_KEY;
		RELEASE_RESOURCE(ks->dev_lock);
		return -EINVAL;
	}

    reset_switch();

	LOCK_RESOURCE(ks->ksz8864_lock);

	/* Start the switch in the appropriate mode and restore the device MAC addresses to the static MAC address table */
	ksz8864_start(ks);

  LOCK_RESOURCE(ks->dev_lock);
  (void)tailTagEnable(ks, true);
  RELEASE_RESOURCE(ks->dev_lock);

	ksz8864_setStaticEntries(
	  ks,
	  ks->smac_entries,
	  sizeof(ks->smac_entries) / sizeof(ks->smac_entries[0]) );

	RELEASE_RESOURCE(ks->ksz8864_lock);

	LOCK_RESOURCE(ks->dev_lock);
	ks->reset_result = ETH_RESET_DONE;
	RELEASE_RESOURCE(ks->dev_lock);

	return count;
}

static int ks_proc_ctl_reset_open(struct inode *inode, struct file *file)
{
	return single_open(file, ks_proc_ctl_reset_read, PDE_DATA(inode));
}

/**
 * Handles read requests on the flush dynamic MAC control procfs entry
 * @param [out]    page   Where the data goes
 * @param [in,out] start
 * @param [in]     off    Read offset
 * @param [in]     count  Number of bytes to read
 * @param [in,out] eof    End-of-file flag
 * @param [in,out] data   Data pointer registered with this function
 * @return                Number of bytes read
 */
static int ks_proc_ctl_fldmac_read(struct seq_file *m, void *v)
{
	struct ethsw_net* const ks  = &ethsw;

	if(m->read_pos == 0) {
		LOCK_RESOURCE(ks->dev_lock);
		if(ks->dmac_flush_result != NULL) {
			seq_printf(m, "%s\n", ks->dmac_flush_result);
		} else {
			seq_printf(m, "%s\n", ETH_DMAC_FLUSH_IDLE);
		}
		ks->dmac_flush_result = NULL;
		RELEASE_RESOURCE(ks->dev_lock);
	}
	return 0;
}

/**
 * Handles flushing of the dynamic routes (dynamic MAC address table)
 * @param [in]     file   The assocated file
 * @param [in]     buffer The userspace buffer containing the write data
 * @param [in]     count  The buffer size
 * @param [in,out] data   Data pointer registered with this function
 * @return                Number of bytes written
 */
static ssize_t ks_proc_ctl_fldmac_write(struct file *file, const char __user *buffer, size_t count, loff_t *ppos)
{
	struct ethsw_net* const ks    = &ethsw;
	static char const       key[] = ETH_DMAC_FLUSH_KEY;
	char                    buf[KS_PROC_MAX_WRITE_SIZE + 1];

	/* limit buffer size */
	count = min(KS_PROC_MAX_WRITE_SIZE, count);

	/* write data to the buffer */
	if ( copy_from_user(buf, buffer, count) ) {
		return -EFAULT;
	}
	buf[count] = '\0';

	if(strncmp(buf, key, sizeof(key)-1) != 0) {
		LOCK_RESOURCE(ks->dev_lock);
		ks->dmac_flush_result = ETH_DMAC_FLUSH_BAD_KEY;
		RELEASE_RESOURCE(ks->dev_lock);
		return -EINVAL;
	}
	ethsw_clearMacTable(ETHSW_MAC_TABLE_DYNAMIC);

	LOCK_RESOURCE(ks->dev_lock);
	ks->dmac_flush_result = ETH_DMAC_FLUSH_DONE;
	RELEASE_RESOURCE(ks->dev_lock);

	return count;
}

static int ks_proc_ctl_fldmac_open(struct inode *inode, struct file *file)
{
	return single_open(file, ks_proc_ctl_fldmac_read, PDE_DATA(inode));
}

/**
 * Handles read requests on the tailtag function procfs entry
 * @param [out]    page   Where the data goes
 * @param [in,out] start
 * @param [in]     off    Read offset
 * @param [in]     count  Number of bytes to read
 * @param [in,out] eof    End-of-file flag
 * @param [in,out] data   Data pointer registered with this function
 * @return                Number of bytes read
 */
static int ks_proc_ctl_tailtag_read(struct seq_file *m, void *v)
{
	struct ethsw_net* const ks  = &ethsw;

	if(m->read_pos == 0) {
		seq_printf(m, "%u\n", ks->tail_tag_enabled ? 1 : 0);
	}
	return 0;
}

/**
 * Handles enabling/disabling of the tailtag function procfs entry
 * @param [in]     file   The assocated file
 * @param [in]     buffer The userspace buffer containing the write data
 * @param [in]     count  The buffer size
 * @param [in,out] data   Data pointer registered with this function
 * @return                Number of bytes written
 */
static ssize_t ks_proc_ctl_tailtag_write(struct file *file, const char __user *buffer, size_t count, loff_t *ppos)
{
	struct ethsw_net* const ks  = &ethsw;
	int                     res;
	int                     value;
	char                    buf[KS_PROC_MAX_WRITE_SIZE + 1];

	/* limit buffer size */
	count = min(KS_PROC_MAX_WRITE_SIZE, count);

	/* write data to the buffer */
	if ( copy_from_user(buf, buffer, count) ) {
		return -EFAULT;
	}
	buf[count] = '\0';

	if(sscanf(buf, "%u", &value) != 1) {
		return -EINVAL;
	}
	if(value < 0 || value > 1) {
		return -EINVAL;
	}

	LOCK_RESOURCE(ks->ksz8864_lock);
	LOCK_RESOURCE(ks->dev_lock);

	res = tailTagEnable(ks, (value == 1) ? true : false );

	RELEASE_RESOURCE(ks->dev_lock);
	RELEASE_RESOURCE(ks->ksz8864_lock);

	if (res < 0) {
		return res;
	}
	return count;
}

static int ks_proc_ctl_tailtag_open(struct inode *inode, struct file *file)
{
	return single_open(file, ks_proc_ctl_tailtag_read, PDE_DATA(inode));
}

/**
 * Handles reading of an MIB counter
 * @param [out]    page   Where the data goes
 * @param [in,out] start
 * @param [in]     off    Read offset
 * @param [in]     count  Number of bytes to read
 * @param [in,out] eof    End-of-file flag
 * @param [in,out] data   Data pointer registered with this function
 * @return                Number of bytes read
 */
static int ks_proc_mib_read(struct seq_file *m, void *v)
{
	unsigned const            offset = (unsigned)m->private;
	struct ethsw_net* const ks  = &ethsw;

	if(m->read_pos == 0) {
		T_KSZ8864_MIB_COUNTER_STATUS status;
		unsigned                     value;
		LOCK_RESOURCE(ks->ksz8864_lock);
		status = ksz8864_getCounterMIB(ks, offset, &value);
		seq_printf(m, "%d\n",
		  (status == KSZ8864_MIB_COUNTER_STATUS_OVERFLOW) ? -(int)value : (int)value );
		RELEASE_RESOURCE(ks->ksz8864_lock);
	}
	return 0;
}

static int ks_proc_mib_open(struct inode *inode, struct file *file)
{
	return single_open(file, ks_proc_mib_read, PDE_DATA(inode));
}

/**
 * Handles reading of the static routes (static MAC address table)
 * @param [out]    page   Where the data goes
 * @param [in,out] start
 * @param [in]     off    Read offset
 * @param [in]     count  Number of bytes to read
 * @param [in,out] eof    End-of-file flag
 * @param [in,out] data   Data pointer registered with this function
 * @return                Number of bytes read
 */
static int ks_proc_sroute_read(struct seq_file *m, void *v)
{
	struct ethsw_net* const ks  = &ethsw;
	int                     i;

	if(m->read_pos == 0) {
		LOCK_RESOURCE(ks->dev_lock);
		seq_printf(m, "MacAddress    Valid  Override  DestMask\n");
		for(i = 0; i < sizeof(ks->smac_entries) / sizeof(ks->smac_entries[0]); i++) {

			if(ks->smac_entries[i].valid == KSZ8864_SMAC_INVALID) {
				continue;
			}

			seq_printf(m, "%012"PRIx64"  %u      %u         %01x\n",
			  ks->smac_entries[i].macAddr,
			  ks->smac_entries[i].valid,
			  ks->smac_entries[i].override,
			  ks->smac_entries[i].forwardMask );
		}
		RELEASE_RESOURCE(ks->dev_lock);
	}
	return 0;
}

static int ks_proc_sroute_open(struct inode *inode, struct file *file)
{
	return single_open(file, ks_proc_sroute_read, PDE_DATA(inode));
}


/**
 * Write a single entry to the static MAC table. When the entry does not yet exist it will be added,
 * otherwise it will overwrite an existing entry. Writing an entry with valid set to false will effectively
 * delete the entry from the static MAC table.
 * @param [in,out] ks           The device instance
 * @param [in]     valid        Valid-flag for the MAC entry
 * @param [in]     override     RSTP override-flag for the MAC entry.
 * @param [in]     forwardMask  Mask indicating which ports to forward to, for the MAC entry.
 * @param [in]     macAddr      MAC address.
 * @return @p 0 when entry succesfull written, @p <0 indicates error.
 */
static int srouteWrite( struct ethsw_net *ks, const bool valid, const bool override,
							const unsigned forwardMask,	unsigned long long macAddr )
{
	int entryIdx = -1;
	int i;
	int res = 0;

	LOCK_RESOURCE(ks->dev_lock);

	for(i = 0; i < sizeof(ks->smac_entries) / sizeof(ks->smac_entries[0]); i++) {
		if(ks->smac_entries[i].valid == KSZ8864_SMAC_INVALID) {
			if(entryIdx < 0) {
				entryIdx = i; /* Got the first free entry */
			}
			continue;
		}
		if(macAddr == ks->smac_entries[i].macAddr) {
			entryIdx = i;        /* Got the matching entry */
			break;
		}
	}
	if(entryIdx < 0) {
		KSMSG("error: Static MAC address table full\n");
		res = -EFAULT;
		goto out;
	}
	ks->smac_entries[entryIdx].macAddr     = macAddr;
	ks->smac_entries[entryIdx].valid       = valid;
	ks->smac_entries[entryIdx].override    = override;
	ks->smac_entries[entryIdx].forwardMask = forwardMask;

	LOCK_RESOURCE(ks->ksz8864_lock);
	ksz8864_addStaticMacAddr(ks, entryIdx, ks->smac_entries + entryIdx);
	RELEASE_RESOURCE(ks->ksz8864_lock);

	KSDBG("Wrote static MAC table entry %012"PRIx64", valid %d, override %d, forward 0x%.02x\n",
				macAddr, (int)valid, (int)override, forwardMask );

out:
	RELEASE_RESOURCE(ks->dev_lock);
	return res;
}

/**
 * Handles modifications to the static routes (static MAC address table)
 * @param [in]     file   The assocated file
 * @param [in]     buffer The userspace buffer containing the write data
 * @param [in]     count  The buffer size
 * @param [in,out] data   Data pointer registered with this function
 * @return                Number of bytes written
 */
static ssize_t ks_proc_sroute_write(struct file *file, const char __user *buffer, size_t count, loff_t *ppos)
{
	struct ethsw_net* const    ks  = &ethsw;
	int                        res;
	int                        items;
	unsigned int               valid;
	unsigned int               override;
	unsigned int               forwardMask;
	unsigned long long         macAddr;

	char                      buf[KS_PROC_MAX_WRITE_SIZE + 1];

	/* limit buffer size */
	count = min(KS_PROC_MAX_WRITE_SIZE, count);

	/* write data to the buffer */
	if ( copy_from_user(buf, buffer, count) ) {
		return -EFAULT;
	}
	buf[count] = '\0';

	items = sscanf(buf, "%llx %u %u %x\n",
	  &macAddr,
	  &valid,
	  &override,
	  &forwardMask );

	/* Abort on illegal format / values */
	if(macAddr > ((1LL << (8 * KSZ8864_MAC_LEN)) - 1LL)) {
		KSMSG("error: Illegal MAC address updating static MAC address entry\n");
		return -EFAULT;
	}
	if(items < 4) {
		if((items < 2) || (valid != 0)) {
			KSMSG("error: Format violation updating static MAC address entry\n");
			return -EFAULT;
		}
		override    = 0;
		forwardMask = 0;
	} else {
		if( (valid > 1) || (override > 1) ||
		    (forwardMask < KSZ8864_FORWARD_PORT1) || (forwardMask > KSZ8864_FORWARD_PORT_ALL) ) {
			KSMSG("error: Format violation updating static MAC address entry\n");
			return -EFAULT;
		}
	}

	res = srouteWrite( ks, (bool)valid, (bool)override, forwardMask, macAddr );
	if (res < 0) {
		return res;
	}

	return count;
}

/**
 * Handles reading of the dynamic routing table (dynamic MAC address table)
 * @param [out]    page   Where the data goes
 * @param [in,out] start
 * @param [in]     off    Read offset
 * @param [in]     count  Number of bytes to read
 * @param [in,out] eof    End-of-file flag
 * @param [in,out] data   Data pointer registered with this function
 * @return                Number of bytes read
 */
static int ks_proc_droute_read(struct seq_file *m, void *v)
{
	struct ethsw_net* const ks  = &ethsw;
	int                     i   = 0;

	if(m->read_pos == 0) {
		int result;
		LOCK_RESOURCE(ks->dev_lock);
		seq_printf(m, "MacAddress    Port  Timestamp\n");

		LOCK_RESOURCE(ks->ksz8864_lock);
		do {
			T_KSZ8864_DYNAMIC_MAC_ENTRY entry;
			result = ksz8864_readDynamicMacAddr(ks, i, &entry);
			if(result > 0) {
				seq_printf(m, "%02x%02x%02x%02x%02x%02x  %u     %u\n",
				  entry.macAddr[0],
				  entry.macAddr[1],
				  entry.macAddr[2],
				  entry.macAddr[3],
				  entry.macAddr[4],
				  entry.macAddr[5],
				  (unsigned) entry.sourcePort,
				  entry.timestamp );
			}
			i++;
		} while((result > 0) && ((result - i) > 0));

		RELEASE_RESOURCE(ks->ksz8864_lock);
		RELEASE_RESOURCE(ks->dev_lock);
	}
	return 0;
}

static int ks_proc_droute_open(struct inode *inode, struct file *file)
{
	return single_open(file, ks_proc_droute_read, PDE_DATA(inode));
}

/**
 * Function which will be called when the KSZ8864 issues an interrupt.
 * @param [in] irq     The interrupt number;
 * @param [in] cookie  A pointer to an arbitrary object;
 * @return status of interrupt handling.
 */
static irqreturn_t ethsw_isr(int irq, void* cookie)
{
	int const pinLevel = gpio_get_value(SWITCH_IRQ_PIN);


	if(!pinLevel) { /* Something change: Wake up the reader */
		struct ethsw_net *ks = &ethsw;
		schedule_work(&ks->ethsw_isr_wq);
		ks->ethsw_wq_count++;
	}
	return IRQ_HANDLED;
}

static void ethsw_update_linkstatus(void)
{
	struct ethsw_net*       ks = &ethsw;
	int                     j;
	unsigned                regValue;
#ifdef HAS_MAC
	unsigned int            linkUpChange = 0, linkDownChange = 0;
#endif
	static unsigned const   portOffsets[] = {
		KSZ8864_SMI_PORT1_OFFSET, /* ETHSW_EXT_PORT_1 */
		KSZ8864_SMI_PORT2_OFFSET, /* ETHSW_EXT_PORT_2 */
	};
	static int const speedLookup[]  = { 0,  0,  10,          100,         0,  10,          100,         0 };
	static int const duplexLookup[] = { -1, -1, DUPLEX_HALF, DUPLEX_HALF, -1, DUPLEX_FULL, DUPLEX_FULL, -1 };

	/* Append status data for the internal PHY's (Port 1 & 2) */
	for (j = ETHSW_EXT_PORT_1; j < (ETHSW_EXT_PORT_1 + ARRAY_SIZE(portOffsets)); j++) {

		regValue = ksz8864_readRegister(ks, KSZ8864_SMI_PORTX_STATUS_1           + portOffsets[j - ETHSW_EXT_PORT_1]);
		ks->link_statuses[ETHSW_EXT_PORT_1 + j].link = ((regValue & KSZ8864_SMI_PORTX_STATUS1_LINK_GOOD) > 0) ? 1 : 0;

		regValue = ksz8864_readRegister(ks, KSZ8864_SMI_PORTX_CONTROL_7_STATUS_2 + portOffsets[j - ETHSW_EXT_PORT_1]);
		ks->link_statuses[ETHSW_EXT_PORT_1 + j].speed  = speedLookup[KSZ8864_SMI_PORTX_STATUS_2_PORT_MODE(regValue)];
		ks->link_statuses[ETHSW_EXT_PORT_1 + j].duplex = duplexLookup[KSZ8864_SMI_PORTX_STATUS_2_PORT_MODE(regValue)];

#ifdef HAS_MAC
		if (ks->link_statuses[ETHSW_EXT_PORT_1 + j].link != ks->prev_link_statuses[ETHSW_EXT_PORT_1 + j]) {
			/* Link status change detected */
			if (ks->link_statuses[ETHSW_EXT_PORT_1 + j].link) {
				linkUpChange = 1;
			} else {
				linkDownChange = 1;
			}
		}
#endif
		ks->prev_link_statuses[ETHSW_EXT_PORT_1 + j] = ks->link_statuses[ETHSW_EXT_PORT_1 + j].link;
	}
}

/**
 * Preemptible tasklet that handles deferred KSZ8864 interrupt work.
 * @param [in] work    Unused
 */
static void ethsw_isr_tasklet(struct work_struct *work)
{
	struct ethsw_net*       ks = &ethsw;
#ifdef HAS_MAC
	unsigned int            linkUpChange = 0, linkDownChange = 0;
#endif

	LOCK_RESOURCE(ks->dev_lock);
	ks->ethsw_wq_count = 0;

	LOCK_RESOURCE(ks->ksz8864_lock);

	/* Clear the interrupt cause (Setting no bits results in setting the read value, which is what we want) */
	kszRegSetBits8(ks, KSZ8864_SMI_INTERRUPT_STATUS, 0);

    ethsw_update_linkstatus();
	
    RELEASE_RESOURCE(ks->ksz8864_lock);
	RELEASE_RESOURCE(ks->dev_lock);

#ifdef HAS_MAC
	if (linkUpChange || linkDownChange) {
		rtnl_lock();
		(void)netdev_bonding_change(ks->net_dev, NETDEV_UP);
		rtnl_unlock();
	}
#endif

	ks->read_flag = 1;
	wake_up_interruptible(&ks->read_queue);
}

#ifdef HAS_EXT_PHY
/**
 * Function which will be called when the PHY issues an interrupt.
 * @param [in] irq     The interrupt number;
 * @param [in] cookie  A pointer to an arbitrary object;
 * @return status of interrupt handling.
 */
static irqreturn_t phy_isr(int irq, void* cookie)
{
	int const pinLevel = gpio_get_value(PHY_IRQ_PIN);

	if(!pinLevel) { /* Something change: Wake up the reader */
		struct ethsw_net *ks = &ethsw;
		schedule_work(&ks->phy_isr_wq);
		ks->phy_wq_count++;
	}
	return IRQ_HANDLED;
}
#endif

/**
 * Preemptible tasklet that handles deferred PHY interrupt work.
 * @param [in] work    Unused
 */
static void phy_isr_tasklet(struct work_struct *work)
{
	struct ethsw_net* ks = &ethsw;
	unsigned          regValue;
	int               advertise, lpa, media;

	ks->phy_wq_count = 0;

	LOCK_RESOURCE(ks->dev_lock);
	LOCK_RESOURCE(ks->mii_lock);

	/* Clear the interrupt cause */
	regValue = p3phy_read(ks, KSZ8051_IRQ_STATUS_CONTROL);

	/* Append the auto-negotiation and link statuses */
	regValue = p3phy_read(ks, MII_BMSR);
	ks->link_statuses[ETHSW_EXT_PORT_3].link = (regValue & BMSR_LSTATUS) ? 1 : 0;

	/* Append the port mode */
	advertise = p3phy_read(ks, MII_ADVERTISE);
	lpa = p3phy_read(ks, MII_LPA);

	RELEASE_RESOURCE(ks->mii_lock);

	media = mii_nway_result(advertise & lpa);
	if(regValue & BMSR_LSTATUS)
	{
		switch(media) {
		case LPA_100FULL:
			ks->link_statuses[ETHSW_EXT_PORT_3].speed  = 100;
			ks->link_statuses[ETHSW_EXT_PORT_3].duplex = DUPLEX_FULL;
			break;
		case LPA_100HALF:
			ks->link_statuses[ETHSW_EXT_PORT_3].speed  = 100;
			ks->link_statuses[ETHSW_EXT_PORT_3].duplex = DUPLEX_HALF;
			break;
		case LPA_10FULL:
			ks->link_statuses[ETHSW_EXT_PORT_3].speed  = 10;
			ks->link_statuses[ETHSW_EXT_PORT_3].duplex = DUPLEX_FULL;
			break;
		case LPA_10HALF:
			ks->link_statuses[ETHSW_EXT_PORT_3].speed  = 10;
			ks->link_statuses[ETHSW_EXT_PORT_3].duplex = DUPLEX_HALF;
			break;
		default:
			ks->link_statuses[ETHSW_EXT_PORT_3].speed  = 0;
			ks->link_statuses[ETHSW_EXT_PORT_3].duplex = -1;
			break;
		}
	}
	else {
		ks->link_statuses[ETHSW_EXT_PORT_3].speed  = 0;
		ks->link_statuses[ETHSW_EXT_PORT_3].duplex = -1;
	}
	RELEASE_RESOURCE(ks->dev_lock);

	ks->read_flag = 1;
	wake_up_interruptible(&ks->read_queue);
}

/**
 * calc_txlen - calculate size of message to send packet
 * @len: Lenght of data
 *
 * Returns the size of the TXFIFO message needed to send
 * this packet.
 */
static inline unsigned calc_txlen(unsigned len) {
	return ALIGN(len + 4, 4);
}

/**
 * Interface function which is called when userspace function @p open() is called.
 * @param [in,out] inode  The inode
 * @param [in,out] fp     The file pointer;
 * @return Linux error status or 0 for none
 */
static int user_ethsw_open(struct inode *inode, struct file *fp)
{
	struct ethsw_net *ks = &ethsw;
	ks->read_flag = 1;
	wake_up_interruptible(&ks->read_queue);

	KSMSG("Device opened...\n");

	return 0;
}

/**
 * Interface function which is called when userspace function @p fread() is called.
 * @param [in]  fp     The file pointer;
 * @param [out] buff   The buffer in which the data needs to be put;
 * @param [in]  count  The number of bytes available in the @p buff;
 * @param [in]  offp   ;
 * @return Linux error status or otherwise the number of bytes copied to @p buff;
 */
static ssize_t user_ethsw_read(struct file *fp, char __user *buff, size_t count, loff_t *offp)
{
	char tmpBuf[DATA_SIZE];
	unsigned i = 0;
	unsigned j;
	struct ethsw_net *ks = &ethsw;

	/* Block until something changes (Unless file has been opened non-blocking) */
	if(!(fp->f_flags & O_NONBLOCK))	{
		wait_event_interruptible(ks->read_queue, ks->read_flag != 0);
		ks->read_flag = 0;
	}

	LOCK_RESOURCE(ks->dev_lock);

	/* Append status data for the internal PHY's (Port 1 & 2) */
	for (j = ETHSW_EXT_PORT_1; j <= ETHSW_EXT_PORT_3; j++) {

		/* Append the auto-negotiation and link statuses */
		tmpBuf[i++] = (ks->link_statuses[j].link != 0) ? LINK_STATUS_GOOD : LINK_STATUS_NONE;

		if(ks->link_statuses[j].speed == 100) {
			if(ks->link_statuses[j].duplex == 1) {
				tmpBuf[i++] = SPEED_MODE_100MB_FULL;
			} else {
				tmpBuf[i++] = SPEED_MODE_100MB_HALF;
			}
		} else if(ks->link_statuses[j].speed == 10) {
			if(ks->link_statuses[j].duplex == 1) {
				tmpBuf[i++] = SPEED_MODE_10MB_FULL;
			} else {
				tmpBuf[i++] = SPEED_MODE_10MB_HALF;
			}
		} else {
			tmpBuf[i++] = SPEED_MODE_UNDEFINED;
		}

		/* Append ':' To indicate next port */
		if(j < ETHSW_EXT_PORT_3) {
		tmpBuf[i++] = ':';
		}
	}
	RELEASE_RESOURCE(ks->dev_lock);

	/* Add CR to indicate end */
	tmpBuf[i++] = '\n';

	if (copy_to_user(buff, tmpBuf, i)) {
		return -EFAULT;
	}

	return i;
}

#ifdef HAS_MAC
static int dev_is_class(struct device *dev, void *class) {
	if (dev->class != NULL && !strcmp(dev->class->name, class))
		return 1;

	return 0;
}

static struct device *dev_find_class(struct device *parent, char *class) {
	if (dev_is_class(parent, class)) {
		get_device(parent);
		return parent;
	}

	return device_find_child(parent, class, dev_is_class);
}

static struct mii_bus *dev_to_mii_bus(struct device *dev) {
	struct device *d;

	d = dev_find_class(dev, "mdio_bus");
	if (d != NULL) {
		struct mii_bus *bus;

		bus = to_mii_bus(d);
		put_device(d);

		return bus;
	}

	return NULL;
}
#endif

/**
 Get Link status of specified MAC port.
 @param[in]  port   Port number to get link status for.
 @param[out] state  Current state of the port link.
 @return @li ETHSW_SUCCESS  Port link state has been retrieved correctly.
 	 	 @li ETHSW_FAIL     An invalid port number has been passed.
*/
int ethsw_getLinkStatus(const T_ETHSW_PORT port, T_ETHSW_LINK_STATE * const state)
{
	struct ethsw_net *ks = &ethsw;
	int portIdx;
#ifdef ETHSW_TRACE_ENABLE
	static T_ETHSW_LINK_STATE prevStates[KSZ8864_MAX_PORTS] =
		{ (T_ETHSW_LINK_STATE)-1, (T_ETHSW_LINK_STATE)-1, (T_ETHSW_LINK_STATE)-1, (T_ETHSW_LINK_STATE)-1 };
#endif

	if ((int)port < (int)ETHSW_PORT1 || (int)port > (int)ETHSW_PORT4)
	{
		KSERR("Invalid port number %d\n", (int)port);
		return ETHSW_FAIL;
	}
	portIdx = (int)port-(int)ETHSW_PORT1;

	LOCK_RESOURCE(ks->dev_lock);
	if (ks->link_statuses[portIdx].link) {
		*state = ETHSW_LINK_STATE_UP;
	} else {
		*state = ETHSW_LINK_STATE_DOWN;
	}
	RELEASE_RESOURCE(ks->dev_lock);

#ifdef ETHSW_TRACE_ENABLE
	if (prevStates[portIdx] != *state)
	{
		KSDBG("Link of port %d is %s\n", (int)port, (*state == ETHSW_LINK_STATE_UP) ? "Up" : "Down");
		prevStates[portIdx] = *state;
	}
#endif
	return ETHSW_SUCCESS;
}

/**
 * Get the negotiated link speed of the specified MAC port.
 * @param[in]  port   Port number to get link status for.
 * @param[out] speed  Current speed of the port link.
 * @return @li ETHSW_SUCCESS  Port link state has been retrieved correctly.
 *         @li ETHSW_FAIL     An invalid port number has been passed.
 */
int ethsw_getLinkInfo(const T_ETHSW_PORT port, T_ETHSW_LINK_SPEED * const speed)
{
	struct ethsw_net *ks = &ethsw;
	int portIdx;
#ifdef ETHSW_TRACE_ENABLE
	static T_ETHSW_LINK_SPEED prevSpeed[KSZ8864_MAX_PORTS] =
		{ (T_ETHSW_LINK_SPEED)-1, (T_ETHSW_LINK_SPEED)-1, (T_ETHSW_LINK_SPEED)-1, (T_ETHSW_LINK_SPEED)-1 };
#endif

	if ((int)port < (int)ETHSW_PORT1 || (int)port > (int)ETHSW_PORT4)
	{
		KSERR("Invalid port number %d\n", (int)port);
		return ETHSW_FAIL;
	}
	portIdx = (int)port-(int)ETHSW_PORT1;

	LOCK_RESOURCE(ks->dev_lock);
	switch(ks->link_statuses[portIdx].speed) {
	case 10:
		if(ks->link_statuses[portIdx].duplex == DUPLEX_FULL) {
			*speed = ETHSW_SPEED_10MB_FULL;
		} else {
			*speed = ETHSW_SPEED_10MB_HALF;
		}
		break;
	case 100:
		if(ks->link_statuses[portIdx].duplex == DUPLEX_FULL) {
			*speed = ETHSW_SPEED_100MB_FULL;
		} else {
			*speed = ETHSW_SPEED_100MB_HALF;
		}
		break;
	default:
		*speed = ETHSW_SPEED_UNKNOWN;
		break;
	}
	RELEASE_RESOURCE(ks->dev_lock);

#ifdef ETHSW_TRACE_ENABLE
	if (prevSpeed[portIdx] != *speed)
	{
		static char const* const speedLookup[] =
		{
			"Unknown",
			"10Mb half",
			"10Mb full",
			"100Mb half",
			"100Mb full",
		};
		unsigned int const speedIdx = ((*speed) - ETHSW_SPEED_UNKNOWN) % (sizeof(speedLookup) / sizeof(speedLookup[0]));
		KSDBG("Speed of port %d is %s\n", (int)port, speedLookup[speedIdx]);
		prevSpeed[portIdx] = *speed;
	}
#endif
	return ETHSW_SUCCESS;
}

/**
 Get MAC address of specified port.
 @param[in]  port  Port number to get MAC address for.
 @param[out] mac   Buffer to store MAC address in, minimum size ETHSW_MAC_LEN bytes, mac[0] = MSB, mac[5] = LSB.
 @return @li ETHSW_SUCCESS    MAC address was retrieval succeeded.
         @li ETHSW_FAILURE    Failed to retrieve MAC address, or an invalid port number has been passed.
*/
int ethsw_getMacAddr(const T_ETHSW_PORT port, unsigned char* mac)
{
	struct ethsw_net *ks = &ethsw;
	int idx;
	const unsigned smac_entry_indices[] = {
		SMAC_ENTRY_IDX_PORT1,	/* ETHSW_PORT1 */
		SMAC_ENTRY_IDX_PORT2,	/* ETHSW_PORT2 */
		SMAC_ENTRY_IDX_PORT3,	/* ETHSW_PORT3 */
	};

	if ((int)port < (int)ETHSW_PORT1 || (int)port > (int)ETHSW_PORT3)
	{
		/* Note port 4 has no MAC! */
		KSERR("Invalid port number %d to get MAC address for\n", (int)port);
		return ETHSW_FAIL;
	}

	idx = smac_entry_indices[(int)port-(int)ETHSW_PORT1];

	LOCK_RESOURCE(ks->dev_lock);
	MAC_LONGLONG_TO_CHAR_ARRAY(mac, ks->smac_entries[idx].macAddr);
	RELEASE_RESOURCE(ks->dev_lock);

	return ETHSW_SUCCESS;
}

/**
  Clear static and/or dynamic MAC tables.
  @param[in] tableMask	Tables to clear.
*/
void ethsw_clearMacTable(const unsigned tableMask)
{
	struct ethsw_net *ks = &ethsw;
	unsigned bits = 0;
	int i;
	const unsigned control2Regs[] = {
	  KSZ8864_SMI_PORTX_CONTROL_2 + KSZ8864_SMI_PORT1_OFFSET,
	  KSZ8864_SMI_PORTX_CONTROL_2 + KSZ8864_SMI_PORT2_OFFSET,
	  KSZ8864_SMI_PORTX_CONTROL_2 + KSZ8864_SMI_PORT3_OFFSET,
	  KSZ8864_SMI_PORTX_CONTROL_2 + KSZ8864_SMI_PORT4_OFFSET,
	};
	unsigned keepControl2RegValues[KSZ8864_MAX_PORTS];

	LOCK_RESOURCE(ks->ksz8864_lock);

	if (tableMask & ETHSW_MAC_TABLE_STATIC) {
		bits |= KSZ8864_SMI_CONTROL_0_GLOBAL_FLUSH_SMAC_TABLE;
		KSDBG("Clear static MAC table\n");
	}
	if (tableMask & ETHSW_MAC_TABLE_DYNAMIC) {
		bits |= KSZ8864_SMI_CONTROL_0_GLOBAL_FLUSH_DMAC_TABLE;
		/* First disable learning on ports, enable it again after flushing. */
		for (i = 0; i < KSZ8864_MAX_PORTS; i++)	{
			/* Change STP mode of port. */
			keepControl2RegValues[i] = ksz8864_readRegister(ks, control2Regs[i]);
			kszRegSetBits8(ks, control2Regs[i], KSZ8864_SMI_PORTX_CONTROL_2_LEARNING_DISABLE);
		}
		KSDBG("Clear dynamic MAC table\n");
	}
	if (bits) {
		/* Trigger the static/dynamic MAC table clear */
		kszRegSetBits8(ks, KSZ8864_SMI_CONTROL_0_GLOBAL, bits);
	}
	if (bits & KSZ8864_SMI_CONTROL_0_GLOBAL_FLUSH_DMAC_TABLE) {
		/* Restore learning feature after flushing. */
		for (i = 0; i < KSZ8864_MAX_PORTS; i++)	{
			ksz8864_writeRegister(ks, control2Regs[i], keepControl2RegValues[i]);
		}
	}
	if (bits & KSZ8864_SMI_CONTROL_0_GLOBAL_FLUSH_SMAC_TABLE) {
		int i;
		LOCK_RESOURCE(ks->dev_lock);
		/* Keep the first entries in the static table as they contain the MAC addresses
		   of the internal MACs, mark all other static MAC table entries as invalid.
		   Mark all other static MAC table entries as invalid. */
		for(i = KSZ8864_MAX_PORTS; i < sizeof(ks->smac_entries) / sizeof(ks->smac_entries[0]); i++) {
			ks->smac_entries[i].valid = KSZ8864_SMAC_INVALID;
		}
		/* Write the table. */
		(void)ksz8864_setStaticEntries( ks,
										ks->smac_entries,
										sizeof(ks->smac_entries)/sizeof(ks->smac_entries[0]) );
		RELEASE_RESOURCE(ks->dev_lock);
	}
	RELEASE_RESOURCE(ks->ksz8864_lock);
}

/**
  Add an entry to the static MAC table.
  @param[in] mac          MAC address to add, mac[0] = MSB, mac[5] = LSB.
  @param[in] forwardMask  Bitmask indicating which ports to forward packets with given destination MAC
                          address to.
  @param[in] override     Override the RTSP setting.
  @return @li ETHSW_SUCCESS   Adding of MAC table entry successful.
          @li ETHSW_FAIL      Failed to add MAC table entry (e.g. table full).
*/
int	ethsw_addStaticMacTableEntry(unsigned char const * const mac, const unsigned forwardMask, const bool override )
{
	struct ethsw_net *ks = &ethsw;
	return srouteWrite( ks, true /*valid*/, override, forwardMask, MAC_CHAR_ARRAY_TO_LONGLONG(mac) );
}

/**
  Delete an entry from the static MAC table. The entry will not really be deleted, but only marked invalid.
  @param[in] mac          MAC address to remove, mac[0] = MSB, mac[5] = LSB.
  @return @li ETHSW_SUCCESS   Removal of MAC table entry successful.
          @li ETHSW_FAILURE   Failed to remove MAC table entry.
*/
int ethsw_delStaticMacTableEntry(unsigned char const * const mac)
{
	struct ethsw_net *ks = &ethsw;
	return srouteWrite( ks, false /*valid*/, 0, 0, MAC_CHAR_ARRAY_TO_LONGLONG(mac) );
}

/**
 Set port RTSP state.
 @param[in] port   Port number to set state for.
 @param[in] state  New port state.
 @return @li ETHSW_SUCCESS   Port mode setting successful.
         @li ETHSW_FAIL      An invalid port number/state value has been passed.
 */
int	ethsw_setPortState(const T_ETHSW_PORT port, const T_ETHSW_PORT_STATE state)
{
	struct ethsw_net *ks = &ethsw;
	unsigned regValue, portIdx, stateIdx;
	const unsigned port_stp_bits[] = {
		KSZ8864_SMI_PORTX_CONTROL_2_RSTP_STATE_FORWARDING,     /* T_ETHSW_PORT_STATE_FORWARDING */
		KSZ8864_SMI_PORTX_CONTROL_2_RSTP_STATE_DISABLE,        /* T_ETHSW_PORT_STATE_DISABLE */
		KSZ8864_SMI_PORTX_CONTROL_2_RSTP_STATE_BLOCKING,       /* T_ETHSW_PORT_STATE_BLOCKING */
		KSZ8864_SMI_PORTX_CONTROL_2_RSTP_STATE_LISTENING,      /* T_ETHSW_PORT_STATE_LISTENING */
		KSZ8864_SMI_PORTX_CONTROL_2_RSTP_STATE_LEARNING,       /* T_ETHSW_PORT_STATE_LEARNING */
	};
	const unsigned control2Regs[] = {
		KSZ8864_SMI_PORTX_CONTROL_2 + KSZ8864_SMI_PORT1_OFFSET,
		KSZ8864_SMI_PORTX_CONTROL_2 + KSZ8864_SMI_PORT2_OFFSET,
		KSZ8864_SMI_PORTX_CONTROL_2 + KSZ8864_SMI_PORT3_OFFSET,
		KSZ8864_SMI_PORTX_CONTROL_2 + KSZ8864_SMI_PORT4_OFFSET,
	};
#ifdef ETHSW_TRACE_ENABLE
	const char* port_state_names[] = {
		"FORWARDING",
		"DISABLE",
		"BLOCKING",
		"LISTENING",
		"LEARNING",
	};
#endif

	if ((int)port < (int)ETHSW_PORT1 || (int)port > (int)ETHSW_PORT4)
	{
		KSERR("Invalid port number %d\n", (int)port);
		return ETHSW_FAIL;
	}
	if ((int)state >= (int)T_ETHSW_PORT_STATE_MAX)
	{
		KSERR("Invalid port state %d\n", (int)state);
		return ETHSW_FAIL;
	}

	/* Change RSTP mode of port. */
	portIdx  = (unsigned)port - (unsigned)ETHSW_PORT1;
	stateIdx = (unsigned)state - (unsigned)T_ETHSW_PORT_STATE_FORWARDING;

	LOCK_RESOURCE(ks->ksz8864_lock);

	regValue = ksz8864_readRegister(ks, control2Regs[portIdx]);
	regValue &= ~(KSZ8864_SMI_PORTX_CONTROL_2_RSTP_STATE_MASK);
	regValue |= port_stp_bits[stateIdx];
	ksz8864_writeRegister(ks, control2Regs[portIdx], regValue);

	RELEASE_RESOURCE(ks->ksz8864_lock);

#ifdef ETHSW_TRACE_ENABLE
	KSDBG("Port %d set to %s state\n", (int)port, port_state_names[stateIdx]);
#endif
	return ETHSW_SUCCESS;
}


#ifdef HAS_MAC
/**
 * Retrieves the MAC addresses configured by the boot-loader from static MAC table
 * @param [in,out] ks   The device instance
 */
static int probe_smac_table(struct ethsw_net* const ks)
{
	int i;

	BUG_ON_NOT_LOCKED(ks->ksz8864_lock);
	BUG_ON_NOT_LOCKED(ks->dev_lock);

	for(i = 0; i < sizeof(ks->smac_entries) / sizeof(ks->smac_entries[0]); i++) {
		if(ksz8864_readStaticMacAddr(ks, i, ks->smac_entries + i) == KSZ8864_SMAC_VALID) {
			if(i < KSZ8864_MAX_PORTS) {
				KSMSG("port %d MAC address: %012"PRIx64, i, ks->smac_entries[i].macAddr);
			}
		} else {
			if(i < KSZ8864_MAX_PORTS) {
				KSMSG("Missing mandatory static MAC address table entry for port %d\n", i);
				return -1;
			}
		}
	}
	return 0;
}
#endif

/**
 * Populates the procfs interfaces for the driver
 * @param [in,out] ks   The device instance
 */
static void probe_proc_entries(struct ethsw_net* const ks)
{
	int i;

	BUG_ON_NOT_LOCKED(ks->dev_lock);

	ks->proc_dir = proc_mkdir(KS_PROC_DIR, NULL);
	if(ks->proc_dir == NULL) {
		KSMSG("Unable to create /proc/%s\n", KS_PROC_DIR);
		return;
	}
	ks->regs_dir = proc_mkdir(KS_REGS_DIR, ks->proc_dir);
	if(ks->regs_dir == NULL) {
		KSMSG("Unable to create /proc/%s\n", KS_REGS_DIR);
		return;
	}
	
	for(i = 0; i < sizeof(ks_proc_reg_entries) / sizeof(ks_proc_reg_entries[0]); i++) {
		if(proc_create_data(ks_proc_reg_entries[i].name, S_IRUGO | S_IWUGO, ks->regs_dir, &ks_proc_reg_entries[i].op, ks_proc_reg_entries[i].data) == NULL) {
			KSMSG("error: Unable to create /proc/%s/%s", KS_PROC_DIR, ks_proc_reg_entries[i].name);
		}
	}
	ks->mii_dir = proc_mkdir(KS_MII_DIR, ks->proc_dir);
	if(ks->mii_dir == NULL) {
		KSMSG("Unable to create /proc/%s\n", KS_MII_DIR);
		return;
	}
	for(i = 0; i < sizeof(ks_proc_mii_entries) / sizeof(ks_proc_mii_entries[0]); i++) {
		if(proc_create_data(ks_proc_mii_entries[i].name, S_IRUGO | S_IWUGO, ks->mii_dir, &ks_proc_mii_entries[i].op, ks_proc_mii_entries[i].data) == NULL) {
			KSMSG("error: Unable to create /proc/%s/%s", KS_PROC_DIR, ks_proc_mii_entries[i].name);
		}
	}
	ks->ctrl_dir = proc_mkdir(KS_CTRL_DIR, ks->proc_dir);
	if(ks->ctrl_dir == NULL) {
		KSMSG("Unable to create /proc/%s\n", KS_CTRL_DIR);
		return;
	}
	for(i = 0; i < sizeof(ks_proc_ctl_entries) / sizeof(ks_proc_ctl_entries[0]); i++) {
		if(proc_create_data(ks_proc_ctl_entries[i].name, S_IRUGO | S_IWUGO, ks->ctrl_dir, &ks_proc_ctl_entries[i].op, ks_proc_ctl_entries[i].data) == NULL) {
			KSMSG("error: Unable to create /proc/%s/%s", KS_PROC_DIR, ks_proc_ctl_entries[i].name);
		}
	}
	ks->mib_dir = proc_mkdir(KS_MIB_DIR, ks->proc_dir);
	if(ks->mib_dir == NULL) {
		KSMSG("Unable to create /proc/%s\n", KS_MIB_DIR);
		return;
	}
	for(i = 0; i < sizeof(ks_proc_mib_entries) / sizeof(ks_proc_mib_entries[0]); i++) {
		if(proc_create_data(ks_proc_mib_entries[i].name, S_IRUGO | S_IWUGO, ks->mib_dir, &ks_proc_mib_entries[i].op, ks_proc_mib_entries[i].data) == NULL) {
			KSMSG("error: Unable to create /proc/%s/%s", KS_PROC_DIR, ks_proc_mib_entries[i].name);
		}
	}
	for(i = 0; i < sizeof(ks_proc_entries) / sizeof(ks_proc_entries[0]); i++) {
		if(proc_create_data(ks_proc_entries[i].name, S_IRUGO | S_IWUGO, ks->proc_dir, &ks_proc_entries[i].op, ks_proc_entries[i].data) == NULL) {
			KSMSG("error: Unable to create /proc/%s/%s", KS_PROC_DIR, ks_proc_entries[i].name);
		}
	}
}

static void ethsw_probe_setmac(void)
{
    int slen, res;
	struct ethsw_net *ks = &ethsw;

    res = 0;
    slen = strlen(am18xx_mac);

    if (slen == 17) {
        unsigned long arr[6];
        am18xx_mac[2] = am18xx_mac[5] = am18xx_mac[8] = am18xx_mac[11] = am18xx_mac[14] = '\0';
        res += kstrtoul/*strict_strtoul*/(&am18xx_mac[0], 16, &arr[0]);
        res += kstrtoul/*strict_strtoul*/(&am18xx_mac[3], 16, &arr[1]);
        res += kstrtoul/*strict_strtoul*/(&am18xx_mac[6], 16, &arr[2]);
        res += kstrtoul/*strict_strtoul*/(&am18xx_mac[9], 16, &arr[3]);
        res += kstrtoul/*strict_strtoul*/(&am18xx_mac[12], 16, &arr[4]);
        res += kstrtoul/*strict_strtoul*/(&am18xx_mac[15], 16, &arr[5]);

        if (!res) {
            T_KSZ8864_STATIC_MAC_ENTRY const    entry   = {
                .valid       = KSZ8864_SMAC_VALID,
                .override    = KSZ8864_PORT_STATE_DONT_OVERRIDE,
                .forwardMask = KSZ8864_FORWARD_PORT4,
                .macAddr     = ((((unsigned long long)arr[0] >>  0U) & 0xFF) << 40U) |
                               ((((unsigned long long)arr[1] >>  8U) & 0xFF) << 32U) |
                               ((((unsigned long long)arr[2] >> 16U) & 0xFF) << 24U) |
                               ((((unsigned long long)arr[3] >> 24U) & 0xFF) << 16U) |
                               ((((unsigned long long)arr[4] >>  0U) & 0xFF) <<  8U) |
                               ((((unsigned long long)arr[5] >>  8U) & 0xFF) <<  0U),
            };
            
            KSMSG("Programming static route for own mac='%s'\n", am18xx_mac);
            ksz8864_addStaticMacAddr(ks, 0, &entry);
        }
    }
}

/**
 * This function is called by Linux during installation of the driver.
 * Its function is to see if the KSZ8864 is connected and if so, register the driver.
 * @param [in] spi    A pointer to the BSP SPI-driver which will be used during communication with the switch;
 * @return A linux error code or @p 0 when successful.
 */
static int /*__devinit */ethsw_probe(struct spi_device *spi)
{
	struct ethsw_net *ks = &ethsw;
	unsigned revId;
#ifdef HAS_EXT_PHY
    unsigned phyId;
#endif
	int ret;
    int irq;

	KSMSG("[2] detecting Micrel KSZ8864 on SPI bus using SPI controller %d and CS %d\n",
			spi->master->bus_num, spi->chip_select);

	ks->reset_result      = NULL;
	ks->dmac_flush_result = NULL;
	ks->spidev = spi;

	mutex_init(&ks->dev_lock);
	LOCK_RESOURCE(ks->dev_lock);

	mutex_init(&ks->ksz8864_lock);
	LOCK_RESOURCE(ks->ksz8864_lock);

	mutex_init(&ks->mii_lock);
	LOCK_RESOURCE(ks->mii_lock);

	ks->read_flag = 0;
	init_waitqueue_head(&ks->read_queue);

	/* Clear the transfer struct. */
	memset(&ks->spi_xfer, 0x00, sizeof(ks->spi_xfer));

	/* initialise pre-made spi transfer messages */
	spi_message_init(&ks->spi_msg);
	spi_message_add_tail(&ks->spi_xfer, &ks->spi_msg);

	/* simple check for a valid chip being connected to the bus */
	if ((ksz8864_readRegister(ks, KSZ8864_SMI_CHIP_ID0)) != KSZ8864_CHIP_ID) {
		KSERR("failed to read device ID\n");
		ret = -ENODEV;
		goto err;
	}
	else {
		KSMSG("found the device\n");
	}

	/* Read the revision */
	revId = KSZ8864_SMI_REV_ID_START_REV_ID(ksz8864_readRegister(ks, KSZ8864_SMI_REV_ID_START));
	KSMSG("revision of switch device: %d\n",revId);


	/* Acquire the switch IRQ pin and set the mask */
	if(gpio_request(SWITCH_IRQ_PIN, SWITCH_IRQ_NAME)) {
		KSERR("Couldn't get switch IRQ pin\n");
		ret = -ENODEV;
		goto err_with_reset_pin;
	}
    //gpio_set_value(SWITCH_IRQ_PIN, 0);

	if(gpio_direction_input(SWITCH_IRQ_PIN)) {
		KSERR("Couldn't set switch IRQ pin to input\n");
		ret = -ENODEV;
		goto err_with_ksz_irq_pin;
	}

    /* Issue 7492: start-up sequence of switch */
	ksz8864_start(ks);

    /* Program our MAC address to the static routes table */
    ethsw_probe_setmac();

	kszRegSetBits8(ks, KSZ8864_SMI_INTERRUPT_MASK, KSZ8864_SMI_INTERRUPT_MASK_PORT1 | KSZ8864_SMI_INTERRUPT_MASK_PORT2);
	/* Clear the interrupt cause (Setting no bits results in setting the read value, which is what we want) */
	kszRegSetBits8(ks, KSZ8864_SMI_INTERRUPT_STATUS, 0);
	dev_set_drvdata(&spi->dev, ks);


#ifdef HAS_MAC
	/* Probe the 3rd port PHY, retrieve the Ethernet device struct and increase usage count */
	ks->net_dev = dev_get_by_name(&init_net, KS_INTERFACE_ETH);
	if (ks->net_dev == NULL) {
		KSERR("error: Ethernet device not found\n");
		ret = -ENODEV;
		goto err_with_ksz_irq_pin;
	}
	ks->mii_bus = dev_to_mii_bus(&ks->net_dev->dev);
	if (ks->mii_bus == NULL) {
		KSERR("error: MII bus not found\n");
		ret = -ENODEV;
		goto err_with_ksz_irq_pin;
	}
	ks->phydev = ks->mii_bus->phy_map[PHY_ADDR];
	if (ks->phydev == NULL) {
		KSERR("PHY not found at %s, MDIO ID %d\n", KS_INTERFACE_ETH, PHY_ADDR);
		ret = -ENODEV;
		goto err_with_ksz_irq_pin;
	}
	phyId = KSZ8051_GET_ID(
	          mdiobus_read(ks->mii_bus, PHY_ADDR, KSZ8051_PHY_ID_1),
	          mdiobus_read(ks->mii_bus, PHY_ADDR, KSZ8051_PHY_ID_2) );
	if(phyId != KSZ8051_ID) {
		KSERR("Unsupported PHY (ID = 0x%x) at %s, MDIO ID %d\n", phyId, KS_INTERFACE_ETH, PHY_ADDR);
		ret = -ENODEV;
		goto err_with_ksz_irq_pin;
	}

	/* Acquire the port 3 PHY IRQ pin and set the mask */
	/* TODO: Check PHY ID's */
#ifdef HAS_EXT_PHY
	KSERR("PHY found for third port\n");
	if(gpio_request(PHY_IRQ_PIN, PHY_IRQ_NAME)) {
		KSERR("Couldn't get PHY IRQ pin\n");
		ret = -ENODEV;
		goto err_with_ksz_irq_pin;
	}
	if(at91_set_GPIO_periph(PHY_IRQ_PIN, 0)) {
		KSERR("Couldn't disable peripheral on PHY IRQ pin\n");
		ret = -ENODEV;
		goto err_with_ksz_irq_pin;
	}
	if(gpio_direction_input(PHY_IRQ_PIN)) {
		KSERR("Couldn't set PHY IRQ pin to input\n");
		ret = -ENODEV;
		goto err_with_ksz_irq_pin;
	}
#endif
	phyRegSetBits(ks, KSZ8051_IRQ_STATUS_CONTROL, KSZ8051_IRQ_CONTROL_LINK_DOWN | KSZ8051_IRQ_CONTROL_LINK_UP);

	if(probe_smac_table(ks))
	{
		KSERR("Couldn't retrieve port MAC addresses from static MAC address table\n");
		ret = -ENODEV;
		goto err_with_ksz_irq_pin;
	}
#endif
	probe_proc_entries(ks);

	irq = gpio_to_irq(SWITCH_IRQ_PIN);
	if(request_irq(irq, ethsw_isr, /*IRQF_DISABLED*/0x00000020 | /*IRQF_TRIGGER_RISING |*/ IRQF_TRIGGER_FALLING,
                "ethsw", NULL)) {
		KSERR("Couldn't request interrupt on switch IRQ pin\n");
		ret = -ENODEV;
		goto err_with_ksz_irq_pin;
	}
    ethsw_update_linkstatus();


#ifdef HAS_MAC
#ifdef HAS_EXT_PHY
	if(request_irq(PHY_IRQ_PIN, phy_isr, /*IRQF_DISABLED*/0x00000020, "ethsw ext PHY", NULL)) {
		KSERR("Couldn't request interrupt on switch IRQ pin\n");
		ret = -ENODEV;
		goto err_with_ksz_irq;
	}
#endif

#ifdef HAS_MAC
	/* Register callbacks in MAC driver to process tail tags on incoming & outgoing packets */
	macb_registerRxCallback(ks->net_dev, onRxMac, (void*)ks);
	macb_registerTxCallback(ks->net_dev, onTxMac, (void*)ks);
	(void)tailTagEnable(ks, true);
#endif

	/* Queue the ISR Retrieve initial link statuses and clear IRQ's */
	ks->phy_wq_count++;
	schedule_work(&ks->phy_isr_wq);
	ks->ethsw_wq_count++;
	schedule_work(&ks->ethsw_isr_wq);
#endif
	ret = 0;
	goto out;

#ifdef HAS_MAC
err_with_ksz_irq:
	free_irq(SWITCH_IRQ_PIN, NULL);
#endif

#ifdef HAS_EXT_PHY
err_with_phy_irq_pin:
	phyRegClearBits(ks, KSZ8051_IRQ_STATUS_CONTROL, KSZ8051_IRQ_CONTROL_LINK_DOWN | KSZ8051_IRQ_CONTROL_LINK_UP);
	gpio_free(PHY_IRQ_PIN);
#endif

err_with_ksz_irq_pin:
	kszRegClearBits8(ks, KSZ8864_SMI_INTERRUPT_MASK, KSZ8864_SMI_INTERRUPT_MASK_PORT1 | KSZ8864_SMI_INTERRUPT_MASK_PORT2);
	gpio_free(SWITCH_IRQ_PIN);

err_with_reset_pin:
	gpio_free(ETH_RESET_PIN);

err:
	KSERR("error!\n");

out:
	RELEASE_RESOURCE(ks->mii_lock);
	RELEASE_RESOURCE(ks->ksz8864_lock);
	RELEASE_RESOURCE(ks->dev_lock);

	return ret;
}

/*
 * Cleans up the procfs interfaces for the driver
 */
static void remove_proc_entries(struct ethsw_net* const ks)
{
	int i;

	BUG_ON_NOT_LOCKED(ks->dev_lock);

	if(ks->proc_dir == NULL)
		return;

	for(i = 0; i < sizeof(ks_proc_entries) / sizeof(ks_proc_entries[0]); i++) {
		remove_proc_entry(ks_proc_entries[i].name, ks->proc_dir);
	}

	for(i = 0; i < sizeof(ks_proc_mib_entries) / sizeof(ks_proc_mib_entries[0]); i++) {
		remove_proc_entry(ks_proc_mib_entries[i].name, ks->mib_dir);
	}
	remove_proc_entry(KS_MIB_DIR, ks->proc_dir);

	for(i = 0; i < sizeof(ks_proc_ctl_entries) / sizeof(ks_proc_ctl_entries[0]); i++) {
		remove_proc_entry(ks_proc_ctl_entries[i].name, ks->ctrl_dir);
	}
	remove_proc_entry(KS_CTRL_DIR, ks->proc_dir);

	for(i = 0; i < sizeof(ks_proc_mii_entries) / sizeof(ks_proc_mii_entries[0]); i++) {
		remove_proc_entry(ks_proc_mii_entries[i].name, ks->mii_dir);
	}
	remove_proc_entry(KS_MII_DIR, ks->proc_dir);

	for(i = 0; i < sizeof(ks_proc_reg_entries) / sizeof(ks_proc_reg_entries[0]); i++) {
		remove_proc_entry(ks_proc_reg_entries[i].name, ks->regs_dir);
	}
	remove_proc_entry(KS_REGS_DIR, ks->proc_dir);

	remove_proc_entry(KS_PROC_DIR, NULL);
}

/**
 * Interface function for Linux to uninstall the driver.
 * @param [in] spi    A pointer to the BSP SPI-driver which will be used during communication with the switch;
 * @return A linux error code or @p 0 when successful.
 */
static int /*__devexit */ethsw_remove(struct spi_device *spi)
{
	struct ethsw_net *ks = &ethsw;
	dev_t devno = MKDEV(ETHSW_DEV_MAJOR, ETHSW_DEV_MINOR);

	LOCK_RESOURCE(ks->dev_lock);
	LOCK_RESOURCE(ks->ksz8864_lock);
	LOCK_RESOURCE(ks->mii_lock);

#ifdef HAS_MAC
	(void)tailTagEnable(ks, false);
	/* Uregister callbacks in MAC driver */
	macb_registerRxCallback(ks->net_dev, NULL, NULL);
	macb_registerTxCallback(ks->net_dev, NULL, NULL);
#endif

	remove_proc_entries(ks);
	unregister_chrdev_region(devno, ETHSW_NUM_MINORS);
	cdev_del(dev);
	kszRegClearBits8(ks, KSZ8864_SMI_INTERRUPT_MASK, KSZ8864_SMI_INTERRUPT_MASK_PORT1 | KSZ8864_SMI_INTERRUPT_MASK_PORT2);
	free_irq(SWITCH_IRQ_PIN, NULL);
	gpio_free(SWITCH_IRQ_PIN);
#ifdef HAS_EXT_PHY
	phyRegClearBits(ks, KSZ8051_IRQ_STATUS_CONTROL, KSZ8051_IRQ_CONTROL_LINK_DOWN | KSZ8051_IRQ_CONTROL_LINK_UP);
	free_irq(PHY_IRQ_PIN, NULL);
	gpio_free(PHY_IRQ_PIN);
#endif
	gpio_free(ETH_RESET_PIN);

	RELEASE_RESOURCE(ks->mii_lock);
	RELEASE_RESOURCE(ks->ksz8864_lock);
	RELEASE_RESOURCE(ks->dev_lock);

	KSMSG("unregistered driver\n");

	return 0;
}

/*=== Global function definitions (Global system or Local sub-system scope) ==========================================*/

/**
 * Interface function for Linux to initialise the driver and register the appropriate interface functions.
 * @return A linux error code or @p 0 when successful.
 */
static int __init ethsw_init(void)
{
	int spiStatus;
	int err, err_res;
	dev_t devno;

    err_res = 0;

	ethsw.rxd = (u8*)kmalloc(ETHSW_SMI_RX_BUF_SIZE, GFP_KERNEL);
	if (!ethsw.rxd) {
		KSERR("could not reserve memory for rx buffer\n");
	}

	ethsw.txd = (u8*)kmalloc(ETHSW_SMI_TX_BUF_SIZE, GFP_KERNEL);
	if (!ethsw.txd) {
		KSERR("could not reserve memory for tx buffer\n");
	} else if (ethsw.rxd) {
		KSMSG("reserved memory for spi buffers\n");
	}

    /* Implement behavior of #7492; Micrell power-on reset sequence */
	if(gpio_request(ETH_RESET_PIN, ETH_RESET_NAME)) {
		KSERR("Couldn't get reset pin\n");
		err_res = -ENODEV;
	} else {
        /* Issue reset pulse, and release from reset. */
        gpio_direction_input(ETH_RESET_PIN);
        udelay(100);
        gpio_direction_output(ETH_RESET_PIN, 0);
        udelay(100);
        gpio_direction_input(ETH_RESET_PIN);
    }

	ethsw.phy_wq_count   = 0;
	INIT_WORK(&ethsw.phy_isr_wq,   phy_isr_tasklet);

	ethsw.ethsw_wq_count = 0;
	INIT_WORK(&ethsw.ethsw_isr_wq, ethsw_isr_tasklet);

	spiStatus = spi_register_driver(&ethsw_driver);

	devno = MKDEV(ETHSW_DEV_MAJOR, ETHSW_DEV_MINOR);

	err = register_chrdev_region(devno, ETHSW_NUM_MINORS, "Omron_KSZ8864_driver");
	if (!err) {
		dev        = cdev_alloc();
		dev->ops   = &user_ethsw_file_ops;
		dev->owner = THIS_MODULE;
		err = cdev_add(dev, devno, 1);
		printk(KERN_ALERT "cdev_add\n");
		if (err) {
		printk(KERN_ALERT "cdev_add error\n");
			unregister_chrdev_region(devno, ETHSW_NUM_MINORS);
		}
	} else {
		unregister_chrdev_region(devno, ETHSW_NUM_MINORS);
	}

  if (err) 
		return err;
  else if (err_res)
      return err_res;
	else if (spiStatus)
		return spiStatus;

	KSMSG("initialisation ready\n");

	return 0;
}

/**
 * Interface function for Linux to free and unload the driver.
 */
static void __exit ethsw_exit(void)
{
	struct ethsw_net *ks = &ethsw;


	if (ethsw.rxd) {
		kfree(ethsw.rxd);
	}
	else
	KSERR("no rx memory to free\n");

	if (ethsw.txd) {
		kfree(ethsw.txd);
	}
	else
	KSERR("no tx memory to free\n");

	KSMSG("driver removed\n");
	/* release usage of network device */
	if (ks->net_dev != NULL) {
		dev_put(ks->net_dev);
	}
	spi_unregister_driver(&ethsw_driver);
}

module_init(ethsw_init);
module_exit(ethsw_exit);

EXPORT_SYMBOL (ethsw_getLinkStatus);
EXPORT_SYMBOL (ethsw_getLinkInfo);
EXPORT_SYMBOL (ethsw_getMacAddr);
EXPORT_SYMBOL (ethsw_clearMacTable);
EXPORT_SYMBOL (ethsw_addStaticMacTableEntry);
EXPORT_SYMBOL (ethsw_delStaticMacTableEntry);
EXPORT_SYMBOL (ethsw_setPortState);

//MODULE_PARAM_DESC(param_mac, "Mac address to statically route");

MODULE_DESCRIPTION("KSZ8864 Switch driver");
MODULE_AUTHOR("Pieter Smith <pieter@boesman.nl>");
MODULE_LICENSE("GPL");

MODULE_ALIAS("spi:ethsw");

